/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>
             Michael Tross, digitalSTROM AG <michael.tross@digitalstrom.com>
             Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.
    

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#include "light.h"
#include "scene.h"
#include "dsm.h"
#include "selectA.h"
#include "datamodel.h"
#include "registers.h"
#include "bsp.h"
#include "bitset8.h"
#include "button.h"
#include "ds485.h"

Q_DEFINE_THIS_MODULE(light)

/* FSM state-handler functions */
static QState Light_initial(Light *me, QEvent const *e);

static QState Light_top(Light *me, QEvent const *e);

static QState Light_special_room_state(Light *me, QEvent const *e);
static QState Light_special_room_state_return_scene_0(Light *me, QEvent const *e);
static QState Light_special_room_state_return_scene_1(Light *me, QEvent const *e);
static QState Light_special_room_state_return_scene_2(Light *me, QEvent const *e);
static QState Light_special_room_state_return_scene_3(Light *me, QEvent const *e);
static QState Light_special_room_state_return_scene_4(Light *me, QEvent const *e);

static QState Light_off(Light *me, QEvent const *e);
static QState Light_scene_0(Light *me, QEvent const *e);
static QState Light_scene_0_local(Light *me, QEvent const *e);
static QState Light_local_prio_off_hold(Light *me, QEvent const *e);
static QState Light_standby_hold(Light *me, QEvent const *e);
static QState Light_deep_off_hold(Light *me, QEvent const *e);

static QState Light_on(Light *me, QEvent const *e);
static QState Light_scene_1(Light *me, QEvent const *e);
static QState Light_scene_1_local(Light *me, QEvent const *e);
static QState Light_scene_2(Light *me, QEvent const *e);
static QState Light_scene_2_local(Light *me, QEvent const *e);
static QState Light_scene_3(Light *me, QEvent const *e);
static QState Light_scene_3_local(Light *me, QEvent const *e);
static QState Light_scene_4(Light *me, QEvent const *e);
static QState Light_scene_4_local(Light *me, QEvent const *e);

static QState Light_dimming_down(Light *me, QEvent const *e);
static QState Light_dimming_up(Light *me, QEvent const *e);
static QState Light_scene_0_dimming_down(Light *me, QEvent const *e);
static QState Light_scene_0_dim_up(Light *me, QEvent const *e);
static QState Light_scene_0_dimming_up(Light *me, QEvent const *e);
static QState Light_scene_1_dim_down(Light *me, QEvent const *e);
static QState Light_scene_1_dimming_down(Light *me, QEvent const *e);
static QState Light_scene_1_dim_up(Light *me, QEvent const *e);
static QState Light_scene_1_dimming_up(Light *me, QEvent const *e);
static QState Light_scene_2_dim_down(Light *me, QEvent const *e);
static QState Light_scene_2_dimming_down(Light *me, QEvent const *e);
static QState Light_scene_2_dim_up(Light *me, QEvent const *e);
static QState Light_scene_2_dimming_up(Light *me, QEvent const *e);
static QState Light_scene_3_dim_down(Light *me, QEvent const *e);
static QState Light_scene_3_dimming_down(Light *me, QEvent const *e);
static QState Light_scene_3_dim_up(Light *me, QEvent const *e);
static QState Light_scene_3_dimming_up(Light *me, QEvent const *e);
static QState Light_scene_4_dim_down(Light *me, QEvent const *e);
static QState Light_scene_4_dimming_down(Light *me, QEvent const *e);
static QState Light_scene_4_dim_up(Light *me, QEvent const *e);
static QState Light_scene_4_dimming_up(Light *me, QEvent const *e);

static QState Light_prog_mode(Light *me, QEvent const *e);
static QState Light_prog_mode_exit_blink_group(Light *me, QEvent const *e);
static QState Light_prog_get_device_count(Light *me, QEvent const *e);
static QState Light_prog_blink_group(Light *me, QEvent const *e);
static QState Light_prog_get_output_device(Light *me, QEvent const *e);
static QState Light_prog_device(Light *me, QEvent const *e);
static QState Light_prog_device_on(Light *me, QEvent const *e);
static QState Light_prog_device_on_dim_down(Light *me, QEvent const *e);
static QState Light_prog_device_on_dimming_down(Light *me, QEvent const *e);
static QState Light_prog_device_on_dim_up(Light *me, QEvent const *e);
static QState Light_prog_device_on_dimming_up(Light *me, QEvent const *e);
static QState Light_prog_device_off(Light *me, QEvent const *e);
static QState Light_prog_device_off_action(Light *me, QEvent const *e);
static QState Light_prog_no_device(Light *me, QEvent const *e);

static QState Light_remote_prog_mode(Light *me, QEvent const *e);
static QState Light_remote_prog_mode_blink_zone(Light *me, QEvent const *e);
static QState Light_prog_get_next_dsm_address(Light *me, QEvent const *e);
static QState Light_prog_get_remote_device_count(Light *me, QEvent const *e);
static QState Light_prog_get_next_remote_device(Light *me, QEvent const *e);

#define GetStateHandlerOffsetFromState(state_)  ((QStateHandler)state_)

static QStateHandler LightOffStates[6];

bool Light_is_off_state(const QStateHandler state) {
    uint8_t i;
    for (i = 0; i < (sizeof(LightOffStates) / sizeof(uint16_t)); ++i) {
        if (state == LightOffStates[i]) {
            return TRUE;
        }
    }
    return FALSE;
}

static void Light_RestartSTOPTimer(Light *me);
static void Light_PostGroupSignal(Light *me, QSignal signal, bool force);
static void Light_StopSTOPTimer(Light *me);
static bool Light_SendSTOP(Light *me);
static bool Light_SendSTOPProgrammingMode(Light *me);
static bool Light_SendSTOPForGroup(Light *me);
static QTimeEvtCtr Light_HandleDimmingAndGetTimeout(Light *me, Scene scene, OPCOperation operation);
static QTimeEvtCtr Light_ProgDeviceHandleDimmingAndGetTimeout(Light *me, Scene scene, OPCOperation operation);
static void Light_IdentifyNextRemoteDeviceIndex(Light *me);

/** Duration of a long downstream message */
#define LONG_DOWNSTREAM_MESSAGE_MS          (480)

/** Duration of a short downstream message */
#define SHORT_DOWNSTREAM_MESSAGE_MS         (240)

/** Duration of device blinking */
#define DEVICE_BLINK_DURATION_MS            (4000)

/* timeout for generating internally a INC/DEC signal */
#define LIGHT_DIM_TIMEOUT_TICKS             (250/BSP_MS_PER_TICKS)

/* timeout for first INC/DEC signal after calling min dim */
#define LIGHT_SHORT_MESSAGE_TIMEOUT_TICKS   (SHORT_DOWNSTREAM_MESSAGE_MS/BSP_MS_PER_TICKS)

/* timeout for generating internally a OPC_INC/OPC_DEC signal */
#define LIGHT_OPC_DIM_TIMEOUT_TICKS         (5000/BSP_MS_PER_TICKS)

/* timeout for generating internally a INC/DEC signal in programming mode */
#define LIGHT_PROG_DIM_TIMEOUT_TICKS        (250/BSP_MS_PER_TICKS)

/* timeout for generating internally a OPC_INC/OPC_DEC signal in programming mode */
#define LIGHT_PROG_OPC_DIM_TIMEOUT_TICKS    (5000/BSP_MS_PER_TICKS)

/* timeout for generating internally a TIP_1X signal in programming mode */
#define LIGHT_PROG_TIP_1X_TIMEOUT_TICKS     (1250/BSP_MS_PER_TICKS)

/* timeout for ending hold */
#define LIGHT_HOLD_END_TIMEOUT_TICKS        (2100/BSP_MS_PER_TICKS)

/* timeout for setting local priority */
#define LIGHT_SET_LOCAL_PRIO_TIMEOUT_TICKS  (2000/BSP_MS_PER_TICKS)

/* timeout for going from local priority off into standby */
#define LIGHT_STANDBY_TIMEOUT_TICKS         (800/BSP_MS_PER_TICKS)

/* timeout for going from standby into deep off */
#define LIGHT_DEEP_OFF_TIMEOUT_TICKS        (1000/BSP_MS_PER_TICKS)

/**
 * programming mode:
 * PROG ON command time (blinking duration)
 */
#define LIGHT_PROG_ON_TIMEOUT_TICKS         ((DEVICE_BLINK_DURATION_MS + 2*LONG_DOWNSTREAM_MESSAGE_MS - 2*LONG_DOWNSTREAM_MESSAGE_MS)/BSP_MS_PER_TICKS)

/**
 * programming mode:
 * PROG OFF command time (blinking duration)
 */
#define LIGHT_PROG_OFF_TIMEOUT_TICKS        ((DEVICE_BLINK_DURATION_MS + 2*LONG_DOWNSTREAM_MESSAGE_MS - SHORT_DOWNSTREAM_MESSAGE_MS)/BSP_MS_PER_TICKS)

/**
 * programming mode:
 * timeout in case the data model sends an error response which
 * can't be received by the light state machine
 */
#define LIGHT_PROG_DATAMODEL_TIMEOUT_TICKS  (3000/BSP_MS_PER_TICKS)

/**
 * programming mode:
 * re-arm exit timer
 */
#define Light_RearmExitTimer(me_) \
    QTimeEvt_rearm(&((me_)->progMode->exitTimer), PROG_EXIT_TIMEOUT_TICKS)


/*..........................................................................*/
static void Light_RestartSTOPTimer(Light *me)
{
    uint8_t stopTimeout = Datamodel_GetStopTimeoutZoneGroup(me->localZoneId, me->groupId);
    if (stopTimeout != 0) {
        if (me->stopTimer != NULL) {
            /* stop timer */
            QTimeEvt_disarm(&me->stopTimer->super);
        } else {
            me->stopTimer = SelectA_AllocateStopTimer(me->localZoneId, me->groupId, 0);
            if (me->stopTimer == NULL) {
                /* TODO: send error event if no more resources are available */
                return;
            }
        }
        QTimeEvt_postIn(&me->stopTimer->super, me->parent, stopTimeout * 1000 / BSP_MS_PER_TICKS);
    }
}

/*..........................................................................*/
static void Light_StopSTOPTimer(Light *me)
{
    if (me->stopTimer != NULL) {
        SelectA_ReleaseStopTimer(me->stopTimer);
        me->stopTimer = NULL;
    }
    Light_PostGroupSignal(me, CALL_ROOM_STOP_SIG, FALSE);
}

/*..........................................................................*/
static void Light_PostGroupSignal(Light *me, QSignal signal, bool force)
{
    /* no need to post group signals to non standard groups */
    if (me->groupId > NUM_GROUPS_WITH_AREA) {
        return;
    }
    /* Set the localZoneId to 0 to handle the signal in the group state machines.
     * Otherwise signal could be falsely routed in programming mode. */
    uint8_t zone = me->progMode->localZoneId;
    me->progMode->localZoneId = 0;
    SelectA_PostGroupSignal(me->groupId, me->localZoneId, signal, force);
    me->progMode->localZoneId = zone;
}

/*..........................................................................*/
static bool Light_SendSTOP(Light *me)
{
    bool sendStop = FALSE;

    if (me->stopTimer != NULL){
        /* release timer */
        SelectA_ReleaseStopTimer(me->stopTimer);
        me->stopTimer = NULL;
        sendStop = TRUE;
    } else if (SelectA_ButtonGroupTimerRunning(me->localZoneId, me->groupId)) {
        sendStop = TRUE;
    }
    if (sendStop) {
        Scene scene = Scene_GetStopScene(me->buttonId);
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
        Light_PostGroupSignal(me, CALL_ROOM_STOP_SIG, FALSE);
    }
    return sendStop;
}

/*..........................................................................*/
static bool Light_SendSTOPProgrammingMode(Light *me)
{
    if (me->stopTimer != NULL) {
        SelectA_ReleaseStopTimer(me->stopTimer);
        me->stopTimer = NULL;
        if (me->progMode->dSMAddress == DS485_INVALID_ADDRESS) {
            Scene scene = Scene_GetStopScene(me->progMode->buttonId);
            SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, FALSE, me->deviceId);
        } else {
            SelectA_SendGeneralAck();
            DsmApi_SendDeviceActionCallSceneRequest(me->progMode->dSMAddress, me->progMode->deviceId, STOP_S);
        }
        return TRUE;
    }
    return FALSE;
}

/*..........................................................................*/
static bool Light_SendSTOPForGroup(Light *me)
{
    if (SelectA_ButtonGroupTimerRunning(me->localZoneId, me->groupId)) {
        Scene scene = Scene_GetStopScene(me->buttonId);
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
        Light_PostGroupSignal(me, CALL_ROOM_STOP_SIG, FALSE);
        return TRUE;
    }
    return FALSE;
}

/*..........................................................................*/
void Light_ctor(Light *me)
{
    QHsm_ctor(&me->super, (QStateHandler)&Light_initial);

    LightOffStates[0] = GetStateHandlerOffsetFromState(&Light_off);
    LightOffStates[1] = GetStateHandlerOffsetFromState(&Light_scene_0);
    LightOffStates[2] = GetStateHandlerOffsetFromState(&Light_scene_0_local);
    LightOffStates[3] = GetStateHandlerOffsetFromState(&Light_local_prio_off_hold);
    LightOffStates[4] = GetStateHandlerOffsetFromState(&Light_standby_hold);
    LightOffStates[5] = GetStateHandlerOffsetFromState(&Light_deep_off_hold);
}

/* HSM definition */

/*..........................................................................*/
static QState Light_initial(Light *me, QEvent const *e)
{
    /* QS dictionary entries */
    QS_FUN_DICTIONARY(&Light_top);

    QS_FUN_DICTIONARY(&Light_off);
    QS_FUN_DICTIONARY(&Light_scene_0);
    QS_FUN_DICTIONARY(&Light_scene_0_local);
    QS_FUN_DICTIONARY(&Light_scene_0_hold);
    QS_FUN_DICTIONARY(&Light_local_prio_off_hold);
    QS_FUN_DICTIONARY(&Light_standby);
    QS_FUN_DICTIONARY(&Light_standby_local);
    QS_FUN_DICTIONARY(&Light_deep_off);
    QS_FUN_DICTIONARY(&Light_deep_off_local);

    QS_FUN_DICTIONARY(&Light_on);
    QS_FUN_DICTIONARY(&Light_scene_1);
    QS_FUN_DICTIONARY(&Light_scene_1_local);
    QS_FUN_DICTIONARY(&Light_scene_2);
    QS_FUN_DICTIONARY(&Light_scene_2_local);
    QS_FUN_DICTIONARY(&Light_scene_3);
    QS_FUN_DICTIONARY(&Light_scene_3_local);
    QS_FUN_DICTIONARY(&Light_scene_4);
    QS_FUN_DICTIONARY(&Light_scene_4_local);

    QS_FUN_DICTIONARY(&Light_dimming_down);
    QS_FUN_DICTIONARY(&Light_dimming_up);
    QS_FUN_DICTIONARY(&Light_scene_1_dim_down);
    QS_FUN_DICTIONARY(&Light_scene_1_dimming_down);
    QS_FUN_DICTIONARY(&Light_scene_1_dim_up);
    QS_FUN_DICTIONARY(&Light_scene_1_dimming_up);
    QS_FUN_DICTIONARY(&Light_scene_2_dim_down);
    QS_FUN_DICTIONARY(&Light_scene_2_dimming_down);
    QS_FUN_DICTIONARY(&Light_scene_2_dim_up);
    QS_FUN_DICTIONARY(&Light_scene_2_dimming_up);
    QS_FUN_DICTIONARY(&Light_scene_3_dim_down);
    QS_FUN_DICTIONARY(&Light_scene_3_dimming_down);
    QS_FUN_DICTIONARY(&Light_scene_3_dim_up);
    QS_FUN_DICTIONARY(&Light_scene_3_dimming_up);
    QS_FUN_DICTIONARY(&Light_scene_4_dim_down);
    QS_FUN_DICTIONARY(&Light_scene_4_dimming_down);
    QS_FUN_DICTIONARY(&Light_scene_4_dim_up);
    QS_FUN_DICTIONARY(&Light_scene_4_dimming_up);

    QS_FUN_DICTIONARY(&Light_prog_mode);
    QS_FUN_DICTIONARY(&Light_prog_get_device_count);
    QS_FUN_DICTIONARY(&Light_prog_blink_group);
    QS_FUN_DICTIONARY(&Light_prog_get_output_device);
    QS_FUN_DICTIONARY(&Light_prog_device);
    QS_FUN_DICTIONARY(&Light_prog_device_on);
    QS_FUN_DICTIONARY(&Light_prog_device_on_dim_down);
    QS_FUN_DICTIONARY(&Light_prog_device_on_dimming_down);
    QS_FUN_DICTIONARY(&Light_prog_device_on_dim_up);
    QS_FUN_DICTIONARY(&Light_prog_device_on_dimming_up);
    QS_FUN_DICTIONARY(&Light_prog_device_off);
    QS_FUN_DICTIONARY(&Light_prog_device_off_action);
    QS_FUN_DICTIONARY(&Light_prog_no_device);

    QS_FUN_DICTIONARY(&Light_remote_prog_mode);

    return Q_TRAN(&Light_scene_0);
}

/*..........................................................................*/
static QState Light_top(Light *me, QEvent const *e)
{
    /* implements default Light_off behavior */
    switch (e->sig) {
    case Q_INIT_SIG:
        return Q_TRAN(&Light_off);
    case CT_SHORT_LONG_SIG:
        /* fall through */
    case CT_LOCAL_SHORT_LONG_SIG: {
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        /* is other zone in programming mode? */
        if (me->progMode->localZoneId != 0) {
            /* Initiate device pairing, i.e. move device to programming zone */
            SelectA_InitiateDevicePairing(me->deviceId, me->progMode->localZoneId);
        }
        return Q_HANDLED();
    }
    case CT_2WAY_UP_TIP_1X_SIG: {
        if (Light_SendSTOP(me)) {
            return Q_HANDLED();
        }
        return Q_TRAN(&Light_scene_1_local);
    }
    case CT_2WAY_DOWN_TIP_1X_SIG: {
        if (Light_SendSTOP(me)) {
            return Q_HANDLED();
        }
        return Q_TRAN(&Light_scene_0_local);
    }
    case CT_2WAY_UP_TIP_2X_SIG:
        /* fall through */
    case CT_CLICK_2X_SIG:
        /* fall through */
    case CT_TIP_2X_SIG: {
        return Q_TRAN(&Light_scene_2_local);
    }
    case CT_2WAY_UP_TIP_3X_SIG:
        /* fall through */
    case CT_CLICK_3X_SIG:
        /* fall through */
    case CT_TIP_3X_SIG: {
        return Q_TRAN(&Light_scene_3_local);
    }
    case CT_2WAY_UP_TIP_4X_SIG:
        /* fall through */
    case CT_TIP_4X_SIG: {
        return Q_TRAN(&Light_scene_4_local);
    }
    case BUTTON_GROUP_FORCE_SCENE_0_SIG:
    case CT_2WAY_DOWN_TIP_2X_SIG: {
        /* force local priority off */
        SelectA_ForceCallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, 0, me->apartmentWide, me->deviceId);
        Light_PostGroupSignal(me, CALL_ROOM_OFF_SIG, TRUE);
        return Q_TRAN(&Light_scene_0);
    }
    case BUTTON_GROUP_FORCE_SCENE_0_AND_STANDBY_SIG:
    case BUTTON_GROUP_STANDBY_SIG:
    case CT_2WAY_DOWN_CLICK_3X_SIG:
    case CT_2WAY_DOWN_TIP_3X_SIG: {
        if ((me->stateMachineGroupId != LIGHT_GROUP) || (me->groupId >= USER_GROUP)){
            SelectA_SendGeneralAck();
            return Q_HANDLED();
        }
        if (e->sig == CT_2WAY_DOWN_CLICK_3X_SIG || e->sig == BUTTON_GROUP_FORCE_SCENE_0_AND_STANDBY_SIG) {
            /* force local priority off */
            SelectA_ForceCallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, 0, me->apartmentWide, me->deviceId);
        }
        /* call STANDBY scene */
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, 0, me->localZoneId, me->groupId, STANDBY, me->apartmentWide, me->deviceId);
        Light_PostGroupSignal(me, CALL_ROOM_OFF_SIG, TRUE);
        return Q_TRAN(&Light_scene_0);
    }
    case BUTTON_GROUP_DEEP_OFF_SIG:
    case CT_2WAY_DOWN_TIP_4X_SIG: {
        if ((me->stateMachineGroupId != LIGHT_GROUP) || (me->groupId >= USER_GROUP)) {
            SelectA_SendGeneralAck();
            return Q_HANDLED();
        }
        /* call DEEP OFF scene */
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, 0, me->localZoneId, me->groupId, DEEP_OFF, me->apartmentWide, me->deviceId);
        Light_PostGroupSignal(me, CALL_ROOM_OFF_SIG, TRUE);
        return Q_TRAN(&Light_scene_0);
    }
    case CT_2WAY_UP_HOLD_END_SIG:
        /* fall through */
    case CT_2WAY_DOWN_HOLD_END_SIG: {
        /* send general acknowledge if hold end signal of 2-way button is not handled */
        SelectA_SendGeneralAck();
        return Q_HANDLED();
    }
    case CT_LOCAL_2WAY_DOWN_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_LOCAL_HOLD_START_SIG: {
        return Q_TRAN(&Light_local_prio_off_hold);
    }
    case CALL_SCENE_ZONE_GROUP_SIG: {
        SceneZoneGroupEvt * const sceneZoneGroupEvt = (SceneZoneGroupEvt * const)e;
        Scene scene = (Scene)sceneZoneGroupEvt->basic.sceneId;
        bool force = sceneZoneGroupEvt->basic.force;
        /* unconditionally call scene locally */
        SelectA_SendSceneZoneGroup(me->localZoneId, me->groupId, scene, force);
        /* fall through */
    }
    case CALL_SCENE_BROADCAST_SIG: {
        SceneBasicEvt const * sceneBasicEvt = (SceneBasicEvt const *)e;
        if ((sceneBasicEvt->sourceZoneId != me->localZoneId) || (sceneBasicEvt->sourceGroupId != me->groupId)) {
            bool force = sceneBasicEvt->force;
            uint8_t sceneId = sceneBasicEvt->sceneId;
            if (Scene_IsStopScene(sceneId)) {
                Light_StopSTOPTimer(me);
                return Q_HANDLED();
            }
            if (Scene_IsOffScene(sceneId)) {
                Light_PostGroupSignal(me, CALL_ROOM_OFF_SIG, force);
                return Q_TRAN(&Light_scene_0);
            }
            if (Scene_IsOnScene(sceneId)) {
                Light_PostGroupSignal(me, CALL_ROOM_ON_SIG, force);
                return Q_TRAN(&Light_scene_1);
            }
            if (Scene_IsScene2(sceneId)) {
                Light_PostGroupSignal(me, CALL_ROOM_ON_SIG, force);
                return Q_TRAN(&Light_scene_2);
            }
            if (Scene_IsScene3(sceneId)) {
                Light_PostGroupSignal(me, CALL_ROOM_ON_SIG, force);
                return Q_TRAN(&Light_scene_3);
            }
            if (Scene_IsScene4(sceneId)) {
                Light_PostGroupSignal(me, CALL_ROOM_ON_SIG, force);
                return Q_TRAN(&Light_scene_4);
            }
            if (((sceneId == AUTO_STANDBY) || (sceneId == STANDBY) || (sceneId == DEEP_OFF) || (sceneId == SLEEPING) || (sceneId == ABSENT)) &&
                (me->stateMachineGroupId != SHADE_GROUP)) {
                Light_PostGroupSignal(me, CALL_ROOM_OFF_SIG, TRUE);
                return Q_TRAN(&Light_scene_0);
            }
        }
        /* don't change state */
        return Q_HANDLED();
    }
    case UNDO_SCENE_NUMBER_ZONE_GROUP_SIG: {
        SceneZoneGroupEvt * const sceneZoneGroupEvt = (SceneZoneGroupEvt * const)e;
        Scene scene = (Scene)sceneZoneGroupEvt->basic.sceneId;
        /* unconditionally undo scene number locally */
        SelectA_SendUndoSceneNumberZoneGroup(me->localZoneId, me->groupId, scene);
        /* fall through */
    }
    case UNDO_SCENE_NUMBER_BROADCAST_SIG: {
        Light_RestartSTOPTimer(me);
        return Q_HANDLED();
    }
    case BUTTON_GROUP_PROGRAMM_MODE_SIG: {
        if (me->progMode->localZoneId == 0) {
            return Q_TRAN(&Light_prog_mode);
        }
        return Q_HANDLED();
    }
    case BUTTON_GROUP_LOCAL_PRIO_OFF_SIG: {
        return Q_TRAN(&Light_local_prio_off_hold);
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState Light_special_room_state(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_CLICK_1X_SIG:
        /* fall through */
    case CT_TIP_1X_SIG: {
        return Q_TRAN(&Light_scene_0_local);
    }
    }
    return Q_SUPER(&Light_top);
}

/*..........................................................................*/
static QState Light_special_room_state_return_scene_0(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case UNDO_SCENE_NUMBER_ZONE_GROUP_SIG: {
        SceneZoneGroupEvt * const sceneZoneGroupEvt = (SceneZoneGroupEvt * const)e;
        Scene scene = (Scene)sceneZoneGroupEvt->basic.sceneId;
        /* unconditionally undo scene number locally */
        SelectA_SendUndoSceneNumberZoneGroup(me->localZoneId, me->groupId, scene);
        /* fall through */
    }
    case UNDO_SCENE_NUMBER_BROADCAST_SIG: {
        SceneZoneGroupEvt * const sceneZoneGroupEvt = (SceneZoneGroupEvt * const)e;
        Scene scene = (Scene)sceneZoneGroupEvt->basic.sceneId;
        if (SelectA_LightAlarmScene() == scene) {
            return Q_TRAN(&Light_scene_0);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_special_room_state);
}

/*..........................................................................*/
static QState Light_special_room_state_return_scene_1(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case UNDO_SCENE_NUMBER_ZONE_GROUP_SIG: {
        SceneZoneGroupEvt * const sceneZoneGroupEvt = (SceneZoneGroupEvt * const)e;
        Scene scene = (Scene)sceneZoneGroupEvt->basic.sceneId;
        /* unconditionally undo scene number locally */
        SelectA_SendUndoSceneNumberZoneGroup(me->localZoneId, me->groupId, scene);
        /* fall through */
    }
    case UNDO_SCENE_NUMBER_BROADCAST_SIG: {
        SceneZoneGroupEvt * const sceneZoneGroupEvt = (SceneZoneGroupEvt * const)e;
        Scene scene = (Scene)sceneZoneGroupEvt->basic.sceneId;
        if (SelectA_LightAlarmScene() == scene) {
            return Q_TRAN(&Light_scene_1);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_special_room_state);
}

/*..........................................................................*/
static QState Light_special_room_state_return_scene_2(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case UNDO_SCENE_NUMBER_ZONE_GROUP_SIG: {
        SceneZoneGroupEvt * const sceneZoneGroupEvt = (SceneZoneGroupEvt * const)e;
        Scene scene = (Scene)sceneZoneGroupEvt->basic.sceneId;
        /* unconditionally undo scene number locally */
        SelectA_SendUndoSceneNumberZoneGroup(me->localZoneId, me->groupId, scene);
        /* fall through */
    }
    case UNDO_SCENE_NUMBER_BROADCAST_SIG: {
        SceneZoneGroupEvt * const sceneZoneGroupEvt = (SceneZoneGroupEvt * const)e;
        Scene scene = (Scene)sceneZoneGroupEvt->basic.sceneId;
        if (SelectA_LightAlarmScene() == scene) {
            return Q_TRAN(&Light_scene_2);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_special_room_state);
}

/*..........................................................................*/
static QState Light_special_room_state_return_scene_3(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case UNDO_SCENE_NUMBER_ZONE_GROUP_SIG: {
        SceneZoneGroupEvt * const sceneZoneGroupEvt = (SceneZoneGroupEvt * const)e;
        Scene scene = (Scene)sceneZoneGroupEvt->basic.sceneId;
        /* unconditionally undo scene number locally */
        SelectA_SendUndoSceneNumberZoneGroup(me->localZoneId, me->groupId, scene);
        /* fall through */
    }
    case UNDO_SCENE_NUMBER_BROADCAST_SIG: {
        SceneZoneGroupEvt * const sceneZoneGroupEvt = (SceneZoneGroupEvt * const)e;
        Scene scene = (Scene)sceneZoneGroupEvt->basic.sceneId;
        if (SelectA_LightAlarmScene() == scene) {
            return Q_TRAN(&Light_scene_3);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_special_room_state);
}

/*..........................................................................*/
static QState Light_special_room_state_return_scene_4(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case UNDO_SCENE_NUMBER_ZONE_GROUP_SIG: {
        SceneZoneGroupEvt * const sceneZoneGroupEvt = (SceneZoneGroupEvt * const)e;
        Scene scene = (Scene)sceneZoneGroupEvt->basic.sceneId;
        /* unconditionally undo scene number locally */
        SelectA_SendUndoSceneNumberZoneGroup(me->localZoneId, me->groupId, scene);
        /* fall through */
    }
    case UNDO_SCENE_NUMBER_BROADCAST_SIG: {
        SceneZoneGroupEvt * const sceneZoneGroupEvt = (SceneZoneGroupEvt * const)e;
        Scene scene = (Scene)sceneZoneGroupEvt->basic.sceneId;
        if (SelectA_LightAlarmScene() == scene) {
            return Q_TRAN(&Light_scene_4);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_special_room_state);
}

/*..........................................................................*/
static QState Light_off(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_TIP_1X_SIG:
        /* fall through */
    case CT_CLICK_1X_SIG: {
        /* scene 1 */
        return Q_TRAN(&Light_scene_1_local);
    }
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_LOCAL_HOLD_END_SIG:
        if (Datamodel_DsmPropertiesRoomDimmingDisabled()) {
            break;
        }
        /* fall through */
    case CT_HOLD_END_SIG: {
        /* send general acknowledge if short-long, hold start and end signal is not handled */
        SelectA_SendGeneralAck();
        return Q_HANDLED();
    }
    case CT_2WAY_UP_HOLD_START_SIG:
        /* fall through */
    case CT_2WAY_UP_HOLD_REPEAT_SIG: {
        return Q_TRAN(&Light_scene_0_dimming_up);
    }
    case CT_2WAY_DOWN_HOLD_START_SIG:
        /* fall through */
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG: {
        if (me->stateMachineGroupId == SHADE_GROUP){
            return Q_TRAN(&Light_scene_0_dimming_down);
        } else {
            return Q_TRAN(&Light_local_prio_off_hold);
        }
    }
    case CALL_SCENE_BROADCAST_SIG: {
        /* handle only broadcast call scene */
        SceneBasicEvt const * sceneBasicEvt = (SceneBasicEvt const *)e;
        if (SelectA_LightAlarmScene() == sceneBasicEvt->sceneId) {
            return Q_TRAN(&Light_special_room_state_return_scene_0);
        }
        /* don't return Q_HANDLED()! */
        break;
    }
    case PROG_ON_ZONE_SIG: {
        /* this case should only be necessary when a new zone was created during device pairing */
        if (me->progMode->localZoneId == 0) {
            /* save initiating device ID, button ID and scene */
            me->progMode->initiatingDeviceId = 0;
            me->progMode->buttonId = me->buttonId;
            me->progMode->sceneIndex = 0;
            return Q_TRAN(&Light_remote_prog_mode);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_top);
}

/*..........................................................................*/
static QState Light_scene_0(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Light_RestartSTOPTimer(me);
        return Q_HANDLED();
    }
    case CT_TIP_1X_SIG:
        /* fall through */
    case CT_CLICK_1X_SIG: {
        if (Light_SendSTOP(me)) {
            return Q_HANDLED();
        }
        break;
    }
    case CT_HOLD_START_SIG: {
        /* force local priority off */
        return Q_TRAN(&Light_local_prio_off_hold);
    }
    }
    return Q_SUPER(&Light_off);
}

/*..........................................................................*/
static QState Light_scene_0_local(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* handle registered (pending) local priority of groups */
        SelectA_HandleRegisteredLocalPrio(me->apartmentWide ? 0 : me->localZoneId, me->groupId, me->deviceId);
        Scene scene = Scene_GetOffScene(me->buttonId);
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
        Light_PostGroupSignal(me, CALL_ROOM_OFF_SIG, FALSE);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_scene_0);
}

/*..........................................................................*/
static QState Light_local_prio_off_hold(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* force local priority off */
        Scene scene = 0;
        SelectA_ForceCallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
        /* arm timers */
        if ((me->stateMachineGroupId == LIGHT_GROUP)  && (me->groupId < USER_GROUP)) {
            /* arm standby timer only for light group */
            Op_QTimeEvt_postIn(me->opTimer1, me->parent, LIGHT_STANDBY_TIMEOUT_TICKS, STANDBY_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        }
        Op_QTimeEvt_postIn(me->opTimer2, me->parent, LIGHT_HOLD_END_TIMEOUT_TICKS, END_STANDBY_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        Light_RestartSTOPTimer(me);
        Light_PostGroupSignal(me, CALL_ROOM_OFF_SIG, TRUE);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_disarm(me->opTimer2);
        return Q_HANDLED();
    }
    case CT_LOCAL_2WAY_DOWN_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_LOCAL_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        /* re-arm timer */
        Op_QTimeEvt_rearm(me->opTimer2, LIGHT_HOLD_END_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case CT_LOCAL_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_LOCAL_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG: {
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        /* fall through */
    }
    case END_STANDBY_SIG: {
        return Q_TRAN(&Light_scene_0);
    }
    case STANDBY_SIG: {
        /* call standby scene on all devices in the zone */
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, 0, me->localZoneId, me->groupId, STANDBY, me->apartmentWide, me->deviceId);
        return Q_TRAN(&Light_standby_hold);
    }
    }
    return Q_SUPER(&Light_off);
}

/*..........................................................................*/
static QState Light_standby_hold(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* arm timers */
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, LIGHT_DEEP_OFF_TIMEOUT_TICKS, DEEP_OFF_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        Op_QTimeEvt_postIn(me->opTimer2, me->parent, LIGHT_HOLD_END_TIMEOUT_TICKS, END_DEEP_OFF_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        Light_RestartSTOPTimer(me);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_disarm(me->opTimer2);
        return Q_HANDLED();
    }
    case CT_LOCAL_2WAY_DOWN_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_LOCAL_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        /* re-arm timer */
        Op_QTimeEvt_rearm(me->opTimer2, LIGHT_HOLD_END_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case CT_LOCAL_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_LOCAL_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG: {
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        /* fall through */
    }
    case END_DEEP_OFF_SIG: {
        return Q_TRAN(&Light_scene_0);
    }
    case DEEP_OFF_SIG: {
        /* call deep off scene on all devices in the zone */
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, 0, me->localZoneId, me->groupId, DEEP_OFF, me->apartmentWide, me->deviceId);
        return Q_TRAN(&Light_deep_off_hold);
    }
    }
    return Q_SUPER(&Light_off);
}

/*..........................................................................*/
static QState Light_deep_off_hold(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Light_RestartSTOPTimer(me);
        return Q_HANDLED();
    }
    case CT_HOLD_START_SIG: {
        /* force local priority off */
        return Q_TRAN(&Light_local_prio_off_hold);
    }
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        return Q_HANDLED();
    }
    case CT_LOCAL_2WAY_DOWN_HOLD_START_SIG:
        /* fall through */
    case CT_LOCAL_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_LOCAL_HOLD_START_SIG:
        /* fall through */
    case CT_LOCAL_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG: {
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        return Q_TRAN(Light_scene_0);
    }
    }
    return Q_SUPER(&Light_off);
}

/*..........................................................................*/
static QState Light_on(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_TIP_1X_SIG:
        /* fall through */
    case CT_CLICK_1X_SIG: {
        /* off scene */
        return Q_TRAN(&Light_scene_0_local);
    }
    case CT_LOCAL_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_LOCAL_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG: {
        /* send general acknowledge if hold end signal is not handled */
        SelectA_SendGeneralAck();
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_top);
}

/*..........................................................................*/
static QState Light_scene_1(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Light_RestartSTOPTimer(me);
        return Q_HANDLED();
    }
    case CT_TIP_1X_SIG:
        /* fall through */
    case CT_CLICK_1X_SIG: {
        if (Light_SendSTOP(me)) {
            return Q_HANDLED();
        }
        break;
    }
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG:
        if (Datamodel_DsmPropertiesRoomDimmingDisabled()) {
            break;
        }
        /* fall through */
    case CT_2WAY_DOWN_HOLD_START_SIG:
        /* fall through */
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG: {
        Light_SendSTOPForGroup(me);
        return Q_TRAN(&Light_scene_1_dimming_down);
    }
    case CT_2WAY_UP_HOLD_START_SIG:
        /* fall through */
    case CT_2WAY_UP_HOLD_REPEAT_SIG: {
        Light_SendSTOPForGroup(me);
        return Q_TRAN(&Light_scene_1_dimming_up);
    }
    case CT_SHORT_LONG_SIG: {
        if (me->progMode->localZoneId == 0) {
            /* save initiating device ID, button ID and scene */
            me->progMode->initiatingDeviceId = me->deviceId;
            me->progMode->buttonId = me->buttonId;
            me->progMode->sceneIndex = 1;
            return Q_TRAN(&Light_prog_mode);
        }
        /* handle short long click for device pairing in super state */
        break;
    }
    case PROG_ON_ZONE_SIG: {
        if (me->progMode->localZoneId == 0) {
            /* save initiating device ID, button ID and scene */
            me->progMode->initiatingDeviceId = 0;
            me->progMode->buttonId = me->buttonId;
            me->progMode->sceneIndex = 1;
            return Q_TRAN(&Light_remote_prog_mode_blink_zone);
        }
        return Q_HANDLED();
    }
    case CALL_SCENE_BROADCAST_SIG: {
        /* handle only broadcast call scene */
        SceneBasicEvt const * sceneBasicEvt = (SceneBasicEvt const *)e;
        if (SelectA_LightAlarmScene() == sceneBasicEvt->sceneId) {
            return Q_TRAN(&Light_special_room_state_return_scene_1);
        }
        /* don't return Q_HANDLED()! */
        break;
    }
    }
    return Q_SUPER(&Light_on);
}

/*..........................................................................*/
static QState Light_scene_1_local(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Scene scene;
        if (SelectA_ButtonGroupTimerRunning(me->localZoneId, me->groupId)) {
            scene = Scene_GetStopScene(me->buttonId);
            SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
            Light_PostGroupSignal(me, CALL_ROOM_STOP_SIG, FALSE);
        }
        /* handle registered (pending) local priority of groups */
        SelectA_HandleRegisteredLocalPrio(me->apartmentWide ? 0 : me->localZoneId, me->groupId, me->deviceId);
        /* call scene */
        scene = Scene_GetOnScene(me->buttonId);
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
        Light_PostGroupSignal(me, CALL_ROOM_ON_SIG, FALSE);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_scene_1);
}

/*..........................................................................*/
static QState Light_scene_2(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Light_RestartSTOPTimer(me);
        return Q_HANDLED();
    }
    case CT_TIP_1X_SIG:
        /* fall through */
    case CT_CLICK_1X_SIG: {
        if (Light_SendSTOP(me)) {
            return Q_HANDLED();
        }
        break;
    }
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG:
        if (Datamodel_DsmPropertiesRoomDimmingDisabled()) {
            break;
        }
        /* fall through */
    case CT_2WAY_DOWN_HOLD_START_SIG:
        /* fall through */
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG: {
        Light_SendSTOPForGroup(me);
        return Q_TRAN(&Light_scene_2_dimming_down);
    }
    case CT_2WAY_UP_HOLD_START_SIG:
        /* fall through */
    case CT_2WAY_UP_HOLD_REPEAT_SIG: {
        Light_SendSTOPForGroup(me);
        return Q_TRAN(&Light_scene_2_dimming_up);
    }
    case CT_SHORT_LONG_SIG: {
        if (me->progMode->localZoneId == 0) {
            /* save initiating device ID, button ID and scene */
            me->progMode->initiatingDeviceId = me->deviceId;
            me->progMode->buttonId = me->buttonId;
            me->progMode->sceneIndex = 2;
            return Q_TRAN(&Light_prog_mode);
        }
        /* handle short long click for device pairing in super state */
        break;
    }
    case PROG_ON_ZONE_SIG: {
        if (me->progMode->localZoneId == 0) {
            /* save initiating device ID, button ID and scene */
            me->progMode->initiatingDeviceId = 0;
            me->progMode->buttonId = me->buttonId;
            me->progMode->sceneIndex = 2;
            return Q_TRAN(&Light_remote_prog_mode_blink_zone);
        }
        return Q_HANDLED();
    }
    case CALL_SCENE_BROADCAST_SIG: {
        /* handle only broadcast call scene */
        SceneBasicEvt const * sceneBasicEvt = (SceneBasicEvt const *)e;
        if (SelectA_LightAlarmScene() == sceneBasicEvt->sceneId) {
            return Q_TRAN(&Light_special_room_state_return_scene_2);
        }
        /* don't return Q_HANDLED()! */
        break;
    }
    }
    return Q_SUPER(&Light_on);
}

/*..........................................................................*/
static QState Light_scene_2_local(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Scene scene;
        if (SelectA_ButtonGroupTimerRunning(me->localZoneId, me->groupId)) {
            scene = Scene_GetStopScene(me->buttonId);
            SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
            Light_PostGroupSignal(me, CALL_ROOM_STOP_SIG, FALSE);
        }
        /* handle registered (pending) local priority of groups */
        SelectA_HandleRegisteredLocalPrio(me->apartmentWide ? 0 : me->localZoneId, me->groupId, me->deviceId);
        /* call scene */
        scene = Scene_GetScene2(me->buttonId);
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
        Light_PostGroupSignal(me, CALL_ROOM_ON_SIG, FALSE);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_scene_2);
}

/*..........................................................................*/
static QState Light_scene_3(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Light_RestartSTOPTimer(me);
        return Q_HANDLED();
    }
    case CT_TIP_1X_SIG:
        /* fall through */
    case CT_CLICK_1X_SIG: {
        if (Light_SendSTOP(me)) {
            return Q_HANDLED();
        }
        break;
    }
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG:
        if (Datamodel_DsmPropertiesRoomDimmingDisabled()) {
            break;
        }
        /* fall through */
    case CT_2WAY_DOWN_HOLD_START_SIG:
        /* fall through */
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG: {
        Light_SendSTOPForGroup(me);
        return Q_TRAN(&Light_scene_3_dimming_down);
    }
    case CT_2WAY_UP_HOLD_START_SIG:
        /* fall through */
    case CT_2WAY_UP_HOLD_REPEAT_SIG: {
        Light_SendSTOPForGroup(me);
        return Q_TRAN(&Light_scene_3_dimming_up);
    }
    case CT_SHORT_LONG_SIG: {
        if (me->progMode->localZoneId == 0) {
            /* save initiating device ID, button ID and scene */
            me->progMode->initiatingDeviceId = me->deviceId;
            me->progMode->buttonId = me->buttonId;
            me->progMode->sceneIndex = 3;
            return Q_TRAN(&Light_prog_mode);
        }
        /* handle short long click for device pairing in super state */
        break;
    }
    case PROG_ON_ZONE_SIG: {
        if (me->progMode->localZoneId == 0) {
            /* save initiating device ID, button ID and scene */
            me->progMode->initiatingDeviceId = 0;
            me->progMode->buttonId = me->buttonId;
            me->progMode->sceneIndex = 3;
            return Q_TRAN(&Light_remote_prog_mode_blink_zone);
        }
        return Q_HANDLED();
    }
    case CALL_SCENE_BROADCAST_SIG: {
        /* handle only broadcast call scene */
        SceneBasicEvt const * sceneBasicEvt = (SceneBasicEvt const *)e;
        if (SelectA_LightAlarmScene() == sceneBasicEvt->sceneId) {
            return Q_TRAN(&Light_special_room_state_return_scene_3);
        }
        /* don't return Q_HANDLED()! */
        break;
    }
    }
    return Q_SUPER(&Light_on);
}

/*..........................................................................*/
static QState Light_scene_3_local(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Scene scene;
        if (SelectA_ButtonGroupTimerRunning(me->localZoneId, me->groupId)) {
            scene = Scene_GetStopScene(me->buttonId);
            SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
            Light_PostGroupSignal(me, CALL_ROOM_STOP_SIG, FALSE);
        }
        /* handle registered (pending) local priority of groups */
        SelectA_HandleRegisteredLocalPrio(me->apartmentWide ? 0 : me->localZoneId, me->groupId, me->deviceId);
        /* call scene */
        scene = Scene_GetScene3(me->buttonId);
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
        Light_PostGroupSignal(me, CALL_ROOM_ON_SIG, FALSE);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_scene_3);
}

/*..........................................................................*/
static QState Light_scene_4(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Light_RestartSTOPTimer(me);
        return Q_HANDLED();
    }
    case CT_TIP_1X_SIG:
        /* fall through */
    case CT_CLICK_1X_SIG: {
        if (Light_SendSTOP(me)) {
            return Q_HANDLED();
        }
        break;
    }
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG:
        if (Datamodel_DsmPropertiesRoomDimmingDisabled()) {
            break;
        }
        /* fall through */
    case CT_2WAY_DOWN_HOLD_START_SIG:
        /* fall through */
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG: {
        Light_SendSTOPForGroup(me);
        return Q_TRAN(&Light_scene_4_dimming_down);
    }
    case CT_2WAY_UP_HOLD_START_SIG:
        /* fall through */
    case CT_2WAY_UP_HOLD_REPEAT_SIG: {
        Light_SendSTOPForGroup(me);
        return Q_TRAN(&Light_scene_4_dimming_up);
    }
    case CT_SHORT_LONG_SIG: {
        if (me->progMode->localZoneId == 0) {
            /* save initiating device ID, button ID and scene */
            me->progMode->initiatingDeviceId = me->deviceId;
            me->progMode->buttonId = me->buttonId;
            me->progMode->sceneIndex = 4;
            return Q_TRAN(&Light_prog_mode);
        }
        /* handle short long click for device pairing in super state */
        break;
    }
    case PROG_ON_ZONE_SIG: {
        if (me->progMode->localZoneId == 0) {
            /* save initiating device ID, button ID and scene */
            me->progMode->initiatingDeviceId = 0;
            me->progMode->buttonId = me->buttonId;
            me->progMode->sceneIndex = 4;
            return Q_TRAN(&Light_remote_prog_mode_blink_zone);
        }
        return Q_HANDLED();
    }
    case CALL_SCENE_BROADCAST_SIG: {
        /* handle only broadcast call scene */
        SceneBasicEvt const * sceneBasicEvt = (SceneBasicEvt const *)e;
        if (SelectA_LightAlarmScene() == sceneBasicEvt->sceneId) {
            return Q_TRAN(&Light_special_room_state_return_scene_4);
        }
        /* don't return Q_HANDLED()! */
        break;
    }
    }
    return Q_SUPER(&Light_on);
}

/*..........................................................................*/
static QState Light_scene_4_local(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Scene scene;
        if (SelectA_ButtonGroupTimerRunning(me->localZoneId, me->groupId)) {
            scene = Scene_GetStopScene(me->buttonId);
            SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
            Light_PostGroupSignal(me, CALL_ROOM_STOP_SIG, FALSE);
        }
        /* handle registered (pending) local priority of groups */
        SelectA_HandleRegisteredLocalPrio(me->apartmentWide ? 0 : me->localZoneId, me->groupId, me->deviceId);
        /* call scene */
        scene = Scene_GetScene4(me->buttonId);
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
        Light_PostGroupSignal(me, CALL_ROOM_ON_SIG, FALSE);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_scene_4);
}

/*..........................................................................*/
static QState Light_scene_0_dimming_down(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        QTimeEvtCtr dimTimeout;
        /* handle registered (pending) local priority of groups */
        SelectA_HandleRegisteredLocalPrio(me->apartmentWide ? 0 : me->localZoneId, me->groupId, me->deviceId);
        /* call scene DEC or OPC_DEC */
        dimTimeout = Light_HandleDimmingAndGetTimeout(me, Scene_GetDecScene(me->buttonId), OPC_DEC);
        /* arm timers */
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, dimTimeout, CALL_DEC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        Op_QTimeEvt_postIn(me->opTimer2, me->parent, LIGHT_HOLD_END_TIMEOUT_TICKS, END_DEC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_disarm(me->opTimer2);
        return Q_HANDLED();
    }
    case CALL_DEC_SIG: {
        /* call scene DEC (continue) and re-arm timer */
        QTimeEvtCtr dimTimeout = Light_HandleDimmingAndGetTimeout(me, Scene_GetDecContinueScene(me->buttonId), OPC_DEC);
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, dimTimeout, CALL_DEC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        return Q_HANDLED();
    }
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        /* re-arm timer */
        Op_QTimeEvt_rearm(me->opTimer2, LIGHT_HOLD_END_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case CT_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG:
        /* fall through */
    case END_DEC_SIG: {
        if (me->outputChannel != 0) {
            /* send opc stop command */
            SelectA_CallOPCDimZoneGroup(me->localZoneId, me->groupId, OPC_STOP, me->outputChannel, me->deviceId);
        } else if (e->sig != END_DEC_SIG) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        return Q_TRAN(&Light_scene_0_dim_up);
    }
    }
    return Q_SUPER(&Light_off);
}

/*..........................................................................*/
static QState Light_scene_0_dim_up(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Light_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        if (Datamodel_DsmPropertiesRoomDimmingDisabled()) {
            break;
        }
        Light_SendSTOPForGroup(me);
        return Q_TRAN(&Light_scene_0_dimming_up);
    }
    }
    return Q_SUPER(&Light_scene_0);
}

/*..........................................................................*/
static QState Light_scene_0_dimming_up(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        uint8_t numShortDSMsgs;
        /* handle registered (pending) local priority of groups */
        SelectA_HandleRegisteredLocalPrio(me->apartmentWide ? 0 : me->localZoneId, me->groupId, me->deviceId);
        /* call CallSceneMin_A with scene as parameter */
        Scene scene = Scene_GetOnScene(me->buttonId);
        numShortDSMsgs = SelectA_CallSceneMinZoneGroup(me->apartmentWide? 0 : me->localZoneId, me->groupId, scene, me->deviceId);
        /* arm timers */
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, numShortDSMsgs * LIGHT_SHORT_MESSAGE_TIMEOUT_TICKS, CALL_INC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        Op_QTimeEvt_postIn(me->opTimer2, me->parent, LIGHT_HOLD_END_TIMEOUT_TICKS, END_INC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_disarm(me->opTimer2);
        return Q_HANDLED();
    }
    case CALL_INC_SIG: {
        QTimeEvtCtr dimTimeout;
        /* call scene INC (continue) and re-arm timer */
        dimTimeout = Light_HandleDimmingAndGetTimeout(me, Scene_GetIncContinueScene(me->buttonId), OPC_INC);
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, dimTimeout, CALL_INC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        return Q_HANDLED();
    }
    case CT_2WAY_UP_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        /* re-arm timer */
        Op_QTimeEvt_rearm(me->opTimer2, LIGHT_HOLD_END_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case CT_2WAY_UP_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG:
        /* fall through */
    case END_INC_SIG: {
        if (me->outputChannel != 0) {
            /* send opc stop command */
            SelectA_CallOPCDimZoneGroup(me->localZoneId, me->groupId, OPC_STOP, me->outputChannel, me->deviceId);
        } else if (e->sig != END_INC_SIG) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        return Q_TRAN(&Light_scene_1_dim_down);
    }
    }
    return Q_SUPER(&Light_on);
}

/*..........................................................................*/
static QTimeEvtCtr Light_HandleDimmingAndGetTimeout(Light *me, Scene scene, OPCOperation operation)
{
    QTimeEvtCtr dimTimeout;
    /* call scene or opc operation */
    if (me->outputChannel == 0) {
        SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->apartmentWide, me->deviceId);
        Light_StopSTOPTimer(me);
        dimTimeout = LIGHT_DIM_TIMEOUT_TICKS;
    } else {
        SelectA_CallOPCDimZoneGroup(me->localZoneId, me->groupId, operation, me->outputChannel, me->deviceId);
        dimTimeout = LIGHT_OPC_DIM_TIMEOUT_TICKS;
    }
    return dimTimeout;
}

/*..........................................................................*/
static QState Light_dimming_down(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        QTimeEvtCtr dimTimeout;
        /* handle registered (pending) local priority of groups */
        SelectA_HandleRegisteredLocalPrio(me->apartmentWide ? 0 : me->localZoneId, me->groupId, me->deviceId);
        /* call scene DEC or OPC_DEC */
        dimTimeout = Light_HandleDimmingAndGetTimeout(me, Scene_GetDecScene(me->buttonId), OPC_DEC);
        /* arm timers */
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, dimTimeout, CALL_DEC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        Op_QTimeEvt_postIn(me->opTimer2, me->parent, LIGHT_HOLD_END_TIMEOUT_TICKS, END_DEC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_disarm(me->opTimer2);
        return Q_HANDLED();
    }
    case CALL_DEC_SIG: {
        /* call scene DEC (continue) and re-arm timer */
        QTimeEvtCtr dimTimeout = Light_HandleDimmingAndGetTimeout(me, Scene_GetDecContinueScene(me->buttonId), OPC_DEC);
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, dimTimeout, CALL_DEC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        return Q_HANDLED();
    }
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        /* re-arm timer */
        Op_QTimeEvt_rearm(me->opTimer2, LIGHT_HOLD_END_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_on);
}

/*..........................................................................*/
static QState Light_dimming_up(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        QTimeEvtCtr dimTimeout;
        /* handle registered (pending) local priority of groups */
        SelectA_HandleRegisteredLocalPrio(me->apartmentWide ? 0 : me->localZoneId, me->groupId, me->deviceId);
        /* call scene INC or OPC_INC */
        dimTimeout = Light_HandleDimmingAndGetTimeout(me, Scene_GetIncScene(me->buttonId), OPC_INC);
        /* arm timers */
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, dimTimeout, CALL_INC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        Op_QTimeEvt_postIn(me->opTimer2, me->parent, LIGHT_HOLD_END_TIMEOUT_TICKS, END_INC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_disarm(me->opTimer2);
        return Q_HANDLED();
    }
    case CALL_INC_SIG: {
        QTimeEvtCtr dimTimeout;
        /* call scene INC (continue) and re-arm timer */
        dimTimeout = Light_HandleDimmingAndGetTimeout(me, Scene_GetIncContinueScene(me->buttonId), OPC_INC);
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, dimTimeout, CALL_INC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, me->apartmentWide, me->outputChannel);
        return Q_HANDLED();
    }
    case CT_2WAY_UP_HOLD_REPEAT_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        /* re-arm timer */
        Op_QTimeEvt_rearm(me->opTimer2, LIGHT_HOLD_END_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_on);
}

/*..........................................................................*/
static QState Light_scene_1_dim_down(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Light_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    default:
        break;
    }
    return Q_SUPER(&Light_scene_1);
}

/*..........................................................................*/
static QState Light_scene_1_dimming_down(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG:
        /* fall through */
    case END_DEC_SIG: {
        if (me->outputChannel != 0) {
            /* send opc stop command */
            SelectA_CallOPCDimZoneGroup(me->localZoneId, me->groupId, OPC_STOP, me->outputChannel, me->deviceId);
        } else if (e->sig != END_DEC_SIG) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        return Q_TRAN(&Light_scene_1_dim_up);
    }
    }
    return Q_SUPER(&Light_dimming_down);
}

/*..........................................................................*/
static QState Light_scene_1_dim_up(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Light_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        if (Datamodel_DsmPropertiesRoomDimmingDisabled()) {
            break;
        }
        Light_SendSTOPForGroup(me);
        return Q_TRAN(&Light_scene_1_dimming_up);
    }
    }
    return Q_SUPER(&Light_scene_1);
}

/*..........................................................................*/
static QState Light_scene_1_dimming_up(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_2WAY_UP_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG:
        /* fall through */
    case END_INC_SIG: {
        if (me->outputChannel != 0) {
            /* send opc stop command */
            SelectA_CallOPCDimZoneGroup(me->localZoneId, me->groupId, OPC_STOP, me->outputChannel, me->deviceId);
        } else if (e->sig != END_INC_SIG) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        return Q_TRAN(&Light_scene_1_dim_down);
    }
    }
    return Q_SUPER(&Light_dimming_up);
}

/*..........................................................................*/
static QState Light_scene_2_dim_down(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Light_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    default: {
        break;
    }
    }
    return Q_SUPER(&Light_scene_2);
}

/*..........................................................................*/
static QState Light_scene_2_dimming_down(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG:
        /* fall through */
    case END_DEC_SIG: {
        if (me->outputChannel != 0) {
            /* send opc stop command */
            SelectA_CallOPCDimZoneGroup(me->localZoneId, me->groupId, OPC_STOP, me->outputChannel, me->deviceId);
        } else if (e->sig != END_DEC_SIG) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        return Q_TRAN(&Light_scene_2_dim_up);
    }
    }
    return Q_SUPER(&Light_dimming_down);
}

/*..........................................................................*/
static QState Light_scene_2_dim_up(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Light_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        if (Datamodel_DsmPropertiesRoomDimmingDisabled()) {
            break;
        }
        Light_SendSTOPForGroup(me);
        return Q_TRAN(&Light_scene_2_dimming_up);
    }
    }
    return Q_SUPER(&Light_scene_2);
}

/*..........................................................................*/
static QState Light_scene_2_dimming_up(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_2WAY_UP_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG:
        /* fall through */
    case END_INC_SIG: {
        if (me->outputChannel != 0) {
            /* send opc stop command */
            SelectA_CallOPCDimZoneGroup(me->localZoneId, me->groupId, OPC_STOP, me->outputChannel, me->deviceId);
        } else if (e->sig != END_INC_SIG) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        return Q_TRAN(&Light_scene_2_dim_down);
    }
    }
    return Q_SUPER(&Light_dimming_up);
}

/*..........................................................................*/
static QState Light_scene_3_dim_down(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Light_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    default: {
        break;
    }
    }
    return Q_SUPER(&Light_scene_3);
}

/*..........................................................................*/
static QState Light_scene_3_dimming_down(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG:
        /* fall through */
    case END_DEC_SIG: {
        if (me->outputChannel != 0) {
            /* send opc stop command */
            SelectA_CallOPCDimZoneGroup(me->localZoneId, me->groupId, OPC_STOP, me->outputChannel, me->deviceId);
        } else if (e->sig != END_DEC_SIG) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        return Q_TRAN(&Light_scene_3_dim_up);
    }
    }
    return Q_SUPER(&Light_dimming_down);
}

/*..........................................................................*/
static QState Light_scene_3_dim_up(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Light_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        if (Datamodel_DsmPropertiesRoomDimmingDisabled()) {
            break;
        }
        Light_SendSTOPForGroup(me);
        return Q_TRAN(&Light_scene_3_dimming_up);
    }
    }
    return Q_SUPER(&Light_scene_3);
}

/*..........................................................................*/
static QState Light_scene_3_dimming_up(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_2WAY_UP_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG:
        /* fall through */
    case END_INC_SIG: {
        if (me->outputChannel != 0) {
            /* send opc stop command */
            SelectA_CallOPCDimZoneGroup(me->localZoneId, me->groupId, OPC_STOP, me->outputChannel, me->deviceId);
        } else if (e->sig != END_INC_SIG) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        return Q_TRAN(&Light_scene_3_dim_down);
    }
    }
    return Q_SUPER(&Light_dimming_up);
}

/*..........................................................................*/
static QState Light_scene_4_dim_down(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Light_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    default: {
        break;
    }
    }
    return Q_SUPER(&Light_scene_4);
}

/*..........................................................................*/
static QState Light_scene_4_dimming_down(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG:
        /* fall through */
    case END_DEC_SIG: {
        if (me->outputChannel != 0) {
            /* send opc stop command */
            SelectA_CallOPCDimZoneGroup(me->localZoneId, me->groupId, OPC_STOP, me->outputChannel, me->deviceId);
        } else if (e->sig != END_DEC_SIG) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        return Q_TRAN(&Light_scene_4_dim_up);
    }
    }
    return Q_SUPER(&Light_dimming_down);
}

/*..........................................................................*/
static QState Light_scene_4_dim_up(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Light_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        if (Datamodel_DsmPropertiesRoomDimmingDisabled()) {
            break;
        }
        Light_SendSTOPForGroup(me);
        return Q_TRAN(&Light_scene_4_dimming_up);
    }
    }
    return Q_SUPER(&Light_scene_4);
}

/*..........................................................................*/
static QState Light_scene_4_dimming_up(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_2WAY_UP_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG:
        /* fall through */
    case END_INC_SIG: {
        if (me->outputChannel != 0) {
            /* send opc stop command */
            SelectA_CallOPCDimZoneGroup(me->localZoneId, me->groupId, OPC_STOP, me->outputChannel, me->deviceId);
        } else if (e->sig != END_INC_SIG) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        return Q_TRAN(&Light_scene_4_dim_down);
    }
    }
    return Q_SUPER(&Light_dimming_up);
}

/*..........................................................................*/
static QState Light_prog_mode(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* initialize programming mode */
        me->progMode->deviceIndex = 0;
        me->progMode->dSMAddress = DS485_INVALID_ADDRESS;
        me->progMode->selectPreviousDevice = FALSE;
        me->progMode->localZoneId = me->localZoneId;
        me->progMode->groupId = me->groupId;
        QTimeEvt_postIn(&(me->progMode->exitTimer), me->parent, PROG_EXIT_TIMEOUT_TICKS);
        /* start alive timer */
        QTimeEvt_postIn(&(me->progMode->aliveTimer), me->parent, REMOTE_PROG_ALIVE_TICKS);
        return Q_HANDLED();
    }
    case SELECT_A_REMOTE_PROG_MODE_ALIVE_SIG: {
        /* send Prog_ON on group (only remote) */
        DsmApi_SendZoneGroupActionProgOn(me->progMode->localZoneId, me->progMode->groupId, me->deviceId);
        /* rearm alive timer */
        QTimeEvt_rearm(&(me->progMode->aliveTimer), REMOTE_PROG_ALIVE_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        QTimeEvt_disarm(&(me->progMode->exitTimer));
        QTimeEvt_disarm(&(me->progMode->aliveTimer));
        return Q_HANDLED();
    }
    case Q_INIT_SIG: {
        return Q_TRAN(&Light_prog_blink_group);
    }
    case CT_2WAY_UP_TIP_3X_SIG:
    case CT_2WAY_DOWN_TIP_3X_SIG:
    case CT_2WAY_DOWN_CLICK_3X_SIG:
    case CT_CLICK_3X_SIG:
        /* fall through */
    case CT_TIP_3X_SIG: {
        /* send Prog_OFF */
        SelectA_ProgOffZoneGroup(me->progMode->localZoneId, me->progMode->groupId, me->deviceId);
        /* exit programming mode without saving scene */
        return Q_TRAN(&Light_prog_mode_exit_blink_group);
    }
    case CT_SHORT_LONG_SIG: {
        Scene scene;
        /* check initiating device ID */
        if (me->progMode->initiatingDeviceId != me->deviceId) {
            /* just send general acknowledge */
            SelectA_SendGeneralAck();
            return Q_HANDLED();
        }
        /* save scene and exit programming mode */
        switch (me->progMode->sceneIndex) {
        case 1:
            scene = Scene_GetOnScene(me->progMode->buttonId);
            break;
        case 2:
            scene = Scene_GetScene2(me->progMode->buttonId);
            break;
        case 3:
            scene = Scene_GetScene3(me->progMode->buttonId);
            break;
        case 4:
            scene = Scene_GetScene4(me->progMode->buttonId);
            break;
        default:
            /* should never get here */
            Q_ERROR();
            return Q_HANDLED();
        }
        SelectA_SaveSceneZoneGroup(me->progMode->localZoneId, me->progMode->groupId, scene, me->deviceId);
        /* send Prog_OFF */
        SelectA_ProgOffZoneGroup(me->progMode->localZoneId, me->progMode->groupId, me->deviceId);
        /* exit programming mode */
        return Q_TRAN(&Light_prog_mode_exit_blink_group);
    }
    case CT_2WAY_UP_HOLD_START_SIG:
    case CT_2WAY_UP_HOLD_END_SIG:
    case CT_2WAY_DOWN_HOLD_START_SIG:
    case CT_2WAY_DOWN_HOLD_END_SIG:
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_END_SIG: {
        /* send general acknowledge if hold end signal is not handled */
        SelectA_SendGeneralAck();
        return Q_HANDLED();
    }
    case CALL_SCENE_BROADCAST_SIG: {
        /* handle only broadcast call scene */
        SceneBasicEvt const * sceneBasicEvt = (SceneBasicEvt const *)e;
        if (SelectA_LightAlarmScene() == sceneBasicEvt->sceneId) {
            /* save back button ID and zone ID */
            me->buttonId = me->progMode->buttonId;
            me->localZoneId = me->progMode->localZoneId;
            /* reset programming mode */
            me->progMode->localZoneId = 0;
            me->progMode->groupId = 0;
            /* exit programming mode without saving */
            switch (me->progMode->sceneIndex) {
            case 1:
                return Q_TRAN(&Light_special_room_state_return_scene_1);
            case 2:
                return Q_TRAN(&Light_special_room_state_return_scene_2);
            case 3:
                return Q_TRAN(&Light_special_room_state_return_scene_3);
            case 4:
                return Q_TRAN(&Light_special_room_state_return_scene_4);
            default:
                break;
            }
            /* should never get here */
            Q_ERROR();
            return Q_HANDLED();
        }
        /* don't return Q_HANDLED()! */
        break;
    }
    case SELECT_A_EXIT_PROG_MODE_TIMEOUT_SIG: {
        /* send Prog_OFF */
        SelectA_ProgOffZoneGroup(me->progMode->localZoneId, me->progMode->groupId, me->deviceId);
        return Q_TRAN(&Light_prog_mode_exit_blink_group);
    }
    case CT_LOCAL_STOP_SIG:
        /* fall through */
    case LOCAL_STOP_SIG: {
        Light_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState Light_prog_mode_exit_blink_group(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* start timer */
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, LIGHT_PROG_OFF_TIMEOUT_TICKS, PROG_BLINK_TIMEOUT_SIG, 0, me->progMode->localZoneId, me->progMode->groupId, me->deviceId, FALSE, me->outputChannel);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* stop timer */
        Op_QTimeEvt_disarm(me->opTimer1);
        return Q_HANDLED();
    }
    case PROG_BLINK_TIMEOUT_SIG: {
        /* save back button ID and zone ID */
        me->buttonId = me->progMode->buttonId;
        me->localZoneId = me->progMode->localZoneId;
        /* reset programming mode */
        me->progMode->localZoneId = 0;
        me->progMode->groupId = 0;
        if (Datamodel_ButtonId_ButtonGroupId(me->buttonId) == 0) {
            Light_PostGroupSignal(me, CALL_ROOM_ON_SIG, TRUE);

            switch (me->progMode->sceneIndex) {
            case 1:
                return Q_TRAN(&Light_scene_1_local);
            case 2:
                return Q_TRAN(&Light_scene_2_local);
            case 3:
                return Q_TRAN(&Light_scene_3_local);
            case 4:
                return Q_TRAN(&Light_scene_4_local);
            default:
                break;
            }
        } else {
            Scene scene = Scene_GetOnScene(me->buttonId);
            SelectA_CallSceneZoneGroupApartment(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, FALSE, me->deviceId);
            if (me->progMode->setLocalPriority) {
                SelectA_SetSceneLocalPrioZoneGroup(me->localZoneId, me->groupId, scene, me->deviceId);
            }
            return Q_TRAN(&Light_scene_0);
        }
        /* should never get here */
        Q_ERROR();
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState Light_prog_blink_group(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* reset blinking device found flag */
        me->progMode->progDeviceFound = FALSE;
        /* send Prog_ON on group */
        SelectA_ProgOnZoneGroup(me->progMode->localZoneId, me->progMode->groupId, me->deviceId);
        /* start timer */
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, LIGHT_PROG_ON_TIMEOUT_TICKS, PROG_BLINK_TIMEOUT_SIG, 0, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        Light_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* stop timer */
        Op_QTimeEvt_disarm(me->opTimer1);
        return Q_HANDLED();
    }
    case PROG_BLINK_TIMEOUT_SIG: {
        return Q_TRAN(&Light_prog_get_device_count);
    }
    }
    return Q_SUPER(&Light_prog_mode);
}

/*..........................................................................*/
static QState Light_prog_get_device_count(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        DatamodelReqZoneDeviceCountEvt * reqEvt =
            Q_NEW(DatamodelReqZoneDeviceCountEvt, DATAMODEL_REQ_ZONE_DEVICE_COUNT_SIG);
        reqEvt->basic.requester = me->parent;
        reqEvt->zoneIdModifier = ZONE_INFO_BY_ID;
        reqEvt->modifier = ZONE_DEVICE_COUNT_ONLY_ACTIVE;
        reqEvt->zoneId = me->progMode->localZoneId;
        QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
        return Q_HANDLED();
    }
    case DATAMODEL_RES_ZONE_DEVICE_COUNT_SIG: {
        DatamodelResZoneDeviceCountEvt const * resEvt = (DatamodelResZoneDeviceCountEvt const *)e;
        me->progMode->deviceCount = resEvt->deviceCount;
        me->progMode->deviceIndex = me->progMode->selectPreviousDevice && me->progMode->deviceCount? me->progMode->deviceCount-1 : me->progMode->deviceIndex;
        return Q_TRAN(&Light_prog_get_output_device);
    }
    }
    return Q_SUPER(&Light_prog_mode);
}

/*..........................................................................*/
static QState Light_prog_get_output_device(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        DatamodelReqDeviceInfoOfIndexEvt * reqEvt =
            Q_NEW(DatamodelReqDeviceInfoOfIndexEvt, DATAMODEL_REQ_DEVICE_INFO_OF_INDEX_SIG);
        reqEvt->basic.requester = me->parent;
        reqEvt->zoneIdModifier = ZONE_INFO_BY_ID;
        reqEvt->modifier = DEVICE_INFO_BY_INDEX_ONLY_ACTIVE;
        reqEvt->zoneId = me->progMode->localZoneId;
        reqEvt->deviceIndex = me->progMode->deviceIndex;
        reqEvt->privateId = SELECT_A_PRIVATE_ID_PROGRAMMING_MODE;
        QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
        /* arm timer */
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, LIGHT_PROG_DATAMODEL_TIMEOUT_TICKS, PROG_DEVICE_TIMEOUT_SIG, 0, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timer */
        Op_QTimeEvt_disarm(me->opTimer1);
        return Q_HANDLED();
    }
    case DATAMODEL_RES_DEVICE_ENTRY_SIG: {
        /* handle response */
        DatamodelResDeviceEntryEvt const * resEvt = (DatamodelResDeviceEntryEvt const *)e;
        Q_ASSERT(resEvt->privateId == SELECT_A_PRIVATE_ID_PROGRAMMING_MODE);
        uint16_t fid = resEvt->device.fid;
        if (IsBitSet8Set(resEvt->device.groupMembership, me->groupId) &&
            Fid_HasOutput(fid) &&
            MODEHasOutputLoad(resEvt->device.mode) &&
            !resEvt->device.locked) {
            /* light device has an output, a load is configured and it's not locked -> save device ID and let blink device */
            me->progMode->deviceId = resEvt->device.deviceId;
            return Q_TRAN(&Light_prog_device);
        }
        Light_IdentifyNextRemoteDeviceIndex(me);
        if (me->progMode->deviceIndex < me->progMode->deviceCount) {
            /* get next output device */
            return Q_TRAN(&Light_prog_get_output_device);
        }
        /* virtual dsm can not support remote programming */
        return Q_TRAN(&Light_prog_get_device_count);
    }
    case PROG_DEVICE_TIMEOUT_SIG: {
        /* timeout waiting for data model device response */
        return Q_TRAN(&Light_prog_get_device_count);
    }
    }
    return Q_SUPER(&Light_prog_mode);
}

/*..........................................................................*/
static QState Light_prog_device(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* set programming device found flag */
        me->progMode->progDeviceFound = TRUE;
        /* send Prog_ON on device */
        if (me->progMode->dSMAddress == DS485_INVALID_ADDRESS) {
            SelectA_SendProgOnDevice(me->progMode->deviceId);
        } else {
            DsmApi_SendDeviceActionProgOn(me->progMode->dSMAddress, me->progMode->deviceId);
        }
        return Q_HANDLED();
    }
    case Q_INIT_SIG: {
        return Q_TRAN(&Light_prog_device_on);
    }
    case CT_2WAY_UP_TIP_1X_SIG: {
        if (Light_SendSTOPProgrammingMode(me)) {
            return Q_HANDLED();
        }
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        Light_RearmExitTimer(me);
        if (me->progMode->dSMAddress == DS485_INVALID_ADDRESS) {
            SelectA_SendSceneDevice(me->progMode->deviceId, MAX_S);
        } else {
            DsmApi_SendDeviceActionCallSceneRequest(me->progMode->dSMAddress, me->progMode->deviceId, MAX_S);
        }
        Light_RestartSTOPTimer(me);
        return Q_TRAN(&Light_prog_device_on);
    }
    case CT_2WAY_DOWN_TIP_1X_SIG: {
        if (Light_SendSTOPProgrammingMode(me)) {
            return Q_HANDLED();
        }
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        return Q_TRAN(&Light_prog_device_off_action);
    }
    case CT_2WAY_UP_TIP_2X_SIG:
    case CT_2WAY_DOWN_TIP_2X_SIG:
    case CT_TIP_2X_SIG: {
        /* stop timer */
        Op_QTimeEvt_disarm(me->opTimer1);
        /* fall through */
    }
    case CT_CLICK_2X_SIG: {
        Light_RearmExitTimer(me);
        if (e->sig == CT_2WAY_DOWN_TIP_2X_SIG) {
            me->progMode->selectPreviousDevice = TRUE;
        } else {
            me->progMode->selectPreviousDevice = FALSE;
        }
        /* select next output device */
        Light_IdentifyNextRemoteDeviceIndex(me);
        if (me->progMode->dSMAddress == DS485_INVALID_ADDRESS) {
            /* select next output device */
            if (me->progMode->deviceIndex < me->progMode->deviceCount) {
                /* get next output device */
                return Q_TRAN(&Light_prog_get_output_device);
            }
            /* send general acknowledge */
            SelectA_SendGeneralAck();
            /* start with remote devices */
            return Q_TRAN(&Light_prog_get_next_dsm_address);
        }
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        /* continue with remote devices */
        return Q_TRAN(&Light_prog_get_next_remote_device);
    }
    case CT_2WAY_UP_HOLD_START_SIG:
    case CT_2WAY_UP_HOLD_REPEAT_SIG: {
        Light_RearmExitTimer(me);
        if (me->progMode->dSMAddress != DS485_INVALID_ADDRESS) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        return Q_TRAN(&Light_prog_device_on_dimming_up);
    }
    case CT_2WAY_DOWN_HOLD_START_SIG:
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG: {
        Light_RearmExitTimer(me);
        if (me->progMode->dSMAddress != DS485_INVALID_ADDRESS) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        return Q_TRAN(&Light_prog_device_on_dimming_down);
    }
    }
    return Q_SUPER(&Light_prog_mode);
}

/*..........................................................................*/
static QState Light_prog_device_on(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_INIT_SIG: {
        return Q_TRAN(&Light_prog_device_on_dim_down);
    }
    case Q_ENTRY_SIG: {
        /* don't call call scene MAX_S here as this is the entry point when a device is selected! */
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* stop timer */
        Op_QTimeEvt_disarm(me->opTimer1);
        return Q_HANDLED();
    }
    case CT_CLICK_1X_SIG:
        /* fall through */
    case CT_TIP_1X_SIG: {
        if (Light_SendSTOPProgrammingMode(me)) {
            return Q_HANDLED();
        }
        /* stop timer */
        Op_QTimeEvt_disarm(me->opTimer1);
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        /* start tip 1 action timer */
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, LIGHT_PROG_TIP_1X_TIMEOUT_TICKS, TIP_1X_ACTION_SIG, 0, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        return Q_HANDLED();
    }
    case TIP_1X_ACTION_SIG: {
        Light_RearmExitTimer(me);
        return Q_TRAN(&Light_prog_device_off_action);
    }
    }
    return Q_SUPER(&Light_prog_device);
}

/*..........................................................................*/
static QState Light_prog_device_on_dim_down(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        Light_RearmExitTimer(me);
        if (me->progMode->dSMAddress != DS485_INVALID_ADDRESS) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        return Q_TRAN(&Light_prog_device_on_dimming_down);
    }
    }
    return Q_SUPER(&Light_prog_device_on);
}

static QTimeEvtCtr Light_ProgDeviceHandleDimmingAndGetTimeout(Light *me, Scene scene, OPCOperation operation)
{
    QTimeEvtCtr dimTimeout = (me->outputChannel == 0)? LIGHT_PROG_DIM_TIMEOUT_TICKS : LIGHT_PROG_OPC_DIM_TIMEOUT_TICKS;
    /* call scene or opc operation */
    if (me->progMode->dSMAddress == DS485_INVALID_ADDRESS) {
        if (me->outputChannel == 0) {
            SelectA_SendSceneDevice(me->progMode->deviceId, scene);
        } else {
            SelectA_SendOPCDimDevice(me->progMode->deviceId, operation, me->outputChannel);
        }
    } else {
        if (me->outputChannel == 0) {
            DsmApi_SendDeviceActionCallSceneRequest(me->progMode->dSMAddress, me->progMode->deviceId, scene);
        } else {
            DsmApi_SendDeviceActionOPCDimRequest(me->progMode->dSMAddress, me->progMode->deviceId, operation, me->outputChannel);
        }
    }
    return dimTimeout;
}

/*..........................................................................*/
static QState Light_prog_device_on_dimming_down(Light *me, QEvent const *e)
{
    static bool s_localStopReceived = FALSE;
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        s_localStopReceived = FALSE;
        QTimeEvtCtr dimTimeout = Light_ProgDeviceHandleDimmingAndGetTimeout(me, DEC_S, OPC_DEC);
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, dimTimeout, CALL_DEC_SIG, 0, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        Op_QTimeEvt_postIn(me->opTimer2, me->parent, LIGHT_HOLD_END_TIMEOUT_TICKS, END_DEC_SIG, 0, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        Light_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_disarm(me->opTimer2);
        return Q_HANDLED();
    }
    case CALL_DEC_SIG: {
        QTimeEvtCtr dimTimeout = Light_ProgDeviceHandleDimmingAndGetTimeout(me, DEC_S, OPC_DEC);
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, dimTimeout, CALL_DEC_SIG, 0, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        return Q_HANDLED();
    }
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG:
    case CT_HOLD_REPEAT_SIG: {
        Light_RearmExitTimer(me);
        if (me->progMode->dSMAddress != DS485_INVALID_ADDRESS) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        /* re-arm timer */
        Op_QTimeEvt_rearm(me->opTimer2, LIGHT_HOLD_END_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case CT_CLICK_1X_SIG:
        /* fall through */
    case CT_TIP_1X_SIG: {
        /* change state and re-send this signal */
        {
            /* re-post event for main HSM */
            ClickForwardEvt * newEvent = Q_NEW(ClickForwardEvt, INTERNAL_CLICK_FORWARD_SIG);
            newEvent->groupId = me->groupId;
            newEvent->localZoneId = me->localZoneId;
            newEvent->event = e->sig;
            newEvent->buttonId = me->buttonId;
            newEvent->deviceId = me->deviceId;
            newEvent->apartmentWide = FALSE;
            newEvent->outputChannel = me->outputChannel;
//            newEvent->fid = me->fid;
//            newEvent->key = me->key;
            QActive_postLIFO(me->parent, (QEvent *)newEvent);
        }
        /* fall through */
    case CT_2WAY_DOWN_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG: {
        Light_RearmExitTimer(me);
        /* fall through */
    }
    }
    case END_DEC_SIG: {
        if (me->outputChannel != 0) {
            Light_ProgDeviceHandleDimmingAndGetTimeout(me, INVALID_SCENE, OPC_STOP);
        }
        if ((me->progMode->dSMAddress == DS485_INVALID_ADDRESS) && (e->sig != END_DEC_SIG)) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        if (s_localStopReceived) {
            return Q_TRAN(&Light_prog_device_off);
        } else {
            return Q_TRAN(&Light_prog_device_on_dim_up);
        }
    }
    case CT_LOCAL_STOP_SIG:
    case LOCAL_STOP_SIG: {
        s_localStopReceived = TRUE;
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_prog_device_on);
}

/*..........................................................................*/
static QState Light_prog_device_on_dim_up(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        Light_RearmExitTimer(me);
        if (me->progMode->dSMAddress != DS485_INVALID_ADDRESS) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        return Q_TRAN(&Light_prog_device_on_dimming_up);
    }
    }
    return Q_SUPER(&Light_prog_device_on);
}

/*..........................................................................*/
static QState Light_prog_device_on_dimming_up(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        QTimeEvtCtr dimTimeout = Light_ProgDeviceHandleDimmingAndGetTimeout(me, INC_S, OPC_INC);
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, dimTimeout, CALL_INC_SIG, 0, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        Op_QTimeEvt_postIn(me->opTimer2, me->parent, LIGHT_HOLD_END_TIMEOUT_TICKS, END_INC_SIG, 0, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        Light_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_disarm(me->opTimer2);
        return Q_HANDLED();
    }
    case CALL_INC_SIG: {
        QTimeEvtCtr dimTimeout = Light_ProgDeviceHandleDimmingAndGetTimeout(me, INC_S, OPC_INC);
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, dimTimeout, CALL_INC_SIG, 0, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        return Q_HANDLED();
    }
    case CT_2WAY_UP_HOLD_REPEAT_SIG:
    case CT_HOLD_REPEAT_SIG: {
        Light_RearmExitTimer(me);
        if (me->progMode->dSMAddress != DS485_INVALID_ADDRESS) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        /* re-arm timer */
        Op_QTimeEvt_rearm(me->opTimer2, LIGHT_HOLD_END_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case CT_CLICK_1X_SIG:
        /* fall through */
    case CT_TIP_1X_SIG: {
        /* change state and re-send this signal */
        {
            /* re-post event for main HSM */
            ClickForwardEvt * newEvent = Q_NEW(ClickForwardEvt, INTERNAL_CLICK_FORWARD_SIG);
            newEvent->groupId = me->groupId;
            newEvent->localZoneId = me->localZoneId;
            newEvent->event = e->sig;
            newEvent->buttonId = me->buttonId;
            newEvent->deviceId = me->deviceId;
            newEvent->apartmentWide = FALSE;
            newEvent->outputChannel = me->outputChannel;
//            newEvent->fid = me->fid;
//            newEvent->key = me->key;
            QActive_postLIFO(me->parent, (QEvent *)newEvent);
        }
        /* fall through */
    }
    case CT_2WAY_UP_HOLD_END_SIG:
        /* fall through */
    case CT_HOLD_END_SIG: {
        Light_RearmExitTimer(me);
        /* fall through */
    }
    case END_INC_SIG: {
        if (me->outputChannel != 0) {
            Light_ProgDeviceHandleDimmingAndGetTimeout(me, INVALID_SCENE, OPC_STOP);
        }
        if ((me->progMode->dSMAddress == DS485_INVALID_ADDRESS) && (e->sig != END_INC_SIG)) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        return Q_TRAN(&Light_prog_device_on_dim_down);
    }
    }
    return Q_SUPER(&Light_prog_device_on);
}

/*..........................................................................*/
static QState Light_prog_device_off(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_EXIT_SIG: {
        /* stop timer */
        Op_QTimeEvt_disarm(me->opTimer1);
        return Q_HANDLED();
    }
    case CT_CLICK_1X_SIG:
        /* fall through */
    case CT_TIP_1X_SIG: {
        if (Light_SendSTOPProgrammingMode(me)) {
            return Q_HANDLED();
        }
        /* stop timer */
        Op_QTimeEvt_disarm(me->opTimer1);
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        /* start tip 1 action timer */
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, LIGHT_PROG_TIP_1X_TIMEOUT_TICKS, TIP_1X_ACTION_SIG, 0, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        return Q_HANDLED();
    }
    case TIP_1X_ACTION_SIG: {
        Light_RearmExitTimer(me);
        if (me->progMode->dSMAddress == DS485_INVALID_ADDRESS) {
            SelectA_SendSceneDevice(me->progMode->deviceId, MAX_S);
        } else {
            DsmApi_SendDeviceActionCallSceneRequest(me->progMode->dSMAddress, me->progMode->deviceId, MAX_S);
        }
        Light_RestartSTOPTimer(me);
        return Q_TRAN(&Light_prog_device_on);
    }
    case CT_HOLD_START_SIG:
    case CT_HOLD_REPEAT_SIG: {
        Light_RearmExitTimer(me);
        if (me->progMode->dSMAddress != DS485_INVALID_ADDRESS) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
        }
        return Q_TRAN(&Light_prog_device_on_dimming_up);
    }
    }
    return Q_SUPER(&Light_prog_device);
}

/*..........................................................................*/
static QState Light_prog_device_off_action(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        if (me->progMode->dSMAddress == DS485_INVALID_ADDRESS) {
            SelectA_SendSceneDevice(me->progMode->deviceId, MIN_S);
        } else {
            DsmApi_SendDeviceActionCallSceneRequest(me->progMode->dSMAddress, me->progMode->deviceId, MIN_S);
        }
        Light_RestartSTOPTimer(me);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_prog_device_off);
}

/*..........................................................................*/
static QState Light_prog_no_device(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_CLICK_1X_SIG:
    case CT_TIP_1X_SIG:
    case CT_CLICK_2X_SIG:
    case CT_TIP_2X_SIG:
    case CT_TIP_4X_SIG: {
        /* send GeneralAck */
        SelectA_SendGeneralAck();
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_prog_mode);
}

/*..........................................................................*/
static QState Light_remote_prog_mode(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* initialize programming mode */
        me->progMode->deviceIndex = 0;
        me->progMode->localZoneId = me->localZoneId;
        me->progMode->groupId = me->groupId;
        QTimeEvt_postIn(&(me->progMode->exitTimer), me->parent, REMOTE_PROG_EXIT_TIMEOUT_TICKS);

        /* reset blinking device found flag */
        me->progMode->progDeviceFound = FALSE;
        return Q_HANDLED();
    }
    case PROG_ON_ZONE_SIG: {
        if (me->localZoneId == me->progMode->localZoneId) {
            QTimeEvt_rearm(&(me->progMode->exitTimer), REMOTE_PROG_EXIT_TIMEOUT_TICKS);
        }
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        QTimeEvt_disarm(&(me->progMode->exitTimer));
        return Q_HANDLED();
    }
    case CALL_SCENE_BROADCAST_SIG: {
        /* handle only broadcast call scene */
        SceneBasicEvt const * sceneBasicEvt = (SceneBasicEvt const *)e;
        if (SelectA_LightAlarmScene() == sceneBasicEvt->sceneId) {
            /* save back button ID and zone ID */
            me->buttonId = me->progMode->buttonId;
            me->localZoneId = me->progMode->localZoneId;
            /* reset programming mode */
            me->progMode->localZoneId = 0;
            me->progMode->groupId = 0;
            /* exit programming mode without saving */
            switch (me->progMode->sceneIndex) {
            case 0:
                return Q_TRAN(&Light_special_room_state_return_scene_0);
            case 1:
                return Q_TRAN(&Light_special_room_state_return_scene_1);
            case 2:
                return Q_TRAN(&Light_special_room_state_return_scene_2);
            case 3:
                return Q_TRAN(&Light_special_room_state_return_scene_3);
            case 4:
                return Q_TRAN(&Light_special_room_state_return_scene_4);
            default:
                break;
            }
            /* should never get here */
            Q_ERROR();
            return Q_HANDLED();
        }
        /* don't return Q_HANDLED()! */
        break;
    }
    case SELECT_A_EXIT_PROG_MODE_TIMEOUT_SIG: {
        /* terminate programming mode */
        /* send Prog_OFF */
        SelectA_SendProgOffZoneGroup(me->progMode->localZoneId, me->progMode->groupId);
        return Q_TRAN(&Light_prog_mode_exit_blink_group);
    }
    case PROG_OFF_ZONE_SIG: {
        /* send Prog_OFF */
        SelectA_SendProgOffZoneGroup(me->progMode->localZoneId, me->progMode->groupId);
        /* PROG_OFF is followed with a scene call, that synchronizes the state machines.
         * Transition to Light_scene_1 as fallback.
         */
        /* reset programming mode */
        me->progMode->localZoneId = 0;
        me->progMode->groupId = 0;
        return Q_TRAN(&Light_scene_1);
    }
    case CT_LOCAL_STOP_SIG: {
        DsmApi_SendZoneGroupActionLocalStop(me->progMode->localZoneId, me->progMode->groupId, me->deviceId);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState Light_remote_prog_mode_blink_zone(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* send Prog_ON on group */
        SelectA_SendProgOnZoneGroup(me->progMode->localZoneId, me->progMode->groupId);
        Light_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_remote_prog_mode);
}

/*..........................................................................*/
static void Light_IdentifyNextRemoteDeviceIndex(Light *me)
{
    if (me->progMode->selectPreviousDevice) {
        me->progMode->deviceIndex = me->progMode->deviceIndex? me->progMode->deviceIndex-1 : me->progMode->deviceCount;
    } else {
        ++me->progMode->deviceIndex;
    }
}

/*..........................................................................*/
static QState Light_prog_get_next_dsm_address(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        QSignal sig = me->progMode->selectPreviousDevice? DS485_REQ_GET_PREVIOUS_ACTIVE_BUS_MEMBER_SIG : DS485_REQ_GET_NEXT_ACTIVE_BUS_MEMBER_SIG;
        DS485ReqGetBusMemberByIndexEvt * reqEvt = Q_NEW(DS485ReqGetBusMemberByIndexEvt, sig);
        reqEvt->requester = me->parent;
        reqEvt->index = me->progMode->dSMAddress;
        QActive_postFIFO(AO_DS485, (QEvent *)reqEvt);
        return Q_HANDLED();
    }
    case DS485_RES_GET_BUS_MEMBER_BY_INDEX_SIG: {
        DS485ResGetBusMemberByIndexEvt const * resEvt = (DS485ResGetBusMemberByIndexEvt const *)e;
        me->progMode->dSMAddress = resEvt->address;
        me->progMode->deviceIndex = 0;
        if (me->progMode->dSMAddress == DS485_INVALID_ADDRESS) {
            if (me->progMode->progDeviceFound) {
                /* no additional dSMs --> return to local devices */
                return Q_TRAN(&Light_prog_get_device_count);
            } else {
                /* no applicable device */
                return Q_TRAN(&Light_prog_no_device);
            }
        } else {
            return Q_TRAN(&Light_prog_get_remote_device_count);
        }
    }
    }
    return Q_SUPER(&Light_prog_mode);
}

/*..........................................................................*/
static QState Light_prog_get_remote_device_count(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        DsmApi_SendZoneDeviceCountRequest(me->parent,
                                          me->progMode->dSMAddress,
                                          me->progMode->dSMAddress, /* transactionId */
                                          ZONE_DEVICE_COUNT_ONLY_ACTIVE,
                                          me->localZoneId);
        return Q_HANDLED();
    }
    case DS485_RX_DATA_SIG: {
        DS485MessageEvt const * msgEvt = (DS485MessageEvt const *)e;
        uint8_t error = DsmApi_GetResError(msgEvt);
        if (error == 0) {
            uint16_t zoneDeviceCount = DsmApi_GetResData16(msgEvt, 0);
            me->progMode->deviceCount = zoneDeviceCount;
            me->progMode->deviceIndex = me->progMode->selectPreviousDevice && me->progMode->deviceCount? me->progMode->deviceCount-1 : 0;
            return Q_TRAN(&Light_prog_get_next_remote_device);
        } else {
            /* dSM returned an error --> all devices done --> next dSM */
            /* fall through */
        }
    }
    case DSMAPI_REQUEST_TIMEOUT_SIG: {
        /* no response from requested dSM --> next dSM */
        return Q_TRAN(&Light_prog_get_next_dsm_address);
    }
    case Q_EXIT_SIG: {
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_prog_mode);
}

/*..........................................................................*/
static QState Light_prog_get_next_remote_device(Light *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        DsmApi_SendDeviceInfoRequest(me->parent,
                                     me->progMode->dSMAddress,
                                     me->progMode->dSMAddress, /* transactionId */
                                     DEVICE_INFO_BY_INDEX_ONLY_ACTIVE,
                                     me->progMode->localZoneId,
                                     me->progMode->deviceIndex);
        return Q_HANDLED();
    }
    case DS485_RX_DATA_SIG: {
        DS485MessageEvt const * msgEvt = (DS485MessageEvt const *)e;
        uint8_t error = DsmApi_GetResError(msgEvt);
        if (error == 0) {
            uint8_t locked = DsmApi_GetResData8(msgEvt, 13);
            uint8_t outputMode = DsmApi_GetResData8(msgEvt, 14);
            uint16_t fid = DsmApi_GetResData16(msgEvt, 6);
            uint8_t groups[8];
            DsmApi_GetResArray8(msgEvt, 16, groups, sizeof(groups));
            if (IsBitSet8Set(groups, me->groupId) && Fid_HasOutput(fid) && MODEHasOutputLoad(outputMode) && !locked) {
                /* device has an output */
                me->progMode->progDeviceFound = TRUE;
                me->progMode->deviceId = DsmApi_GetResData16(msgEvt, 0);
                return Q_TRAN(&Light_prog_device);
            } else {
                /* device has no output --> next device */
                Light_IdentifyNextRemoteDeviceIndex(me);
                if (me->progMode->deviceIndex != me->progMode->deviceCount) {
                    return Q_TRAN(&Light_prog_get_next_remote_device);
                } else {
                    /* zone on this dSM has no more devices --> next dSM */
                    /* fall through */
                }
            }
        }
        /* dSM returned an error --> all devices done --> next dSM */
        /* fall through */
    }
    case DSMAPI_REQUEST_TIMEOUT_SIG: {
        /* no response from requested dSM --> next dSM */
        return Q_TRAN(&Light_prog_get_next_dsm_address);
    }
    case Q_EXIT_SIG: {
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Light_prog_mode);
}
