/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Author: Sergey Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.
    

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>

#include "properties.h"
#include "selectB.h"
#include "dsmapi.h"
#include "registers.h"
#include "button.h"
#include "datamodel.h"
#include "sockutil.h"
#include "uthash.h"
#include "msgutil.h"
#include "proputils.h"
#include "util.h"
#include "channelutil.h"

static void send_sensor_value(uint16_t deviceId, uint8_t index, double value)
{
    uint8_t sensorType;

    if (Datamodel_GetDeviceSensorType(deviceId, index, &sensorType)) {
        int32_t retVal;

        switch (sensorType) {
        case 9:
        case 10:
            /* temperature */
            retVal = (int32_t)((value + 273.15 - 230.0) * 40);
            break;
        case 11:
        case 12:
            /* brightness */
            retVal = 800 * log10(value);
            break;
        case 13:
        case 14:
            /* humidity */
            retVal = (int32_t)(value * 40);
            break;
        case 15:
            /* air pressure */
            retVal = (int32_t)((value - 200) * 4);
            break;
        case 18:
            /* wind speed */
            retVal = (int32_t)(value * 40);
            break;
        case 19:
            /* wind direction */
            retVal = (int32_t)(value * 8);
            retVal %= 360;
            break;
        case 20:
            /* precipitation */
            retVal = (int32_t)(value * 40);
            break;
        case 21:
            /* carbon dioxide */
            retVal = 800 * log10(value);
            break;
        case 50:
            /* temperature set point */
            retVal = (int32_t)((value + 273.15 - 230.0) * 40);
            break;
        case 51:
            /* heating control level */
            retVal = (int32_t)(value + 100.0);
            break;
        case 64:
            /* output current mA */
            retVal = (int32_t)(value / 4);
            break;
        case 65:
            /* power consumption VA */
            retVal = (int32_t)(value);
            break;
        default:
            retVal = value;
            vdsm_report(LOG_WARNING, "send sensor value: sensor type %d not "
                                "handled\n", sensorType);
        }
        /* limit to 12 bit */
        uint16_t upstreamValue;
        if (retVal < 0) {
            upstreamValue = 0;
        } else if (retVal > 0xfff) {
            upstreamValue = 0xfff;
        } else {
            upstreamValue = retVal;
        }
        /* create long upstream event */
        LongUpstreamEvt *filteredUpEvt = Q_NEW(LongUpstreamEvt, LONG_UPSTREAM_SIG);
        filteredUpEvt->source = deviceId;
        filteredUpEvt->mainqualifier = UpstreamMainqualifierSensor;
        filteredUpEvt->subqualifier = UpstreamSubqualifierSensorSensorValues;
        filteredUpEvt->data = ((upstreamValue << 8) & 0xFF00) | ((index << 4) & 0xF0) | ((upstreamValue >> 8) & 0x0F);
        filteredUpEvt->quality = 62;
        filteredUpEvt->flags = 0;
        /* post event */
        QF_publish((QEvent*)filteredUpEvt);
        vdsm_report(LOG_DEBUG, "send sensor value: published long upstream event from device %d (value %d)\n", deviceId, retVal);
    } else {
        vdsm_report(LOG_WARNING, "send sensor value: no sensor found fo device %d and sensor %d\n", deviceId, index);
    }
}

static void prop_process_request_fid(vdc_client_t *vdc_client,
                                     Vdcapi__VdcResponseGetProperty *msg,
                                     Vdcapi__VdsmRequestGetProperty *request,
                                     request_cache_t *rc)
{
    if (rc->count == 0) {
        uint64_t result;
        if (!msgutil_get_uint_property(msg, "primaryGroup", UINT16_MAX,
                                       &result)) {
            vdsm_report(LOG_WARNING, "request fid: no property primaryGroup\n");
            free(rc);
            return;
        }
        rc->result |= ((uint16_t)result << 12);
        rc->result |= FM_VARIABLE_STD_GROUP; // set groupbits read
        rc->count = 1;
        vdc_api_send_get_property(vdc_client, request->dsuid,
                                  "buttonInputDescriptions", rc);
    } else if (rc->count == 1) {

        Vdcapi__PropertyElement *elemButtonInputDescription = NULL;
        if (proputil_get_property_find_element_by_name(msg, "buttonInputDescriptions",
                                                       &elemButtonInputDescription) != 0) {
            vdsm_report(LOG_INFO, "request fid: no property "
                             "buttonInputDescriptions\n");
            goto cleanup_prop_process_request_fid_1;
        }

        Vdcapi__PropertyElement *elemChannel = NULL;
        if (proputil_find_property_element_by_index(elemButtonInputDescription,
                                                    rc->channel,
                                                    &elemChannel) != 0) {
            vdsm_report(LOG_INFO, "request fid: channel %d not found in "
                                "buttonInputDescriptions\n", rc->channel);
            goto cleanup_prop_process_request_fid_1;
        }

        uint64_t buttonType = 0;
        if (get_property_element_name_uint(elemChannel, "buttonType", &buttonType) != 0) {
            vdsm_report(LOG_WARNING, "request fid: no property buttonType\n");
            goto cleanup_prop_process_request_fid_1;
        }

        if (buttonType == 0) {
            rc->result |= 0;
        } else if ( (buttonType == 1) ||
                    (buttonType == 2) ) {
            rc->result |= 1;
        } else if (buttonType >= 3) {
            rc->result |= 2;
        }
        // more button types

cleanup_prop_process_request_fid_1:
        rc->count = 2;
        vdc_api_send_get_property(vdc_client, request->dsuid,
                                  "outputDescription", rc);

    } else if (rc->count == 2) {

        Vdcapi__PropertyElement *elemOutputDescription = NULL;
        if (proputil_get_property_find_element_by_name(msg, "outputDescription",
                                                       &elemOutputDescription) != 0) {
            vdsm_report(LOG_INFO, "request fid: no property outputDescription\n");
            goto cleanup_prop_process_request_fid_2;
        }

        bool variableMap = FALSE;
        if (get_property_element_name_bool(elemOutputDescription, "variableRamp", &variableMap) != 0) {
            vdsm_report(LOG_WARNING, "request fid: no property variableRamp\n");
            goto cleanup_prop_process_request_fid_2;
        }

        rc->result |= 0x10;
        rc->result |= (variableMap) ? 0x20 : 0;

cleanup_prop_process_request_fid_2:
        rc->count = 3;
        vdc_api_send_get_property(vdc_client, request->dsuid,
                                  "binaryInputDescriptions", rc);

    } else if (rc->count == 3) {
        Vdcapi__PropertyElement *elemBinaryInputDescriptions = NULL;
        if (proputil_get_property_find_element_by_name(msg, "binaryInputDescriptions",
                                                       &elemBinaryInputDescriptions) != 0) {
            vdsm_report(LOG_INFO, "request fid: no property binaryInputDescriptions\n");
            goto cleanup_prop_process_request_fid_3;
        }

        Vdcapi__PropertyElement *elemChannel = NULL;
        if (proputil_find_property_element_by_index(elemBinaryInputDescriptions,
                                                    rc->channel,
                                                    &elemChannel) != 0) {
            vdsm_report(LOG_INFO, "request fid: channel %d not found in "
                                "binaryInputDescriptions\n", rc->channel);
            goto cleanup_prop_process_request_fid_3;
        }
        rc->result |= 0x08;

cleanup_prop_process_request_fid_3:
        rc->count = 4;
        vdc_api_send_get_property(vdc_client, request->dsuid,
                                  "sensorDescriptions", rc);
    } else if (rc->count == 4) {
        Vdcapi__PropertyElement *elemSensorDescriptions = NULL;
        if (proputil_get_property_find_element_by_name(msg, "sensorDescriptions",
                                                       &elemSensorDescriptions) != 0) {
            vdsm_report(LOG_INFO, "request fid: no property sensorDescriptions\n");
            goto cleanup_prop_process_request_fid_4;
        }

        Vdcapi__PropertyElement *elemChannel = NULL;
        if (proputil_find_property_element_by_index(elemSensorDescriptions,
                                                    rc->channel,
                                                    &elemChannel) != 0) {
            vdsm_report(LOG_INFO, "request fid: channel %d not found in "
                                "sensorDescriptions\n", rc->channel);
            goto cleanup_prop_process_request_fid_4;
        }
        rc->result |= 0x08;

cleanup_prop_process_request_fid_4:
        {
            LongUpstreamEvt *filteredUpEvt =
                    Q_NEW(LongUpstreamEvt, LONG_UPSTREAM_SIG);

            filteredUpEvt->source = Datamodel_GetDeviceIdFromDsuid(request->dsuid);
            filteredUpEvt->mainqualifier = 0;
            filteredUpEvt->subqualifier = rc->offset / 2;
            filteredUpEvt->data = (uint16_t)rc->result;
            filteredUpEvt->quality = 62;
            filteredUpEvt->flags = 0;
            /* post event */
            QF_publish((QEvent*)filteredUpEvt);
            vdsm_report(LOG_DEBUG, "get property fid: published event\n");
            free(rc);
        }
    } else {
        vdsm_report(LOG_WARNING, "get property fid: unexpected number of replies\n");
        free(rc);
    }
}

static void prop_process_request_mode(vdc_client_t *vdc_client,
                                      Vdcapi__VdcResponseGetProperty *msg,
                                      Vdcapi__VdsmRequestGetProperty *request,
                                      request_cache_t *rc)
{
    if (rc->count == 0) {
        Vdcapi__PropertyElement *elemOutputSettings = NULL;

        if (proputil_get_property_find_element_by_name(msg, "outputSettings",
                                                       &elemOutputSettings) != 0) {
            vdsm_report(LOG_INFO, "request mode: no property outputSettings\n");
            goto cleanup_request_mode_0;
        }

        uint64_t mode = 0;
        if (get_property_element_name_uint(elemOutputSettings, "mode", &mode) != 0) {
            vdsm_report(LOG_WARNING, "request mode: no property mode\n");
            goto cleanup_request_mode_0;
        }

        if (mode == 1) {
            rc->result |= 16;
        } else if (mode == 2) {
            rc->result |= 22;
        } else {
            rc->result |= mode;
        }

cleanup_request_mode_0:
        rc->count = 1;
        vdc_api_send_get_property(vdc_client, request->dsuid,
                                  "buttonInputSettings", rc);

    } else if (rc->count == 1) {
        Vdcapi__PropertyElement *elemButtonInputSettings = NULL;
        Vdcapi__PropertyElement *elemButtonInputId = NULL;

        if (proputil_get_property_find_element_by_name(msg, "buttonInputSettings",
                                                       &elemButtonInputSettings) != 0) {
            vdsm_report(LOG_INFO, "request mode: no property buttonInputSettings\n");
            goto cleanup_request_mode_1;
        }
        if (proputil_find_property_element_by_index(elemButtonInputSettings,
                                                    rc->channel,   // channel id
                                                    &elemButtonInputId) != 0) {
            vdsm_report(LOG_INFO, "request mode: channel %d not found in "
                                "buttonInputSettings\n", rc->channel);
            goto cleanup_request_mode_1;
        }

        uint64_t group = 0;
        uint64_t func = 0;
        uint64_t ltNumGrp = 0;
        if (get_property_element_name_uint(elemButtonInputId,"group", &group) != 0) {
            vdsm_report(LOG_WARNING, "request mode: no property group\n");
            goto cleanup_request_mode_1;
        }
        ltNumGrp |= (group << 4);

        if (get_property_element_name_uint(elemButtonInputId,"function", &func) != 0) {
            vdsm_report(LOG_WARNING, "request mode: no property function\n");
            goto cleanup_request_mode_1;
        }
        ltNumGrp |= (func & 0x0f);
        rc->result |= (ltNumGrp << 8);

cleanup_request_mode_1:
        {
            /* create long upstream event */
            LongUpstreamEvt *filteredUpEvt = Q_NEW(LongUpstreamEvt, LONG_UPSTREAM_SIG);
            filteredUpEvt->source = Datamodel_GetDeviceIdFromDsuid(request->dsuid);
            filteredUpEvt->mainqualifier = 0;
            filteredUpEvt->subqualifier = rc->offset / 2;
            filteredUpEvt->data = (uint16_t)rc->result;
            filteredUpEvt->quality = 62;
            filteredUpEvt->flags = 0;
            /* post event */
            QF_publish((QEvent*)filteredUpEvt);
            vdsm_report(LOG_DEBUG, "get property mode: published event. value: 0x%04x\n",
                   rc->result);
            // free allocated memory
            free(rc);
        }
    } else {
        vdsm_report(LOG_WARNING, "get property mode: unexpected number of "
                            "replies\n");
        free(rc);
    }
}

static void prop_process_request_ltmode(Vdcapi__VdcResponseGetProperty *msg,
                                        Vdcapi__VdsmRequestGetProperty *request,
                                        request_cache_t *rc)
{
    Vdcapi__PropertyElement* elemButtonInputSettings = NULL;
    Vdcapi__PropertyElement* elemChannelButton = NULL;
    uint64_t mode = 0;
    uint16_t retVal = 0;

    if (proputil_get_property_find_element_by_name(msg, "buttonInputSettings",
                                                   &elemButtonInputSettings) != 0) {
        vdsm_report(LOG_INFO, "request ltmode: no property buttonInputSettings\n");
        goto cleanup_request_ltmode;
    }
    if (proputil_find_property_element_by_index(elemButtonInputSettings,
                                                rc->channel,       // channel id
                                                &elemChannelButton) != 0) {
        vdsm_report(LOG_INFO, "request ltmode: channel %d not found in "
                            "buttonInputSettings\n", rc->channel);
        goto cleanup_request_ltmode;
    }
    if (get_property_element_name_uint(elemChannelButton, "mode", &mode) != 0) {
        vdsm_report(LOG_WARNING, "request ltmode: no property mode\n");
        goto cleanup_request_ltmode;
    }

    if ((mode >= 5) && (mode <= 12)) {
        // 2-way button with one dsuid
        retVal = 13;
    } else {
        retVal = (uint16_t)mode;
    }

cleanup_request_ltmode:
    {
        /* create long upstream event */
        LongUpstreamEvt *filteredUpEvt = Q_NEW(LongUpstreamEvt, LONG_UPSTREAM_SIG);
        filteredUpEvt->source = Datamodel_GetDeviceIdFromDsuid(request->dsuid);
        filteredUpEvt->mainqualifier = 0;
        filteredUpEvt->subqualifier = rc->offset / 2;
        filteredUpEvt->data = retVal;
        filteredUpEvt->quality = 62;
        filteredUpEvt->flags = 0;
        /* post event */
        QF_publish((QEvent*)filteredUpEvt);
        vdsm_report(LOG_DEBUG, "request ltmode: published event. value: %d\n", retVal);
        free(rc);
    }
}

static void prop_process_request_pbchannel(Vdcapi__VdcResponseGetProperty *msg,
                                           Vdcapi__VdsmRequestGetProperty *request,
                                           request_cache_t *rc)
{
    // Verify MRY
    Vdcapi__PropertyElement* elemButtonInputSettings = NULL;
    Vdcapi__PropertyElement* elemChannelButton = NULL;
    int64_t pbchannel = 0;
    uint16_t retVal = 0;

    if (proputil_get_property_find_element_by_name(msg, "buttonInputSettings",
                                                   &elemButtonInputSettings) != 0) {
        vdsm_report(LOG_INFO, "request pbchannel: no property buttonInputSettings\n");
        goto cleanup_request_pbchannel;
    }
    if (proputil_find_property_element_by_index(elemButtonInputSettings,
                                                rc->channel,       // channel id
                                                &elemChannelButton) != 0) {
        vdsm_report(LOG_INFO, "request pbchannel: channel %d not found in "
                            "buttonInputSettings\n", rc->channel);
        goto cleanup_request_pbchannel;
    }
    if (get_property_element_name_int(elemChannelButton, "channel", &pbchannel) != 0) {
        vdsm_report(LOG_WARNING, "request pbchannel: no property channel\n");
        goto cleanup_request_pbchannel;
    }
    retVal = (uint16_t)pbchannel;

cleanup_request_pbchannel:
    {
        /* create long upstream event */
        LongUpstreamEvt *filteredUpEvt = Q_NEW(LongUpstreamEvt, LONG_UPSTREAM_SIG);
        filteredUpEvt->source = Datamodel_GetDeviceIdFromDsuid(request->dsuid);
        filteredUpEvt->mainqualifier = 0;
        filteredUpEvt->subqualifier = rc->offset / 2;
        filteredUpEvt->data = retVal;
        filteredUpEvt->quality = 62;
        filteredUpEvt->flags = 0;
        /* post event */
        QF_publish((QEvent*)filteredUpEvt);
        vdsm_report(LOG_DEBUG, "request pbchannel: published event. value: %d\n", retVal);
        free(rc);
    }
}

static void prop_process_request_pbgroup(Vdcapi__VdcResponseGetProperty *msg,
                                         Vdcapi__VdsmRequestGetProperty *request,
                                         request_cache_t *rc)
{
    // Verify MRY
    // Assumption to map "group" to PBGROUP.
    Vdcapi__PropertyElement* elemButtonInputSettings = NULL;
    Vdcapi__PropertyElement* elemChannelButton = NULL;
    int64_t pbgroup = 0;
    uint16_t retVal = 0;

    if (proputil_get_property_find_element_by_name(msg, "buttonInputSettings",
                                                   &elemButtonInputSettings) != 0) {
        vdsm_report(LOG_INFO, "request pbgroup: no property buttonInputSettings\n");
        goto cleanup_request_pbgroup;
    }
    if (proputil_find_property_element_by_index(elemButtonInputSettings,
                                                rc->channel,       // channel id
                                                &elemChannelButton) != 0) {
        vdsm_report(LOG_INFO, "request pbgroup: channel %d not found in "
                            "buttonInputSettings\n", rc->channel);
        goto cleanup_request_pbgroup;
    }
    if (get_property_element_name_int(elemChannelButton, "group", &pbgroup) != 0) {
        vdsm_report(LOG_INFO, "request pbgroup: no property group\n");
        goto cleanup_request_pbgroup;
    }
    retVal = (uint16_t)pbgroup;

cleanup_request_pbgroup:
    {
        /* create long upstream event */
        LongUpstreamEvt *filteredUpEvt = Q_NEW(LongUpstreamEvt, LONG_UPSTREAM_SIG);
        filteredUpEvt->source = Datamodel_GetDeviceIdFromDsuid(request->dsuid);
        filteredUpEvt->mainqualifier = 0;
        filteredUpEvt->subqualifier = rc->offset / 2;
        filteredUpEvt->data = retVal;
        filteredUpEvt->quality = 62;
        filteredUpEvt->flags = 0;
        /* post event */
        QF_publish((QEvent*)filteredUpEvt);
        vdsm_report(LOG_DEBUG, "request pbgroup: published event. value: %d\n", retVal);
        free(rc);
    }
}

static void prop_process_request_outval(Vdcapi__VdcResponseGetProperty *msg,
                                        Vdcapi__VdsmRequestGetProperty *request,
                                        request_cache_t *rc)
{
    uint16_t deviceId = Datamodel_GetDeviceIdFromDsuid(request->dsuid);
    uint64_t retval = 0;
    Vdcapi__PropertyElement* elemChannelStates = NULL;
    Vdcapi__PropertyElement* elemChannelId = NULL;
    Vdcapi__PropertyElement* elemValue = NULL;

    if (proputil_get_property_find_element_by_name(msg, "channelStates", &elemChannelStates) != 0) {
        vdsm_report(LOG_INFO, "request outval: no property channelStates\n");
        goto cleanup_request_outval;
    }

    if (proputil_find_property_element_by_index(elemChannelStates, rc->channel, &elemChannelId) != 0) {
        vdsm_report(LOG_WARNING, "request outval: channel %d not found in "
                            "channelStates\n", rc->channel);
        goto cleanup_request_outval;
    }

    if (proputil_find_property_element_by_name(elemChannelId, "value", &elemValue) != 0) {
        vdsm_report(LOG_WARNING, "request outval: no property value\n");
        goto cleanup_request_outval;
    }

    double vdcapiValue;
    if (!msgutil_get_double(elemValue->value, &vdcapiValue)) {
        vdsm_report(LOG_WARNING, "request outval: value is no double\n");
        goto cleanup_request_outval;
    }

    // convert output data to dsm representation
    uint16_t dsmVal = 0;

    if (!convert_vcdapival_to_dsmval(rc->channel, vdcapiValue, &dsmVal)) {
        vdsm_report(LOG_WARNING, "request outval: could not convert value for "
               "device %d and channel %d\n", deviceId, rc->channel);
        goto cleanup_request_outval;
    }
    retval = dsmVal;

cleanup_request_outval:
    {
        LongUpstreamEvt *filteredUpEvt = Q_NEW(LongUpstreamEvt, LONG_UPSTREAM_SIG);
        filteredUpEvt->source = deviceId;
        filteredUpEvt->mainqualifier = 0;
        filteredUpEvt->subqualifier = rc->channel;
        filteredUpEvt->data = (uint16_t)retval;
        filteredUpEvt->quality = 62;
        filteredUpEvt->flags = 0;
        /* post event */
        QF_publish((QEvent*)filteredUpEvt);
        vdsm_report(LOG_DEBUG, "get property output value: published event. value: %d\n",
               retval);
        free(rc);
    }
}

static void prop_process_request_scenes(Vdcapi__VdcResponseGetProperty *msg,
                                        Vdcapi__VdsmRequestGetProperty *request,
                                        request_cache_t *rc)
{
    uint16_t rval = 0;

    Vdcapi__PropertyElement* elemScenes = NULL;
    Vdcapi__PropertyElement* elemSceneId[2];

    if (proputil_get_property_find_element_by_name(msg, "scenes", &elemScenes) != 0) {
        vdsm_report(LOG_INFO, "request scenes: no property scenes\n");
        goto cleanup_prop_process_request_scenes;
    }

    if (proputil_find_property_element_by_index(elemScenes, (rc->offset&0x7E),
                                                &elemSceneId[0]) != 0) {
        vdsm_report(LOG_WARNING, "request scenes: offset %d not found in "
                            "scenes\n", rc->offset & 0x7E);
        goto cleanup_prop_process_request_scenes;
    }

    if (proputil_find_property_element_by_index(elemScenes, (rc->offset&0x7E)+1,
                                                &elemSceneId[1]) != 0) {
        vdsm_report(LOG_WARNING, "request scenes: offset+1 %d not found in "
                            "scenes\n", (rc->offset & 0x7E) + 1);
        goto cleanup_prop_process_request_scenes;
    }

    if (rc->offset < 128) {
        int i;
        for (i = 0; i < 2; ++i) {
            Vdcapi__PropertyElement* elemChannelValue = NULL;

            if (proputil_find_property_element_by_name_index(elemSceneId[i],
                                                             "channels",
                                                             rc->channel,
                                                             &elemChannelValue) != 0) {
                vdsm_report(LOG_WARNING, "request scenes %d: channel %d not found "
                                    "in channels\n", i, rc->channel);
                continue;
            }

            // no dsm conversion needed. This is only for channel 0: (8 bit only)
            // Bank 128 queries only channel 0.
            double v = 0;
            if (get_property_element_name_double(elemChannelValue,"value",&v) != 0) {
                vdsm_report(LOG_WARNING, "request scenes %d: double value not found "
                                    "in value\n", i);
                continue;
            }
            rval |= ((uint16_t)v << (8 * i));
        }
    } else {
        int i;
        for (i = 0; i < 2; ++i) {

            bool dontCare;
            if (get_property_element_name_bool(elemSceneId[i],"dontCare",&dontCare) != 0) {
                vdsm_report(LOG_WARNING, "request scenes %d: bool value not found "
                                    "in dontCare\n", i);
                goto cleanup_prop_process_request_scenes;
            }

            bool ignoreLocalPriority;
            if (get_property_element_name_bool(elemSceneId[i],"ignoreLocalPriority",&ignoreLocalPriority) != 0) {
                vdsm_report(LOG_WARNING, "request scenes %d: bool value not found "
                                    "in ignoreLocalPriority\n", i);
                goto cleanup_prop_process_request_scenes;
            }

            uint64_t effect;
            if (get_property_element_name_uint(elemSceneId[i],"effect", &effect) != 0) {
                vdsm_report(LOG_WARNING, "request scenes %d: uint64 value not found "
                                    "in effect\n", i);
                goto cleanup_prop_process_request_scenes;
            }

            // ledcon   0: nothing,  1: auto, 2: off
            // dimtime  1: standard, 2: slow

            // effect 4 is "blink/alert" in vDC API
            bool flashmode = (effect==4);
            // dimtime corresponds with effect 1..3, all other effects are mapped to dimtime 1
            uint8_t dimtime = effect==0 || effect>=4 ? 1 : effect;

            uint8_t sceneConfig = 0;
            if (dontCare)            sceneConfig |= 0x01;
            if (ignoreLocalPriority) sceneConfig |= 0x02;
            if (flashmode)           sceneConfig |= 0x08;
            sceneConfig |= (dimtime<<6);

            rval |= (sceneConfig << (8 * i));
        }
    }

cleanup_prop_process_request_scenes:
    {
        /* create long upstream event */
        LongUpstreamEvt *filteredUpEvt = Q_NEW(LongUpstreamEvt, LONG_UPSTREAM_SIG);
        filteredUpEvt->source = Datamodel_GetDeviceIdFromDsuid(request->dsuid);
        filteredUpEvt->mainqualifier = 0;
        filteredUpEvt->subqualifier = rc->offset / 2;
        filteredUpEvt->data = rval;
        filteredUpEvt->quality = 62;
        filteredUpEvt->flags = 0;
        /* post event */
        QF_publish((QEvent*)filteredUpEvt);
        vdsm_report(LOG_DEBUG, "request scenes: published event. value: 0x%04x\n",
               rval);
        free(rc);
    }
}

static void prop_process_request_name(Vdcapi__VdcResponseGetProperty *msg,
                                      Vdcapi__VdsmRequestGetProperty *request)
{
    uint16_t deviceId = Datamodel_GetDeviceIdFromDsuid(request->dsuid);
    const char *name = NULL;

    if (msg->n_properties <= 0) {
        vdsm_report(LOG_WARNING, "request name: no properties in result\n");
    } else {

        if (msgutil_get_string_property(msg, "name", &name)) {
            DatamodelReqDeviceNameEvt * reqEvt =
                    Q_NEW(DatamodelReqDeviceNameEvt, DATAMODEL_REQ_DEVICE_SET_NAME_SIG);
            reqEvt->basic.requester = NULL;
            reqEvt->deviceAddress = deviceId;
            strncpy((char *)reqEvt->name, name, MAX_NAME_LEN);
            QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
        }
    }

    /* create long upstream event (notifies selectB SM to proceed) */
    LongUpstreamEvt *filteredUpEvt = Q_NEW(LongUpstreamEvt, LONG_UPSTREAM_SIG);
    filteredUpEvt->source = deviceId;
    filteredUpEvt->mainqualifier = 0;
    filteredUpEvt->subqualifier = 0;
    filteredUpEvt->data = 0;
    filteredUpEvt->quality = 62;
    filteredUpEvt->flags = 0;
    /* post event */
    QF_publish((QEvent*)filteredUpEvt);
    vdsm_report(LOG_DEBUG, "request name: published event for device %d. "
                      "name: %s\n", deviceId, name ? name:"NULL");
}

static void prop_process_request_sensors(Vdcapi__VdcResponseGetProperty *msg,
                                         Vdcapi__VdsmRequestGetProperty *request)
{
    uint16_t deviceId = Datamodel_GetDeviceIdFromDsuid(request->dsuid);

    Vdcapi__PropertyElement* elemSensorDescriptions = NULL;
    if (proputil_get_property_find_element_by_name(msg,
                                                   "sensorDescriptions",
                                                   &elemSensorDescriptions) != 0) {
        vdsm_report(LOG_INFO, "request sensors: no property sensorDescriptions\n");
        Datamodel_RemoveSpareSensorEntries(deviceId, 0);
        goto cleanup_request_sensors;
    }

    Vdcapi__PropertyElement* elemSensor = NULL;

    size_t i;
    for (i = 0; i < elemSensorDescriptions->n_elements; ++i) {
        elemSensor = elemSensorDescriptions->elements[i];
        if (!elemSensor) {
            vdsm_report(LOG_WARNING, "request sensors %d: invalid element\n", i);
            continue;
        }

        const char *name = NULL;
        if (get_property_element_name_string(elemSensor, "name", &name) != 0) {
            vdsm_report(LOG_WARNING, "request sensors %d: no property name\n", i);
            continue;
        }

        uint64_t sensorType = 0;
        if (get_property_element_name_uint(elemSensor,"sensorType", &sensorType) != 0) {
            vdsm_report(LOG_WARNING, "request sensors %d: no property "
                                "sensorType\n", i);
            continue;
        }

        uint64_t sensorUsage = 0;
        if (get_property_element_name_uint(elemSensor,"sensorUsage",&sensorUsage) != 0) {
            vdsm_report(LOG_WARNING, "request sensorsi %d: no property "
                                "sensorUsage\n", i);
            continue;
        }

        double updateInterval = 0;
        if (get_property_element_name_double(elemSensor,"aliveSignInterval",&updateInterval) != 0) {
            vdsm_report(LOG_WARNING, "request sensors %d: no property "
                                "aliveSignInterval\n", i);
            continue;
        }
        vdsm_report(LOG_DEBUG, "request sensors: found sensor %s (index %d/%d),"
               " type %d, usage %d\n", name, i+1, elemSensorDescriptions->n_elements, sensorType, sensorUsage);

        uint8_t ds_type = 0xff;
        uint8_t precision = 0;
        if ((sensorType == 1) && (sensorUsage == 1)) {
            /* indoor temperature */
            ds_type = 9;
            precision = 0;
        } else if ((sensorType == 1) && (sensorUsage == 2)) {
            /* outdoor temperature */
            ds_type = 10;
            precision = 0;
        } else if ((sensorType == 2) && (sensorUsage == 1)) {
            /* indoor humidity sensor */
            ds_type = 13;
            precision = 0;
        } else if ((sensorType == 2) && (sensorUsage == 2)) {
            /* outdoor humidity sensor */
            ds_type = 14;
            precision = 0;
        } else if ((sensorType == 3) && (sensorUsage == 1)) {
            /* indoor lux sensor */
            ds_type = 11;
            precision = 0;
        } else if ((sensorType == 3) && (sensorUsage == 2)) {
            /* outdoor lux sensor */
            ds_type = 12;
            precision = 0;
        } else if (sensorType == 5) {
            /* Carbon dioxide ppm */
            ds_type = 21;
            precision = 0;
        } else if (sensorType == 13) {
            /* wind speed */
            ds_type = 18;
            precision = 0;
        } else if (sensorType == 14) {
            /* pressure/atmosphere */
            ds_type = 15;
            precision = 0;
        }
        if (ds_type != 0xff) {
            vdsm_report(LOG_DEBUG, "request sensors %d: appropriate ds sensor "
                              "type %d\n", i, ds_type);
        } else {
            vdsm_report(LOG_WARNING, "request sensors %d: warning, no appropriate "
                                "ds sensor type mapped\n", i);
        }
        /* update datamodel */
        Datamodel_SetDeviceSensor(deviceId, i, ds_type, (uint32_t)updateInterval, 0, precision);
        Datamodel_SetDeviceSensorName(deviceId, i, name);
    }
    Datamodel_RemoveSpareSensorEntries(deviceId, elemSensorDescriptions->n_elements);

cleanup_request_sensors:
    {
        /* create long upstream event (notifies selectB SM to proceed) */
        LongUpstreamEvt *filteredUpEvt = Q_NEW(LongUpstreamEvt, LONG_UPSTREAM_SIG);
        filteredUpEvt->source = deviceId;
        filteredUpEvt->mainqualifier = 0;
        filteredUpEvt->subqualifier = 0;
        filteredUpEvt->data = 0;
        filteredUpEvt->quality = 62;
        filteredUpEvt->flags = 0;
        /* post event */
        QF_publish((QEvent*)filteredUpEvt);
        vdsm_report(LOG_DEBUG, "request sensors: published event for device %d\n",
               deviceId);
    }
}

static void prop_process_request_opc_table(vdc_client_t *vdc_client,
                                           Vdcapi__VdcResponseGetProperty *msg,
                                           Vdcapi__VdsmRequestGetProperty *request,
                                           request_cache_t *rc)
{
    uint16_t deviceId = Datamodel_GetDeviceIdFromDsuid(request->dsuid);

    if (!msg) {
        vdsm_report(LOG_WARNING, "request opc table: invalid message\n");
        goto cleanup_request_opc_table;
    }

    Vdcapi__PropertyElement* elemChannelDescriptions = NULL;
    if (proputil_get_property_find_element_by_name(msg, "channelDescriptions", &elemChannelDescriptions) != 0) {
        vdsm_report(LOG_INFO, "request opc table: no property channelDescriptions\n");
        Datamodel_RemoveSpareOpcTableEntries(deviceId, 0);
        goto cleanup_request_opc_table;
    }

    if (rc->count == 0) {

        size_t i;
        for (i = 0; i < elemChannelDescriptions->n_elements; ++i) {
            if (!elemChannelDescriptions->elements[i]->name) {
                vdsm_report(LOG_WARNING, "request opc table %d: invalid name element\n", i);
                continue;
            }

            unsigned long channelId;
            channelId = strtoul(elemChannelDescriptions->elements[i]->name, NULL, 0);
            vdsm_report(LOG_DEBUG, "request opc table %d: converted channelId %d\n", i, channelId);

            const char *name = NULL;
            if (get_property_element_name_string(elemChannelDescriptions->elements[i], "name", &name) != 0) {
                vdsm_report(LOG_WARNING, "request opc table %d: no property name\n", i);
                continue;
            }

            uint64_t index;
            if (get_property_element_name_uint(elemChannelDescriptions->elements[i],"channelIndex",&index) != 0) {
                vdsm_report(LOG_WARNING, "request opc table %d: no property channelIndex\n", i);
                continue;
            }

            double minVal;
            if (get_property_element_name_double(elemChannelDescriptions->elements[i],"min",&minVal) != 0) {
                vdsm_report(LOG_WARNING, "request opc table %d: no property min\n", i);
                continue;
            }

            double maxVal;
            if (get_property_element_name_double(elemChannelDescriptions->elements[i],"max",&maxVal) != 0) {
                vdsm_report(LOG_WARNING, "request opc table %d: no property max\n", i);
                continue;
            }

            double resolution;
            if (get_property_element_name_double(elemChannelDescriptions->elements[i],"resolution",&resolution) != 0) {
                vdsm_report(LOG_WARNING, "request opc table %d: no property resolution\n", i);
                continue;
            }

            if (!util_check_valid_params(channelId, minVal, maxVal)) {
                vdsm_report(LOG_WARNING, "request opc table: Channel ranges are not valid: channel:%i, min:%lld, max:%lld\n",
                            channelId, minVal, maxVal);
            }

            // Update opc channels
            Datamodel_SetOpcTableValue(deviceId, (uint8_t)index, (uint8_t)channelId);
        }
        Datamodel_RemoveSpareOpcTableEntries(deviceId, elemChannelDescriptions->n_elements);


        rc->count = 1;
        // Additional query: get default channel
        // Default channel can be queried with explicitely querying channel "0"
        Vdcapi__Message *msg = NULL;
        Vdcapi__VdsmRequestGetProperty *getprop_submsg = NULL;
        Vdcapi__PropertyElement* elemPropName = NULL;
        Vdcapi__PropertyElement* elemPropIndex = NULL;

        msg = malloc(sizeof(Vdcapi__Message));
        if (!msg) {
            vdsm_report(LOG_ERR, "request opc table: failed to init message\n");
            goto cleanup_request_opc_table;
        }
        vdcapi__message__init(msg);
        if (proputil_get_property_new(request->dsuid, &getprop_submsg) != 0) {
            vdsm_report(LOG_ERR, "request opc table: failed to init submessage\n");
            free(msg);
            goto cleanup_request_opc_table;
        }

        if (proputil_create_property_empty(&elemPropName, "channelDescriptions") != 0) {
            vdsm_report(LOG_ERR, "request opc table: failed to init element name\n");
            proputil_get_property_free(getprop_submsg);
            free(msg);
            goto cleanup_request_opc_table;
        }
        if (proputil_get_property_add(getprop_submsg, elemPropName) !=0) {
            vdsm_report(LOG_ERR, "request opc table: failed to add element name\n");
            proputil_property_recursive_free(elemPropName);
            proputil_get_property_free(getprop_submsg);
            free(msg);
            goto cleanup_request_opc_table;
        }

        if (proputil_create_property_empty(&elemPropIndex, "0") != 0) {
            vdsm_report(LOG_ERR, "request opc table: failed to init element index\n");
            proputil_property_recursive_free(elemPropName);
            proputil_get_property_free(getprop_submsg);
            free(msg);
            goto cleanup_request_opc_table;
        }
        if (proputil_add_property_element(elemPropName, elemPropIndex) != 0) {
            vdsm_report(LOG_ERR, "request opc table: failed to add element index\n");
            proputil_property_recursive_free(elemPropIndex);
            proputil_property_recursive_free(elemPropName);
            proputil_get_property_free(getprop_submsg);
            free(msg);
            goto cleanup_request_opc_table;
        }

        msg->vdsm_request_get_property = getprop_submsg;
        msg->type = VDCAPI__TYPE__VDSM_REQUEST_GET_PROPERTY;
        vdc_api_send_message(vdc_client, msg, rc);
        return;
    } else if (rc->count == 1) {
        // get default channel
        if (elemChannelDescriptions->n_elements > 0) {
            if (!elemChannelDescriptions->elements[0]->name) {
                vdsm_report(LOG_WARNING, "request opc table: invalid element name\n");
                goto cleanup_request_opc_table;
            }

            unsigned long channelId;
            channelId = strtoul(elemChannelDescriptions->elements[0]->name, NULL, 0);
            vdsm_report(LOG_DEBUG, "request opc table: converted channelId %d\n", channelId);

            const char *name = NULL;
            if (get_property_element_name_string(elemChannelDescriptions->elements[0], "name", &name) != 0) {
                vdsm_report(LOG_WARNING, "request opc table: no property name\n");
                goto cleanup_request_opc_table;
            }

            uint64_t channelIndex;
            if (get_property_element_name_uint(elemChannelDescriptions->elements[0],"channelIndex",&channelIndex) != 0) {
                vdsm_report(LOG_WARNING, "request opc table: no property channelIndex\n");
                goto cleanup_request_opc_table;
            }

            double minVal;
            if (get_property_element_name_double(elemChannelDescriptions->elements[0],"min",&minVal) != 0) {
                vdsm_report(LOG_WARNING, "request opc table: no property min\n");
                goto cleanup_request_opc_table;
            }

            double maxVal;
            if (get_property_element_name_double(elemChannelDescriptions->elements[0],"max",&maxVal) != 0) {
                vdsm_report(LOG_WARNING, "request opc table: no property max\n");
                goto cleanup_request_opc_table;
            }

            double resolution;
            if (get_property_element_name_double(elemChannelDescriptions->elements[0],"resolution",&resolution) != 0) {
                vdsm_report(LOG_WARNING, "request opc table: no property resolution\n");
                goto cleanup_request_opc_table;
            }

            if (!util_check_valid_params(channelId, minVal, maxVal)) {
                vdsm_report(LOG_WARNING, "request opc table: Default Channel ranges are not valid: channel:%i, min:%lld, max:%lld\n",
                            channelId, minVal, maxVal);
            }

            Datamodel_SetDefaultChannel(deviceId);
        } else {
            Datamodel_RemoveDefaultChannel(deviceId);
        }
    }

cleanup_request_opc_table:
    {
        free(rc);
        /* create long upstream event (notifies selectB SM to proceed) */
        LongUpstreamEvt *filteredUpEvt = Q_NEW(LongUpstreamEvt, LONG_UPSTREAM_SIG);
        filteredUpEvt->source = deviceId;
        filteredUpEvt->mainqualifier = 0;
        filteredUpEvt->subqualifier = 0;
        filteredUpEvt->data = 0; // return value
        filteredUpEvt->quality = 62;
        filteredUpEvt->flags = 0;
        /* post event */
        QF_publish((QEvent*)filteredUpEvt);
        vdsm_report(LOG_DEBUG, "request opc table: published event for device %d\n",
               deviceId);
    }
}

static void prop_process_request_opc_scene_config(Vdcapi__VdcResponseGetProperty *msg,
                                                  Vdcapi__VdsmRequestGetProperty *request,
                                                  request_cache_t *rc)
{
    uint16_t rval = 0;

    if (msg->n_properties <= 0) {
        vdsm_report(LOG_WARNING, "request opc scene config: no properties in result\n");
    } else {
        Vdcapi__PropertyElement* elemScenes = NULL;
        Vdcapi__PropertyElement* elemSceneId = NULL;

        if (proputil_get_property_find_element_by_name(msg, "scenes", &elemScenes) != 0) {
            vdsm_report(LOG_INFO, "request opc scene config: no property scenes\n");
            goto cleanup_request_opc_scene_config;
        }
        if (proputil_find_property_element_by_index(elemScenes, rc->scene, &elemSceneId) != 0) {
            vdsm_report(LOG_WARNING, "request opc scene config: scene %d not found "
                                "in scenes\n", rc->scene);
            goto cleanup_request_opc_scene_config;
        }

        switch (rc->operation)
        {
        case DEVICE_O_P_C_CONFIG_GET_SCENE_CONFIG:
        {
            bool dontCare = FALSE;
            bool ignoreLocalPrio = FALSE;
            uint64_t effect = 0;

            // Ignore Local Prio
            if (get_property_element_name_bool(elemSceneId, "ignoreLocalPriority", &ignoreLocalPrio) != 0) {
                vdsm_report(LOG_WARNING, "request opc scene config: bool value "
                                    "not found in ignoreLocalPriority\n");
                goto cleanup_request_opc_scene_config;
            }

            // dontCare
            if (get_property_element_name_bool(elemSceneId, "dontCare", &dontCare) != 0) {
                vdsm_report(LOG_WARNING, "request opc scene config: bool value "
                                    "not found in dontCare\n");
                goto cleanup_request_opc_scene_config;
            }

            // effect
            if (get_property_element_name_uint(elemSceneId, "effect", &effect) != 0) {
                vdsm_report(LOG_WARNING, "request opc scene config: uint64 value "
                                    "not found in effect\n");
                goto cleanup_request_opc_scene_config;
            }

            rval = 0;
            if (dontCare) rval |= 0x01;
            if (ignoreLocalPrio) rval |= 0x02;

            // effect in vDC API is different from effect upstream
            // - upstream effect: 0=none, 1=flash
            // - vDC effect: 0..3 specifies transition speed (not used here) and 4=flash
            // So only vDC effect 4 (flash) can be mapped here:
            rval |= ((effect==4 ? 1 : 0)<<8)&0xff00;
        }
        break;

        case DEVICE_O_P_C_CONFIG_GET_DC_FLAGS:
        {
            int i;
            rval = 0;

            uint16_t deviceId = Datamodel_GetDeviceIdFromDsuid(request->dsuid);
            for (i = 0; i < 16; ++i) {

                uint8_t channelId = 0;
                if (!Datamodel_GetOpcChannelFromIndex(deviceId, i, &channelId)) {
                    vdsm_report(LOG_WARNING, "request opc scene config %d: channel "
                           "not found for device %d\n", i, deviceId);
                    continue;
                }

                Vdcapi__PropertyElement* elemChannelOutput = NULL;
                if (proputil_find_property_element_by_name_index(elemSceneId,
                                                                 "channels",
                                                                 channelId,
                                                                 &elemChannelOutput) != 0) {
                    vdsm_report(LOG_WARNING, "request opc scene config %d: "
                           "channel %d not found in channels\n", i, channelId);
                    continue;
                }

                bool dontCare;
                if (get_property_element_name_bool(elemChannelOutput,
                                                   "dontCare",
                                                   &dontCare) != 0) {
                    vdsm_report(LOG_WARNING, "request opc scene config %d: "
                           "bool value not found in dontCare\n", i);
                    continue;
                }
                if (dontCare) {
                    rval += (1 << i);
                }
            }
        }
        break;

        default:
            // oops
            vdsm_report(LOG_WARNING, "request opc scene config: wrong operation\n");
        }
    }

cleanup_request_opc_scene_config:
    {
        /* create long upstream event */
        LongUpstreamEvt *filteredUpEvt = Q_NEW(LongUpstreamEvt, LONG_UPSTREAM_SIG);
        filteredUpEvt->source = Datamodel_GetDeviceIdFromDsuid(request->dsuid);
        filteredUpEvt->mainqualifier = 0;
        filteredUpEvt->subqualifier = rc->scene;
        filteredUpEvt->data = rval;
        filteredUpEvt->quality = 62;
        filteredUpEvt->flags = 0;
        /* post event */
        QF_publish((QEvent*)filteredUpEvt);
        vdsm_report(LOG_DEBUG, "request opc scene config: published event. "
                          "value: 0x%04x\n", rval);
        free(rc);
    }
}

static void prop_process_request_opc_scene_value(Vdcapi__VdcResponseGetProperty *msg,
                                                 Vdcapi__VdsmRequestGetProperty *request,
                                                 request_cache_t *rc)
{
    uint16_t rval = 0;
    uint16_t deviceId = Datamodel_GetDeviceIdFromDsuid(request->dsuid);

    if (msg->n_properties <= 0) {
        vdsm_report(LOG_WARNING, "request opc scene: no properties in result\n");
    } else {
        Vdcapi__PropertyElement* elemScene = NULL;
        Vdcapi__PropertyElement* elemSceneId = NULL;
        Vdcapi__PropertyElement* elemChannelId = NULL;
        Vdcapi__PropertyElement* elemValue = NULL;

        if (proputil_get_property_find_element_by_name(msg, "scenes", &elemScene) != 0) {
            vdsm_report(LOG_INFO, "request opc scene: no property scenes\n");
            goto cleanup_request_opc_scene_value;
        }
        if (proputil_find_property_element_by_index(elemScene, rc->scene, &elemSceneId) != 0) {
            vdsm_report(LOG_WARNING, "request opc scene: scene %d not found "
                                "in scenes\n", rc->scene);
            goto cleanup_request_opc_scene_value;
        }

        if (proputil_find_property_element_by_name_index(elemSceneId, "channels", rc->channel, &elemChannelId) != 0) {
            vdsm_report(LOG_WARNING, "request opc scene: channel %d not found "
                                "in channels\n", rc->channel);
            goto cleanup_request_opc_scene_value;
        }

        if (proputil_find_property_element_by_name(elemChannelId, "value", &elemValue) != 0) {
            vdsm_report(LOG_WARNING, "request opc scene: no property value\n");
            goto cleanup_request_opc_scene_value;
        }

        // convert the data to dsm internal range
        double vdcapiValue;
        if (!msgutil_get_double(elemValue->value, &vdcapiValue)) {
            vdsm_report(LOG_WARNING, "request opc scene: value is no double\n");
            goto cleanup_request_opc_scene_value;
        }

        uint16_t dsmVal = 0;
        if (!convert_vcdapival_to_dsmval(rc->channel, vdcapiValue, &dsmVal)) {
            vdsm_report(LOG_WARNING, "request opc scene: could not convert value "
                   "for device %d and channel %d\n", deviceId, rc->channel);
            goto cleanup_request_opc_scene_value;
        }
        rval = dsmVal;
    }

cleanup_request_opc_scene_value:
    {
        /* create long upstream event */
        LongUpstreamEvt *filteredUpEvt = Q_NEW(LongUpstreamEvt, LONG_UPSTREAM_SIG);
        filteredUpEvt->source = deviceId;
        filteredUpEvt->mainqualifier = 0;
        filteredUpEvt->subqualifier = rc->channel;
        filteredUpEvt->data = rval;
        filteredUpEvt->quality = 62;
        filteredUpEvt->flags = 0;
        /* post event */
        QF_publish((QEvent*)filteredUpEvt);
        vdsm_report(LOG_DEBUG, "request opc scene: published event. value: 0x%04x\n",
               rval);
        free(rc);
    }
}

static void prop_process_request_vdc_name(Vdcapi__VdcResponseGetProperty *msg,
                                          Vdcapi__VdsmRequestGetProperty *request)
{
    const char *name = NULL;
    if (msgutil_get_string_property(msg, "name", &name)) {
        Datamodel_updateVdcName(request->dsuid, name);
    } else {
        vdsm_report(LOG_WARNING, "request vdc name: no property name in result\n");
    }
}

static void prop_process_request_bin_inputs(vdc_client_t *vdc_client,
                                            Vdcapi__VdcResponseGetProperty *msg,
                                            Vdcapi__VdsmRequestGetProperty *request,
                                            request_cache_t *rc)
{
    uint16_t deviceId = Datamodel_GetDeviceIdFromDsuid(request->dsuid);

    if (rc->count == 0) {
        Vdcapi__PropertyElement *elemInputDescriptions = NULL;
        if (proputil_get_property_find_element_by_name(msg, "binaryInputDescriptions",
                                                       &elemInputDescriptions) != 0) {
            vdsm_report(LOG_INFO, "request binary inputs: no property "
                             "binaryInputDescriptions\n");
            Datamodel_RemoveSpareBinaryInputs(deviceId, 0);
            goto cont_request_bin_inputs;
        }

        size_t i;
        Vdcapi__PropertyElement *elemInputDescId = NULL;
        for (i = 0; i < elemInputDescriptions->n_elements; ++i) {
            elemInputDescId = elemInputDescriptions->elements[i];

            if (!elemInputDescId->name) {
                vdsm_report(LOG_WARNING, "request binary inputs %d: invalid element "
                                    "name\n", i);
                continue;
            }

            unsigned long channelId;
            channelId = strtoul (elemInputDescId->name, NULL, 0);

            const char *name = NULL;
            if (get_property_element_name_string(elemInputDescId, "name", &name) != 0) {
                vdsm_report(LOG_WARNING, "request binary inputs %d: no property "
                                    "name\n", i);
                continue;
            }
            Datamodel_SetDeviceBinaryInputName(deviceId, channelId, name);
        }
        Datamodel_RemoveSpareBinaryInputs(deviceId, elemInputDescriptions->n_elements);

cont_request_bin_inputs:
        // get button input settings of device
        rc->count = 1;
        vdc_api_send_get_property(vdc_client, request->dsuid,
                                  "binaryInputSettings", rc);

    } else if (rc->count == 1) {
        Vdcapi__PropertyElement *elemBinaryInputSettings = NULL;
        if (proputil_get_property_find_element_by_name(msg, "binaryInputSettings",
                                                       &elemBinaryInputSettings) != 0) {
            vdsm_report(LOG_INFO, "request binary inputs: no property "
                             "binaryInputSettings\n");
            goto cont_request_bin_inputs_end;
        }

        size_t i;
        Vdcapi__PropertyElement *elemInputSettings = NULL;
        for (i = 0; i < elemBinaryInputSettings->n_elements; ++i) {
            elemInputSettings = elemBinaryInputSettings->elements[i];

            if (!elemInputSettings->name) {
                vdsm_report(LOG_WARNING, "request binary inputs %d: invalid element "
                                    "name (count = 1)\n", i);
                continue;
            }

            unsigned long channelId;
            channelId = strtoul (elemInputSettings->name, NULL, 0);

            uint64_t group;
            if (get_property_element_name_uint(elemInputSettings, "group", &group) != 0) {
                vdsm_report(LOG_WARNING, "request binary inputs %d: no property "
                                    "group\n", i);
                continue;
            }

            uint64_t sensorFunction;
            if (get_property_element_name_uint(elemInputSettings, "sensorFunction", &sensorFunction) != 0) {
                vdsm_report(LOG_WARNING, "request binary inputs %d: no property "
                                         "sensorFunction\n", i);
                continue;
            }
            // without configuration binary inputs should be "generic" (15, BUTTON_SERVER, ButtonId_Generic in modelconst.h of dSS)
            Datamodel_SetDeviceBinaryInput(deviceId, channelId, (uint8_t)group, (uint8_t)sensorFunction, BUTTON_SERVER);
        }

cont_request_bin_inputs_end:
        {
            LongUpstreamEvt *filteredUpEvt = Q_NEW(LongUpstreamEvt,
                                                   LONG_UPSTREAM_SIG);
            filteredUpEvt->source = deviceId;
            filteredUpEvt->mainqualifier = 0;
            filteredUpEvt->subqualifier = 0;
            filteredUpEvt->data = 0;
            filteredUpEvt->quality = 62;
            filteredUpEvt->flags = 0;
            /* post event */
            QF_publish((QEvent*)filteredUpEvt);
            vdsm_report(LOG_DEBUG, "request binary inputs: published event for "
                   "device %d\n", deviceId);
            free(rc);
        }
    }
}

static void prop_process_request_group_membership(
        Vdcapi__VdcResponseGetProperty *msg,
        request_cache_t *rc)
{
    uint16_t deviceId = rc->result;

    Vdcapi__PropertyElement* elemOutputSettings = NULL;
    Vdcapi__PropertyElement* elemGroups = NULL;
    Vdcapi__PropertyElement* elemGroupId = NULL;


    if (proputil_get_property_find_element_by_name(msg, "outputSettings",
                                                   &elemOutputSettings) != 0) {
        vdsm_report(LOG_INFO, "request groupMembership: no property outputSettings."
                              "Assign group membership from button group.\n");
        Datamodel_SetGroupMembershipFromButtonGroup(deviceId);
        goto cleanup_request_request_ltmode;
    }

    if (proputil_find_property_element_by_name(elemOutputSettings, "groups", &elemGroups) != 0) {
        vdsm_report(LOG_WARNING, "request groupMembership: no property groups\n");
        goto cleanup_request_request_ltmode;
    }

    Datamodel_ClearDeviceGroupMembership(deviceId);
    size_t i;
    for (i = 0; i < elemGroups->n_elements; ++i) {
        elemGroupId = elemGroups->elements[i];

        if (!elemGroupId->name) {
            vdsm_report(LOG_WARNING, "request groupMembership %d: invalid element "
                                "name\n", i);
            continue;
        }

        unsigned long groupId = strtoul (elemGroupId->name, NULL, 0);
        if ((groupId > 0) && (groupId < 64)) {
            Datamodel_SetDeviceGroupMembershipAddGroup(deviceId, (size_t)groupId);
        }
    }

cleanup_request_request_ltmode:
    {
        /* create long upstream event (notifies selectB SM to proceed) */
        LongUpstreamEvt *filteredUpEvt = Q_NEW(LongUpstreamEvt, LONG_UPSTREAM_SIG);
        filteredUpEvt->source = deviceId;
        filteredUpEvt->mainqualifier = 0;
        filteredUpEvt->subqualifier = 0;
        filteredUpEvt->data = 0;
        filteredUpEvt->quality = 62;
        filteredUpEvt->flags = 0;
        /* post event */
        QF_publish((QEvent*)filteredUpEvt);
        vdsm_report(LOG_DEBUG, "request groupMembership: published event for "
                          "device %d\n", deviceId);
        free(rc);
    }
}

static void prop_process_request_sensor_val(Vdcapi__VdcResponseGetProperty *msg,
                                            request_cache_t *rc)
{
    uint8_t sensorIndex = rc->channel;
    uint16_t deviceId = rc->result;
    uint16_t retVal = 0;

    /* binary input states */
    if (sensorIndex == 0x20) {
        Vdcapi__PropertyElement* elemBinaryInputStates = NULL;
        if (proputil_get_property_find_element_by_name(msg, "binaryInputStates",
                                                       &elemBinaryInputStates) != 0) {
            vdsm_report(LOG_INFO, "request sensor: no property binaryInputStates\n");
            goto cleanup_request_sensor_val;
        }

        if (elemBinaryInputStates->n_elements == 0) {
            vdsm_report(LOG_INFO, "request sensor: no binary inputs\n");
            goto cleanup_request_sensor_val;
        }

        size_t i;
        for (i = 0; i < elemBinaryInputStates->n_elements; ++i) {
            if (!elemBinaryInputStates->elements[i]->name) {
                vdsm_report(LOG_WARNING, "request sensor %d: invalid name element\n",
                       i);
                continue;
            }

            unsigned long index;
            index = strtoul (elemBinaryInputStates->elements[i]->name, NULL, 0);
            vdsm_report(LOG_DEBUG, "request sensor %d: converted index %d\n",
                   i, index);

            bool value;
            if (get_property_element_name_bool(elemBinaryInputStates->elements[i],
                                               "value", &value) != 0) {
                vdsm_report(LOG_WARNING, "request sensor %d: bool value not found "
                            "in value (binary input state is not known yet)\n", i);
                continue;
            }
            if (value) {
                retVal |= (1 << index); // Check if shift with index needed.
            }
        }

        /* create long upstream event */
        LongUpstreamEvt *filteredUpEvt = Q_NEW(LongUpstreamEvt,
                                               LONG_UPSTREAM_SIG);
        filteredUpEvt->source = deviceId;
        filteredUpEvt->mainqualifier = UpstreamMainqualifierSensor;
        filteredUpEvt->subqualifier = UpstreamSubqualifierBinaryInputStates;
        filteredUpEvt->data = SWAP_BYTES(retVal);
        filteredUpEvt->quality = 62;
        filteredUpEvt->flags = 0;
        /* post event */
        QF_publish((QEvent*)filteredUpEvt);
        vdsm_report(LOG_DEBUG, "request sensor: published long upstream event from "
                          "device %d (value %d)\n", deviceId, retVal);
    } else {
        Vdcapi__PropertyElement* elemSensorStates = NULL;
        Vdcapi__PropertyElement* elemSensor = NULL;
        if (proputil_get_property_find_element_by_name(msg, "sensorStates",
                                                       &elemSensorStates) != 0) {
            vdsm_report(LOG_INFO, "request sensor: no property sensorStates\n");
            goto cleanup_request_sensor_val;
        }

        if (proputil_find_property_element_by_index(elemSensorStates, sensorIndex, &elemSensor) != 0) {
            vdsm_report(LOG_WARNING, "request sensor: index %d not found in "
                                "sensorStates\n", sensorIndex);
            goto cleanup_request_sensor_val;
        }

        double value;
        if (get_property_element_name_double(elemSensor, "value", &value) != 0) {
            vdsm_report(LOG_WARNING, "request sensor: no property value\n");
            goto cleanup_request_sensor_val;
        }

        send_sensor_value(deviceId, sensorIndex, value);
    }

cleanup_request_sensor_val:
    free(rc);
}

uint8_t prop_process_push_property(Vdcapi__Message *msg)
{
    if (!msg->vdc_send_push_property) {
        vdsm_report(LOG_WARNING, "handle pushProperty: missing property message\n");
        return 1;
    }

    if (!msg->vdc_send_push_property->dsuid) {
        vdsm_report(LOG_WARNING, "handle pushProperty: missing dSUID\n");
        return 1;
    }

    if (msg->vdc_send_push_property->n_properties == 0) {
        vdsm_report(LOG_WARNING, "handle pushProperty: missing properties\n");
        return 1;
    }

    char dsuid_upper[DSUID_STR_LEN];
    strncpy(dsuid_upper, msg->vdc_send_push_property->dsuid, DSUID_STR_LEN);
    convert_string_to_upper(dsuid_upper,  DSUID_STR_LEN);

    uint16_t deviceId = Datamodel_GetDeviceIdFromDsuid(dsuid_upper);

    vdsm_report(LOG_DEBUG, "handle pushProperty from dsuid %s deviceId %d\n",
           dsuid_upper, deviceId);

    Vdcapi__PropertyElement *elemBinaryInputStates = NULL;
    Vdcapi__PropertyElement *elemButtonInputStates = NULL;
    Vdcapi__PropertyElement *elemSensorStates = NULL;
    Vdcapi__PropertyElement *elemId = NULL;

    if (proputil_send_push_property_find_element_by_name(msg->vdc_send_push_property,
                                                         "buttonInputStates",
                                                         &elemButtonInputStates) == 0) {
        uint8_t ltmode = Datamodel_GetDeviceLtmode(deviceId);
        uint8_t key = UCHAR_MAX;

        size_t i;
        for (i = 0; i < elemButtonInputStates->n_elements; ++i) {
            elemId = elemButtonInputStates->elements[i];

            if (!elemId->name) {
                vdsm_report(LOG_WARNING, "handle pushProperty button %d: invalid "
                                    "element name\n", i);
                continue;
            }

            unsigned long index = strtoul(elemId->name, NULL, 0);

            uint64_t  clickType = 0;
            if (get_property_element_name_uint(elemId,
                                              "clickType",
                                              &clickType) != 0) {
                vdsm_report(LOG_WARNING, "handle pushProperty button %d: no "
                                    "property clickType\n", i);
                continue;
            }

            // index, clickType
            // setup button key according to ltmode register
            if (ltmode == 0) {
                // normal button
                key = KEY_1WAY;
            } else if ((ltmode >= 5) && (ltmode <= 8)) {
                // 2-way down
                key = KEY_2WAY_A;
            } else if ((ltmode >= 9) && (ltmode <= 12)) {
                // 2-way up
                key = KEY_2WAY_B;
            } else if (ltmode == 13) {
                // 2-way with single dsuid

                // Verify MRY
                key = KEY_2WAY_A + (index&0xff);
            } else {
                vdsm_report(LOG_WARNING, "handle pushProperty button %d: ltmode %d "
                                    "not handled\n", i, ltmode);
            }

            /* allocate short upstream event */
            ShortUpstreamEvt * e = Q_NEW(ShortUpstreamEvt, SHORT_UPSTREAM_SIG);
            e->source = deviceId;
            e->key = key;
            e->click = clickType;
            e->sensor = 0;
            e->resend = 0;
            e->quality = 62;
            e->flags = 0;
            e->circuit = 0;
            /* publish short upstream event */
            QF_publish((QEvent*)e);
            vdsm_report(LOG_DEBUG, "handle pushProperty button %d: published short "
                "upstream event from device %d (click %d, key %d, index %d)\n",
                i, deviceId, clickType, key, index);
        }

    } else if (proputil_send_push_property_find_element_by_name(msg->vdc_send_push_property,
                                                                "sensorStates",
                                                                &elemSensorStates) == 0) {
        // Verify MRY
        // check usage of index
        size_t i;
        for (i = 0; i < elemSensorStates->n_elements; ++i) {
            elemId = elemSensorStates->elements[i];

            if (!elemId->name) {
                vdsm_report(LOG_WARNING, "handle pushProperty sensor %d: invalid "
                                    "element name\n", i);
                continue;
            }

            unsigned long index = strtoul (elemId->name, NULL, 0);

            double  value = 0;
            if (get_property_element_name_double(elemId,
                                              "value",
                                              &value) != 0) {
                vdsm_report(LOG_WARNING, "handle pushProperty sensor %d: "
                                    "no property value\n", i);
                continue;
            }
            send_sensor_value(deviceId, index, value);
        }

    } else if (proputil_send_push_property_find_element_by_name(msg->vdc_send_push_property,
                                                                "binaryInputStates",
                                                                &elemBinaryInputStates) == 0) {
        size_t i;
        for (i = 0; i < elemBinaryInputStates->n_elements; ++i) {
            elemId = elemBinaryInputStates->elements[i];

            if (!elemId->name) {
                vdsm_report(LOG_WARNING, "handle pushProperty binary %d: invalid "
                                    "element name\n", i);
                continue;
            }

            unsigned long index = strtoul (elemId->name, NULL, 0);

            bool  value = 0;
            if (get_property_element_name_bool(elemId,
                                               "value",
                                               &value) != 0) {
                vdsm_report(LOG_WARNING, "handle pushProperty binary %d: "
                                    "no property value\n", i);
                continue;
            }
            uint8_t status = value ? 1 : 0;

            /* allocate short upstream event */
            ShortUpstreamEvt * e = Q_NEW(ShortUpstreamEvt, SHORT_UPSTREAM_SIG);
            e->source = deviceId;
            e->eventType = Sensor_BinaryInputEvent;
            e->data = ((index&0xff) << 1) | (status & 0x01);
            e->sensor = 1;
            e->resend = 0;
            e->quality = 62;
            e->flags = 0;
            e->circuit = 0;
            /* publish short upstream event */
            QF_publish((QEvent*)e);
            vdsm_report(LOG_DEBUG, "handle pushProperty binary %d: published "
                   "short upstream event from device %d (status %d)\n",
                   i, deviceId, status);
        }

    } else {
        vdsm_report(LOG_WARNING, "handle pushProperty: property name not handled\n");
    }
    return 0;
}

uint8_t prop_process_identify(Vdcapi__Message *msg)
{
    if (!msg->vdc_send_identify->dsuid) {
        vdsm_report(LOG_WARNING, "handle identify: missing dSUID\n");
        return 1;
    }

    uint16_t deviceId = Datamodel_GetDeviceIdFromDsuid(
                                        msg->vdc_send_identify->dsuid);

    /* allocate short upstream event */
    ShortUpstreamEvt * e = Q_NEW(ShortUpstreamEvt, SHORT_UPSTREAM_SIG);
    e->source = deviceId;
    e->key = KEY_LOCAL;
    e->click = CT_L1;
    e->sensor = 0;
    e->resend = 0;
    e->quality = 62;
    e->flags = 0;
    /* publish short upstream event */
    QF_publish((QEvent*)e);
    vdsm_report(LOG_DEBUG, "handle identify: published short upstream event "
           "from device %d (click %d, key %d)\n",
           deviceId, e->click, e->key);
    return 0;
}

void prop_process_properties(vdc_client_t *vdc_client,
                             Vdcapi__VdcResponseGetProperty *msg,
                             Vdcapi__Message *cached_msg,
                             request_cache_t *rc)
{
    if (!cached_msg->vdsm_request_get_property) {
        vdsm_report(LOG_ERR, "process known properties: cached message type "
               "mismatch\n");
        free(rc);
        return;
    }

    Vdcapi__VdsmRequestGetProperty *request = cached_msg->vdsm_request_get_property;

    switch (rc->context) {
    case REQUEST_CONTEXT_GET_PROPERTY_FID:
        prop_process_request_fid(vdc_client, msg, request, rc);
        break;

    case REQUEST_CONTEXT_GET_PROPERTY_MODE:
        prop_process_request_mode(vdc_client, msg, request, rc);
        break;

    case REQUEST_CONTEXT_GET_PROPERTY_LTMODE:
        prop_process_request_ltmode(msg, request, rc);
        break;

    case REQUEST_CONTEXT_GET_PROPERTY_PBCHANNEL:
        prop_process_request_pbchannel(msg, request, rc);
        break;

    case REQUEST_CONTEXT_GET_PROPERTY_PBGROUP:
        prop_process_request_pbgroup(msg, request, rc);
        break;

    case REQUEST_CONTEXT_GET_PROPERTY_OUTPUT_VALUE:
        prop_process_request_outval(msg, request, rc);
        break;

    case REQUEST_CONTEXT_GET_PROPERTY_SCENES:
        prop_process_request_scenes(msg, request, rc);
        break;

    case REQUEST_CONTEXT_GET_PROPERTY_NAME:
        prop_process_request_name(msg, request);
        free(rc);
        break;

    case REQUEST_CONTEXT_GET_PROPERTY_SENSORS:
        prop_process_request_sensors(msg, request);
        free(rc);
        break;

    case REQUEST_CONTEXT_GET_PROPERTY_BINARY_INPUTS:
        prop_process_request_bin_inputs(vdc_client, msg, request, rc);
        break;

    case REQUEST_CONTEXT_GET_PROPERTY_GROUP_MEMBERSHIP:
        prop_process_request_group_membership(msg, rc);
        break;

    case REQUEST_CONTEXT_GET_PROPERTY_SENSOR_VALUE:
        prop_process_request_sensor_val(msg, rc);
        break;

    case REQUEST_CONTEXT_GET_PROPERTY_OPC_VALUES:
        prop_process_request_opc_table(vdc_client, msg, request, rc);
        break;

    case REQUEST_CONTEXT_GET_PROPERTY_OPC_SCENE_CONFIG:
        prop_process_request_opc_scene_config(msg, request, rc);
        break;

    case REQUEST_CONTEXT_GET_PROPERTY_OPC_SCENES_VALUE:
        prop_process_request_opc_scene_value(msg, request, rc);
        break;

    case REQUEST_CONTEXT_GET_VDC_NAME:
        prop_process_request_vdc_name(msg, request);
        free(rc);
        break;

    default:
        vdsm_report(LOG_ERR, "process properties: unknown property "
               "request context\n");
        free(rc);
        break;
    }
}
