/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>
             Michael Tross, digitalSTROM AG <michael.tross@digitalstrom.com>
             Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.
    

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#include "dsmapi.h"
#include "datamodel.h"
#include "dsm.h"
#include "selectA.h"
#include "selectB.h"
#include "heating.h"
#include "vdc-api-proto.h"
#include "mmi.h"
#include "bsp.h"
#include "registers.h"
#include "util.h"

Q_DEFINE_THIS_MODULE(dsmapi)

#define DSM_API_VERSION     0x0301      /**< dSM API version xx.yy: corresponds to MAJOR.MINOR of the dsm-api library */

#define DSM_API_REQUEST_TIMEOUT_TICKS   BSP_TICKS(750)

enum {
    MAX_DSM_API_NAME_LEN        = (MAX_NAME_LEN + 1),
};

typedef struct {
    QEvent super;
    QActive * requester;
    QEvent * msgEvt;
} DsmApiSendRequestEvt;

typedef struct {
    uint8_t operation;
    union {
        uint8_t bank;
        uint8_t channel;
    };
    uint16_t value;
    union {
        uint8_t offset;
        uint8_t scene;
    };
    bool is16bit;
} DsmApiDeviceParameterInfo;

typedef struct {
    bool upstreamEnabled;
    bool downstreamEnabled;
} DsmApiCommunicationLog;

typedef struct {
    QActive *requester;             /**< requester */
    QTimeEvt timer;                 /**< dSM API request/response timer */
    dsuid_t busAddress;             /**< cached dS485 bus destination address of request */
    uint8_t transactionId;          /**< cached transaction ID of request */
    uint8_t messageId;              /**< cached message ID of request */
    uint8_t modifierId;             /**< cached modifier ID of request */
} DsmApiPendingRequest;

/* Active object class -----------------------------------------------------*/
typedef struct {
    QActive super;                                  /**< active object base */
    QTimeEvt timer;                                 /**< timer */
    QEQueue deferredQueue;                          /**< deferred queue */
    dsuid_t requestAddress;                         /**< cached request address */
    dsuid_t destinationAddress;                     /**< cached destination address */
    uint8_t messageId;                              /**< cached message ID */
    uint8_t modifierId;                             /**< cached modifier ID */
    uint8_t transactionId;                          /**< cached transaction ID */
    uint16_t zoneIds[MAX_ZONE+1];                   /**< cached zone IDs */
    DsmApiDeviceParameterInfo deviceParameterInfo;  /**< data for the current device parameter request */
    DsmApiCommunicationLog communicationLog;        /**< state of the communication debug log */
    DsmApiPendingRequest pendingRequest;            /**< pending request info */
    bool longPowerDown;                             /**< true if last Power-Down was long */
    uint8_t tableIndex;                             /**< cache table index of binary input/sensor table */
    bool sendResponse;                              /**< true if the request expecting a response */
} DsmApi;

/* Local objects -----------------------------------------------------------*/
static DsmApi l_dsmApi;   /* the single instance of the dSM API active object */
static QEvent const * l_deferredQueueSto[16];
static const QEvent l_sysTermEvt = { SYSTEM_TERM_SIG, 0 };

static const dsuid_t l_broadcastDsuid = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };

static QState DsmApi_initial(DsmApi *me, QEvent const *e);

static QState DsmApi_top(DsmApi *me, QEvent const *e);

static QState DsmApi_idle(DsmApi *me, QEvent const *e);

static QState DsmApi_busy(DsmApi *me, QEvent const *e);
static QState DsmApi_device_info(DsmApi *me, QEvent const *e);
static QState DsmApi_button_info(DsmApi *me, QEvent const *e);
static QState DsmApi_binary_input_count(DsmApi *me, QEvent const *e);
static QState DsmApi_binary_input_table(DsmApi *me, QEvent const *e);
static QState DsmApi_wait_sensor_count(DsmApi *me, QEvent const *e);
static QState DsmApi_wait_sensor_table_entry(DsmApi *me, QEvent const *e);
static QState DsmApi_device_state(DsmApi *me, QEvent const *e);
static QState DsmApi_zone_count(DsmApi *me, QEvent const *e);
static QState DsmApi_zone_device_count(DsmApi *me, QEvent const *e);
static QState DsmApi_zone_info(DsmApi *me, QEvent const *e);
static QState DsmApi_wait_zone_group_info(DsmApi *me, QEvent const *e);
static QState DsmApi_wait_zone_history(DsmApi *me, QEvent const *e);
static QState DsmApi_wait_scene_info(DsmApi *me, QEvent const *e);
static QState DsmApi_wait_dsm_info(DsmApi *me, QEvent const *e);
static QState DsmApi_wait_dsm_apartment_id(DsmApi *me, QEvent const *e);
static QState DsmApi_wait_dsm_temperature_limits(DsmApi *me, QEvent const *e);
static QState DsmApi_wait_dsm_flags(DsmApi *me, QEvent const *e);
static QState DsmApi_wait_dsm_upstream_frequency(DsmApi *me, QEvent const *e);
static QState DsmApi_wait_ping_intervals(DsmApi *me, QEvent const *e);
static QState DsmApi_wait_dsm_state(DsmApi *me, QEvent const *e);
static QState DsmApi_wait_current_temperature(DsmApi *me, QEvent const *e);
static QState DsmApi_wait_max_temperature(DsmApi *me, QEvent const *e);
static QState DsmApi_wait_hash(DsmApi *me, QEvent const *e);
static QState DsmApi_wait_room_states(DsmApi *me, QEvent const *e);
static QState DsmApi_wait_flash_read(DsmApi *me, QEvent const *e);
static QState DsmApi_start_update(DsmApi *me, QEvent const *e);
static QState DsmApi_system_reset(DsmApi *me, QEvent const *e);
static QState DsmApi_system_reset_and_update(DsmApi *me, QEvent const *e);
static QState DsmApi_stopped(DsmApi *me, QEvent const *e);
static QState DsmApi_device_parameter(DsmApi *me, QEvent const *e);
static QState DsmApi_remove_by_serial_number(DsmApi *me, QEvent const *e);
static QState DsmApi_device_output_channel(DsmApi *me, QEvent const *e);
static QState DsmApi_device_output_channel_wait_count(DsmApi *me, QEvent const *e);
static QState DsmApi_device_output_channel_wait_table_entry(DsmApi *me, QEvent const *e);
static QState DsmApi_wait_vdc_user_protobuf_message_request(DsmApi *me, QEvent const *e);

static void DsmApi_SendOPCRequest(DsmApi *me, uint16_t deviceId, uint8_t modifier, uint8_t channelID, uint16_t value, uint8_t sceneID, uint8_t size);

static void DsmApi_ConvertDeviceSerialToDsuid(uint32_t deviceSerial, uint8_t *dsuid);
static void DsmApi_ConvertDsuidToDeviceSerial(uint8_t *dsuid, uint32_t *deviceSerial);
static uint8_t DsmApi_GetLocalZoneId(DsmApi *me, uint16_t zoneId);

static DS485MessageEvt * DsmApi_CreateResponse(DsmApi *me, uint8_t error, uint8_t dataLength);
static DS485NetlibPacketMessageEvt * DsmApi_CreateLongResponse(DsmApi *me, uint8_t error, uint16_t dataLength);
static DS485MessageEvt * DsmApi_CreateRequestFromDsid(dsuid_t const *destination, uint16_t sourceId, uint8_t transactionId, uint8_t msgid, uint8_t modifier, uint8_t dataLength);
static DS485MessageEvt * DsmApi_CreateEvent(uint8_t destination, uint16_t source, uint8_t transactionId, uint8_t msgid, uint8_t modifier, uint8_t dataLength);
static DS485MessageEvt * DsmApi_CreateDsmDoneEvent(uint16_t source, uint8_t modifier, uint8_t error, uint8_t dataLength);
static DS485MessageEvt * DsmApi_CreateRequestFromVdcDsuid(dsuid_t const *destination, dsuid_t const *vdcDsuid, uint8_t transactionId, uint8_t msgid, uint8_t modifier, uint8_t dataLength);

/*..........................................................................*/

#define DsmApi_GetTransactionId(msg_)   ((msg_)->message[DSMAPI_TRANSACTIONID_OFFSET])
#define DsmApi_GetMessageId(msg_)       ((msg_)->message[DSMAPI_MSGID_OFFSET])
#define DsmApi_GetModifierId(msg_)      ((msg_)->message[DSMAPI_MODIFIER_OFFSET])

/*..........................................................................*/

#define DsmApi_GetReqData8(msg_, offset_) \
    (msg_)->message[DSMAPI_REQ_DATA_OFFSET+(offset_)]

#define DsmApi_GetReqData16(msg_, offset_) \
    ((uint16_t)((msg_)->message[DSMAPI_REQ_DATA_OFFSET+(offset_)+0]) << 8) | \
    ((uint16_t)((msg_)->message[DSMAPI_REQ_DATA_OFFSET+(offset_)+1]) << 0)

#define DsmApi_GetReqData32(msg_, offset_) \
    ((uint32_t)((msg_)->message[DSMAPI_REQ_DATA_OFFSET+(offset_)+0]) << 24) | \
    ((uint32_t)((msg_)->message[DSMAPI_REQ_DATA_OFFSET+(offset_)+1]) << 16) | \
    ((uint32_t)((msg_)->message[DSMAPI_REQ_DATA_OFFSET+(offset_)+2]) <<  8) | \
    ((uint32_t)((msg_)->message[DSMAPI_REQ_DATA_OFFSET+(offset_)+3]) <<  0)

#define DsmApi_GetReqArray8(msg_, offset_, array_, size_) \
    memcpy((array_), &((msg_)->message[DSMAPI_REQ_DATA_OFFSET+(offset_)]), size_)

#define DsmApi_GetReqArray16(msg_, offset_, array_, size_) do {\
    uint16_t index_; \
    for (index_ = 0; index_ < size_; ++index_) { \
        array_[index_] = DsmApi_GetReqData16(msg_, offset_+2*index_); \
    } \
} while (0)

/*..........................................................................*/

#define DsmApi_SetResData8(msg_, offset_, data8_) do { \
    (msg_)->message[DSMAPI_RES_DATA_OFFSET+(offset_)] = (data8_); \
} while (0)

#define DsmApi_SetResData16(msg_, offset_, data16_) do { \
    (msg_)->message[DSMAPI_RES_DATA_OFFSET+(offset_)+0] = (uint8_t)((data16_) >> 8); \
    (msg_)->message[DSMAPI_RES_DATA_OFFSET+(offset_)+1] = (uint8_t)((data16_) >> 0); \
} while (0)

#define DsmApi_SetResData32(msg_, offset_, data32_) do { \
    (msg_)->message[DSMAPI_RES_DATA_OFFSET+(offset_)+0] = (uint8_t)((data32_) >> 24); \
    (msg_)->message[DSMAPI_RES_DATA_OFFSET+(offset_)+1] = (uint8_t)((data32_) >> 16); \
    (msg_)->message[DSMAPI_RES_DATA_OFFSET+(offset_)+2] = (uint8_t)((data32_) >>  8); \
    (msg_)->message[DSMAPI_RES_DATA_OFFSET+(offset_)+3] = (uint8_t)((data32_) >>  0); \
} while (0)

#define DsmApi_SetResArray8(msg_, offset_, array_, size_) do { \
    memcpy(&((msg_)->message[DSMAPI_RES_DATA_OFFSET+(offset_)]), (array_), size_); \
} while (0)

#define DsmApi_SetResArray16(msg_, offset_, array_, size_) do { \
    uint16_t index_; \
    for (index_ = 0; index_ < size_; ++index_) { \
        DsmApi_SetResData16(msg_, offset_+2*index_, array_[index_]); \
    } \
} while (0)

#define DsmApi_SetResName(msg_, offset_, name_) do { \
    memcpy(&((msg_)->message[DSMAPI_RES_DATA_OFFSET+(offset_)]), (name_), MAX_NAME_LEN); \
    (msg_)->message[DSMAPI_RES_DATA_OFFSET+(offset_)+MAX_NAME_LEN] = 0; \
} while (0)

/*..........................................................................*/

#define DsmApi_SetReqData8(msg_, offset_, data8_) do { \
    (msg_)->message[DSMAPI_REQ_DATA_OFFSET+(offset_)] = (data8_); \
} while (0)

#define DsmApi_SetReqData16(msg_, offset_, data16_) do { \
    (msg_)->message[DSMAPI_REQ_DATA_OFFSET+(offset_)+0] = (uint8_t)((data16_) >> 8); \
    (msg_)->message[DSMAPI_REQ_DATA_OFFSET+(offset_)+1] = (uint8_t)((data16_) >> 0); \
} while (0)

#define DsmApi_SetReqData32(msg_, offset_, data32_) do { \
    (msg_)->message[DSMAPI_REQ_DATA_OFFSET+(offset_)+0] = (uint8_t)((data32_) >> 24); \
    (msg_)->message[DSMAPI_REQ_DATA_OFFSET+(offset_)+1] = (uint8_t)((data32_) >> 16); \
    (msg_)->message[DSMAPI_REQ_DATA_OFFSET+(offset_)+2] = (uint8_t)((data32_) >>  8); \
    (msg_)->message[DSMAPI_REQ_DATA_OFFSET+(offset_)+3] = (uint8_t)((data32_) >>  0); \
} while (0)

#define DsmApi_SetReqArray8(msg_, offset_, array_, size_) do { \
    memcpy(&((msg_)->message[DSMAPI_REQ_DATA_OFFSET+(offset_)]), (array_), size_); \
} while (0)

/*..........................................................................*/

#define DsmApi_SetEvtData8(msg_, offset_, data8_) do { \
    (msg_)->message[DSMAPI_EVT_DATA_OFFSET+(offset_)] = (data8_); \
} while (0)

#define DsmApi_SetEvtData16(msg_, offset_, data16_) do { \
    (msg_)->message[DSMAPI_EVT_DATA_OFFSET+(offset_)+0] = (uint8_t)((data16_) >> 8); \
    (msg_)->message[DSMAPI_EVT_DATA_OFFSET+(offset_)+1] = (uint8_t)((data16_) >> 0); \
} while (0)

#define DsmApi_SetEvtData32(msg_, offset_, data32_) do { \
    (msg_)->message[DSMAPI_EVT_DATA_OFFSET+(offset_)+0] = (uint8_t)((data32_) >> 24); \
    (msg_)->message[DSMAPI_EVT_DATA_OFFSET+(offset_)+1] = (uint8_t)((data32_) >> 16); \
    (msg_)->message[DSMAPI_EVT_DATA_OFFSET+(offset_)+2] = (uint8_t)((data32_) >>  8); \
    (msg_)->message[DSMAPI_EVT_DATA_OFFSET+(offset_)+3] = (uint8_t)((data32_) >>  0); \
} while (0)

#define DsmApi_SetEvtArray8(msg_, offset_, array_, size_) do { \
    memcpy(&((msg_)->message[DSMAPI_EVT_DATA_OFFSET+(offset_)]), (array_), size_); \
} while (0)

#define DsmApi_GetDataLength(evt_)  (DS485_GetPayloadLength(evt_) - 3)

#define DsmApi_SetLongResArray8(msg_, offset_, array_, size_) do { \
    memcpy(&((msg_)->message[DSMAPI_LONG_RES_DATA_OFFSET+(offset_)]), (array_), size_); \
} while (0)

#define DsmApi_GetNetlibPacketDataLength(evt_)  (DS485_GetNetlibPacketPayloadLength(evt_) - 3)

/* Global-scope objects ----------------------------------------------------*/
QActive * const AO_DsmApi = (QActive *)&l_dsmApi;      /* "opaque" AO pointer */

/*..........................................................................*/
void DsmApi_ctor(bool longPowerDown)
{
    uint8_t i;
    /* set initial state */
    DsmApi *me = &l_dsmApi;
    QActive_ctor(&me->super, (QStateHandler)&DsmApi_initial);
    /* initialize timer */
    QTimeEvt_ctor(&me->timer, DSMAPI_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->pendingRequest.timer, DSMAPI_REQUEST_TIMEOUT_SIG);
    /* initialize deferred queue */
    QEQueue_init(&me->deferredQueue, l_deferredQueueSto, Q_DIM(l_deferredQueueSto));
    /* initialize zone IDs */
    for (i = 0; i < Q_DIM(me->zoneIds); ++i) {
        me->zoneIds[i] = DATAMODEL_INVALID_ZONE_ID;
    }
    /* initialize extended state variables */
    me->longPowerDown = longPowerDown;
    me->pendingRequest.busAddress = DSUID_BROADCAST;
}

/*..........................................................................*/
void DsmApi_SendZoneGroupActionCallSceneRequest(uint8_t localZoneId, uint8_t groupId, Scene scene, uint16_t originDevice, bool force)
{
    DS485MessageEvt * txMsgEvt;
    uint8_t modifier = (force ? ZONE_GROUP_ACTION_REQUEST_ACTION_FORCE_CALL_SCENE : ZONE_GROUP_ACTION_REQUEST_ACTION_CALL_SCENE);
    txMsgEvt = DsmApi_CreateRequestFromDsid(&l_broadcastDsuid, originDevice, DSM_TRANSACTION_ID, ZONE_GROUP_ACTION_REQUEST, modifier, 6);
    /* zone ID */
    DsmApi_SetReqData16(txMsgEvt, 0, l_dsmApi.zoneIds[localZoneId]);
    /* group */
    DsmApi_SetReqData8(txMsgEvt, 2, groupId);
    /* origin device */
    DsmApi_SetReqData16(txMsgEvt, 3, originDevice);
    /* scene */
    DsmApi_SetReqData8(txMsgEvt, 5, scene);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendDeviceActionCallSceneRequest(uint8_t destination, uint16_t deviceId, Scene scene)
{
}

/*..........................................................................*/
void DsmApi_SendZoneGroupActionUndoSceneNumberRequest(uint8_t localZoneId, uint8_t groupId, Scene scene, uint16_t originDevice)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateRequestFromDsid(&l_broadcastDsuid, originDevice, DSM_TRANSACTION_ID, ZONE_GROUP_ACTION_REQUEST, ZONE_GROUP_ACTION_REQUEST_ACTION_UNDO_SCENE_NUMBER, 6);
    /* zone ID */
    DsmApi_SetReqData16(txMsgEvt, 0, l_dsmApi.zoneIds[localZoneId]);
    /* group */
    DsmApi_SetReqData8(txMsgEvt, 2, groupId);
    /* origin device */
    DsmApi_SetReqData16(txMsgEvt, 3, originDevice);
    /* scene */
    DsmApi_SetReqData8(txMsgEvt, 5, scene);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendZoneGroupActionSetSceneLocalPrioRequest(uint8_t localZoneId, uint8_t groupId, Scene scene, uint16_t originDevice)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateRequestFromDsid(&l_broadcastDsuid, originDevice, DSM_TRANSACTION_ID, ZONE_GROUP_ACTION_REQUEST, ZONE_GROUP_ACTION_REQUEST_ACTION_SET_SCENE_LOCAL_PRIORITY, 6);
    /* zone ID */
    DsmApi_SetReqData16(txMsgEvt, 0, l_dsmApi.zoneIds[localZoneId]);
    /* group */
    DsmApi_SetReqData8(txMsgEvt, 2, groupId);
    /* origin device */
    DsmApi_SetReqData16(txMsgEvt, 3, originDevice);
    /* scene */
    DsmApi_SetReqData8(txMsgEvt, 5, scene);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendZoneGroupActionCallSceneMinRequest(uint8_t localZoneId, uint8_t groupId, Scene scene, uint16_t originDevice)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateRequestFromDsid(&l_broadcastDsuid, originDevice, DSM_TRANSACTION_ID, ZONE_GROUP_ACTION_REQUEST, ZONE_GROUP_ACTION_REQUEST_ACTION_CALL_SCENE_MIN, 6);
    /* zone ID */
    DsmApi_SetReqData16(txMsgEvt, 0, l_dsmApi.zoneIds[localZoneId]);
    /* group */
    DsmApi_SetReqData8(txMsgEvt, 2, groupId);
    /* origin device */
    DsmApi_SetReqData16(txMsgEvt, 3, originDevice);
    /* scene */
    DsmApi_SetReqData8(txMsgEvt, 5, scene);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendZoneGroupActionRegisterSceneLocalPrioRequest(uint8_t localZoneId, uint8_t groupId, Scene scene, uint16_t originDevice)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateRequestFromDsid(&l_broadcastDsuid, originDevice, DSM_TRANSACTION_ID, ZONE_GROUP_ACTION_REQUEST, ZONE_GROUP_ACTION_REQUEST_ACTION_REGISTER_SCENE_LOCAL_PRIORITY, 6);
    /* zone ID */
    DsmApi_SetReqData16(txMsgEvt, 0, l_dsmApi.zoneIds[localZoneId]);
    /* group */
    DsmApi_SetReqData8(txMsgEvt, 2, groupId);
    /* origin device */
    DsmApi_SetReqData16(txMsgEvt, 3, originDevice);
    /* scene */
    DsmApi_SetReqData8(txMsgEvt, 5, scene);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendZoneGroupActionProgOn(uint8_t localZoneId, uint8_t groupId, uint16_t originDevice)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateRequestFromDsid(&l_broadcastDsuid, originDevice, DSM_TRANSACTION_ID, ZONE_GROUP_ACTION_REQUEST, ZONE_GROUP_ACTION_REQUEST_ACTION_PROGRAMMING_MODE_ON, 5);
    /* zone ID */
    DsmApi_SetReqData16(txMsgEvt, 0, l_dsmApi.zoneIds[localZoneId]);
    /* group */
    DsmApi_SetReqData8(txMsgEvt, 2, groupId);
    /* origin device */
    DsmApi_SetReqData16(txMsgEvt, 3, originDevice);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendDeviceActionProgOn(uint8_t destination, uint16_t deviceId)
{
}

/*..........................................................................*/
void DsmApi_SendZoneGroupActionProgOff(uint8_t localZoneId, uint8_t groupId, uint16_t originDevice)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateRequestFromDsid(&l_broadcastDsuid, originDevice, DSM_TRANSACTION_ID, ZONE_GROUP_ACTION_REQUEST, ZONE_GROUP_ACTION_REQUEST_ACTION_PROGRAMMING_MODE_OFF, 5);
    /* zone ID */
    DsmApi_SetReqData16(txMsgEvt, 0, l_dsmApi.zoneIds[localZoneId]);
    /* group */
    DsmApi_SetReqData8(txMsgEvt, 2, groupId);
    /* origin device */
    DsmApi_SetReqData16(txMsgEvt, 3, originDevice);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendZoneGroupActionSelectGroup(uint8_t localZoneId, uint8_t groupId, uint16_t originDevice)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateRequestFromDsid(&l_broadcastDsuid, originDevice, DSM_TRANSACTION_ID, ZONE_GROUP_ACTION_REQUEST, ZONE_GROUP_ACTION_REQUEST_ACTION_SELECT_GROUP, 5);
    /* zone ID */
    DsmApi_SetReqData16(txMsgEvt, 0, l_dsmApi.zoneIds[localZoneId]);
    /* group */
    DsmApi_SetReqData8(txMsgEvt, 2, groupId);
    /* origin device */
    DsmApi_SetReqData16(txMsgEvt, 3, originDevice);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendZoneGroupActionSaveScene(uint8_t localZoneId, uint8_t groupId, Scene scene, uint16_t originDevice)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateRequestFromDsid(&l_broadcastDsuid, originDevice, DSM_TRANSACTION_ID, ZONE_GROUP_ACTION_REQUEST, ZONE_GROUP_ACTION_REQUEST_ACTION_SAVE_SCENE, 6);
    /* zone ID */
    DsmApi_SetReqData16(txMsgEvt, 0, l_dsmApi.zoneIds[localZoneId]);
    /* group */
    DsmApi_SetReqData8(txMsgEvt, 2, groupId);
    /* origin device */
    DsmApi_SetReqData16(txMsgEvt, 3, originDevice);
    /* scene */
    DsmApi_SetReqData8(txMsgEvt, 5, scene);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendZoneGroupActionLocalStop(uint8_t localZoneId, uint8_t groupId, uint16_t originDevice)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateRequestFromDsid(&l_broadcastDsuid, originDevice, DSM_TRANSACTION_ID, ZONE_GROUP_ACTION_REQUEST, ZONE_GROUP_ACTION_REQUEST_ACTION_LOCAL_STOP, 5);
    /* zone ID */
    DsmApi_SetReqData16(txMsgEvt, 0, l_dsmApi.zoneIds[localZoneId]);
    /* group */
    DsmApi_SetReqData8(txMsgEvt, 2, groupId);
    /* origin device */
    DsmApi_SetReqData16(txMsgEvt, 3, originDevice);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendZoneGroupSensorPushFromVdc(dsuid_t const *vdcDsuid, dsuid_t const *sensorDsuid,
                                           uint16_t zoneId, uint8_t groupId,
                                           uint8_t sensorType, uint16_t sensorValue,
                                           bool sensorPrecision)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateRequestFromVdcDsuid(&l_broadcastDsuid, vdcDsuid, DSM_TRANSACTION_ID, ZONE_GROUP_SENSOR_PUSH, 0, 24);
    /* zone ID */
    DsmApi_SetReqData16(txMsgEvt, 0, zoneId);
    /* group */
    DsmApi_SetReqData8(txMsgEvt, 2, groupId);
    /* dsuid */
    DsmApi_SetReqArray8(txMsgEvt, 3, sensorDsuid->id, DSUID_SIZE);
    /* sensor type */
    DsmApi_SetReqData8(txMsgEvt, 20, sensorType);
    /* sensor value */
    DsmApi_SetReqData16(txMsgEvt, 21, sensorValue);
    /* sensor precision */
    DsmApi_SetReqData8(txMsgEvt, 23, sensorPrecision? 1 : 0);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendZoneGroupActionOPCDimRequest(uint8_t localZoneId, uint8_t groupId, OPCOperation operation, uint8_t channelId, uint16_t originDevice)
{
    ZoneGroupActionRequestModifiers_t mod = ZONE_GROUP_ACTION_REQUEST_ACTION_OPC_INC;
    switch (operation) {
    case OPC_INC:
        mod = ZONE_GROUP_ACTION_REQUEST_ACTION_OPC_INC;
        break;
    case OPC_DEC:
        mod = ZONE_GROUP_ACTION_REQUEST_ACTION_OPC_DEC;
        break;
    case OPC_STOP:
        mod = ZONE_GROUP_ACTION_REQUEST_ACTION_OPC_STOP;
        break;
    }

    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateRequestFromDsid(&l_broadcastDsuid, originDevice, DSM_TRANSACTION_ID, ZONE_GROUP_ACTION_REQUEST, mod, 6);

    /* zone ID */
    DsmApi_SetReqData16(txMsgEvt, 0, l_dsmApi.zoneIds[localZoneId]);
    /* group */
    DsmApi_SetReqData8(txMsgEvt, 2, groupId);
    /* origin device */
    DsmApi_SetReqData16(txMsgEvt, 3, originDevice);
    /* channel ID */
    DsmApi_SetReqData8(txMsgEvt, 5, channelId);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendDeviceActionOPCDimRequest(uint8_t destination, uint16_t deviceId, OPCOperation operation, uint8_t channelId)
{
    ZoneGroupActionRequestModifiers_t mod = ZONE_GROUP_ACTION_REQUEST_ACTION_OPC_INC;
    switch (operation) {
    case OPC_INC:
        mod = ZONE_GROUP_ACTION_REQUEST_ACTION_OPC_INC;
        break;
    case OPC_DEC:
        mod = ZONE_GROUP_ACTION_REQUEST_ACTION_OPC_DEC;
        break;
    case OPC_STOP:
        mod = ZONE_GROUP_ACTION_REQUEST_ACTION_OPC_STOP;
        break;
    }
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateRequestFromDsid(&l_broadcastDsuid, deviceId, DSM_TRANSACTION_ID, DEVICE_ACTION_REQUEST, mod, 3);
    /* device ID */
    DsmApi_SetReqData16(txMsgEvt, 0, deviceId);
    /* channel ID */
    DsmApi_SetReqData8(txMsgEvt, 2, channelId);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendDeviceAccessibilityEvent(uint16_t deviceId, uint16_t zoneId, uint16_t vendorId, uint16_t productId, uint16_t functionId, uint16_t version, dsuid_t deviceDsuid, bool on)
{
    DS485MessageEvt * txMsgEvt;
    uint8_t modifier = (on ? EVENT_DEVICE_ACCESSIBILITY_ON : EVENT_DEVICE_ACCESSIBILITY_OFF);
    txMsgEvt = DsmApi_CreateEvent(DS485_BROADCAST_ADDRESS, deviceId, DSM_TRANSACTION_ID, EVENT_DEVICE_ACCESSIBILITY, modifier, 29);
    /* device ID */
    DsmApi_SetEvtData16(txMsgEvt, 0, deviceId);
    /* zone ID */
    DsmApi_SetEvtData16(txMsgEvt, 2, zoneId);
    /* vendor ID */
    DsmApi_SetEvtData16(txMsgEvt, 4, vendorId);
    /* product ID */
    DsmApi_SetEvtData16(txMsgEvt, 6, productId);
    /* function ID */
    DsmApi_SetEvtData16(txMsgEvt, 8, functionId);
    /* version */
    DsmApi_SetEvtData16(txMsgEvt, 10, version);
    /* dsuid */
    DsmApi_SetResArray8(txMsgEvt, 12, &deviceDsuid, sizeof(deviceDsuid));
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendDevicePresenceEvent(uint16_t deviceId, bool present)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateEvent(DS485_BROADCAST_ADDRESS, deviceId, DSM_TRANSACTION_ID, EVENT_TEST_DEVICE_PRESENCE, 0, 3);
    /* device ID */
    DsmApi_SetEvtData16(txMsgEvt, 0, deviceId);
    /* device presence */
    DsmApi_SetEvtData8(txMsgEvt, 2, (uint8_t)present);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendDeviceStatisticsEvent(uint16_t deviceId, uint8_t paramOffset, uint16_t paramValue)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateEvent(DS485_BROADCAST_ADDRESS, deviceId, DSM_TRANSACTION_ID, EVENT_DEVICE_STATISTICS, 0, 5);
    /* device ID */
    DsmApi_SetEvtData16(txMsgEvt, 0, deviceId);
    /* parameter offset */
    DsmApi_SetEvtData8(txMsgEvt, 2, paramOffset);
    /* parameter value */
    DsmApi_SetEvtData16(txMsgEvt, 3, paramValue);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendDeviceActionEvent(uint16_t deviceId, uint8_t localZoneId, uint8_t groupId, uint8_t buttonNumber, uint8_t clickType)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateEvent(DS485_BROADCAST_ADDRESS, deviceId, DSM_TRANSACTION_ID, EVENT_DEVICE_ACTION, 0, 7);
    /* device ID */
    DsmApi_SetEvtData16(txMsgEvt, 0, deviceId);
    /* zone ID */
    DsmApi_SetEvtData16(txMsgEvt, 2, l_dsmApi.zoneIds[localZoneId]);
    /* group ID */
    DsmApi_SetEvtData8(txMsgEvt, 4, groupId);
    /* button number */
    DsmApi_SetEvtData8(txMsgEvt, 5, buttonNumber);
    /* click type */
    DsmApi_SetEvtData8(txMsgEvt, 6, clickType);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendDeviceLocalActionEvent(uint16_t deviceId, uint8_t localZoneId, uint8_t localOn)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateEvent(DS485_BROADCAST_ADDRESS, deviceId, DSM_TRANSACTION_ID, EVENT_DEVICE_LOCAL_ACTION, 0, 5);
    /* device ID */
    DsmApi_SetEvtData16(txMsgEvt, 0, deviceId);
    /* zone ID */
    DsmApi_SetEvtData16(txMsgEvt, 2, l_dsmApi.zoneIds[localZoneId]);
    /* local state */
    DsmApi_SetEvtData8(txMsgEvt, 4, localOn);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendDeviceParameterEvent(uint16_t deviceId, uint8_t paramBank, uint8_t paramOffset, uint8_t paramValue)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateEvent(DS485_BROADCAST_ADDRESS, deviceId, DSM_TRANSACTION_ID, EVENT_DEVICE_CONFIG, 0, 5);
    /* device ID */
    DsmApi_SetEvtData16(txMsgEvt, 0, deviceId);
    /* parameter bank */
    DsmApi_SetEvtData8(txMsgEvt, 2, paramBank);
    /* parameter offset */
    DsmApi_SetEvtData8(txMsgEvt, 3, paramOffset);
    /* parameter value */
    DsmApi_SetEvtData8(txMsgEvt, 4, paramValue);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendTestTransmissionQualityEvent(uint16_t deviceId, uint8_t downstreamQuality, uint16_t upstreamQuality)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateEvent(DS485_BROADCAST_ADDRESS, deviceId, DSM_TRANSACTION_ID, EVENT_TEST_TRANSMISSION_QUALITY, 0, 5);
    /* device ID */
    DsmApi_SetEvtData16(txMsgEvt, 0, deviceId);
    /* downstream quality */
    DsmApi_SetEvtData8(txMsgEvt, 2, downstreamQuality);
    /* upstream quality */
    DsmApi_SetEvtData16(txMsgEvt, 3, upstreamQuality);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendSensorListEvent(uint16_t deviceId, uint8_t sensorId, uint8_t sensorPresent, uint8_t sensorType, uint8_t lastSensor)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateEvent(DS485_BROADCAST_ADDRESS, deviceId, DSM_TRANSACTION_ID, EVENT_DEVICE_SENSOR, MOD_DEV_SENSOR_GET_TYPE, 6);
    /* device ID */
    DsmApi_SetEvtData16(txMsgEvt, 0, deviceId);
    /* sensor ID */
    DsmApi_SetEvtData8(txMsgEvt, 2, sensorId);
    /* sensor present flag */
    DsmApi_SetEvtData8(txMsgEvt, 3, sensorPresent);
    /* sensor type */
    DsmApi_SetEvtData8(txMsgEvt, 4, sensorType);
    /* last sensor flag */
    DsmApi_SetEvtData8(txMsgEvt, 5, lastSensor);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendSensorValueEvent(uint16_t deviceId, uint8_t sensorId, uint16_t sensorValue)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateEvent(DS485_BROADCAST_ADDRESS, deviceId, DSM_TRANSACTION_ID, EVENT_DEVICE_SENSOR, MOD_DEV_SENSOR_GET_VALUE, 5);
    /* device ID */
    DsmApi_SetEvtData16(txMsgEvt, 0, deviceId);
    /* sensor ID */
    DsmApi_SetEvtData8(txMsgEvt, 2, sensorId);
    /* sensor value */
    DsmApi_SetEvtData16(txMsgEvt, 3, sensorValue);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendSensorEvent(uint16_t deviceId, uint8_t eventIndex)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateEvent(DS485_BROADCAST_ADDRESS, deviceId, DSM_TRANSACTION_ID, EVENT_DEVICE_SENSOR, MOD_DEV_SENSOR_EVENT, 3);
    /* device ID */
    DsmApi_SetEvtData16(txMsgEvt, 0, deviceId);
    /* event index */
    DsmApi_SetEvtData8(txMsgEvt, 2, eventIndex);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendBinaryInputEvent(uint16_t deviceId, uint8_t binaryInputIndex, uint8_t binaryInputType, uint8_t binaryInputState)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateEvent(DS485_BROADCAST_ADDRESS, deviceId, DSM_TRANSACTION_ID, EVENT_DEVICE_SENSOR, MOD_DEV_BINARY_INPUT_EVENT, 5);
    /* device ID */
    DsmApi_SetEvtData16(txMsgEvt, 0, deviceId);
    /* event index */
    DsmApi_SetEvtData8(txMsgEvt, 2, binaryInputIndex);
    /* event index */
    DsmApi_SetEvtData8(txMsgEvt, 3, binaryInputType);
    /* sensor state */
    DsmApi_SetEvtData8(txMsgEvt, 4, binaryInputState);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendUDIResponseEvent(uint16_t deviceId, uint8_t data)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateEvent(DS485_BROADCAST_ADDRESS, deviceId, DSM_TRANSACTION_ID, EVENT_DSLINK, EVENT_DSLINK_RESPONSE, 3);
    /* device ID */
    DsmApi_SetEvtData16(txMsgEvt, 0, deviceId);
    /* UDI data */
    DsmApi_SetEvtData8(txMsgEvt, 2, data);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendUDIInterruptEvent(uint16_t deviceId, bool highPriority)
{
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DsmApi_CreateEvent(DS485_BROADCAST_ADDRESS, deviceId, DSM_TRANSACTION_ID, EVENT_DSLINK, EVENT_DSLINK_INTERRUPT, 3);
    /* device ID */
    DsmApi_SetEvtData16(txMsgEvt, 0, deviceId);
    /* UDI interrupt priority */
    DsmApi_SetEvtData8(txMsgEvt, 2, (highPriority) ? 1 : 0);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendDeviceGiveUp(uint32_t serialNumber, uint8_t circuitId)
{
}

/*..........................................................................*/
void DsmApi_SendShortUpstreamEvent(ShortUpstreamEvt const *e, bool crosstalk)
{
    DsmApi *me = &l_dsmApi;
    if (me->communicationLog.upstreamEnabled) {
        DS485MessageEvt * txMsgEvt;
        /* create event */
        txMsgEvt = DsmApi_CreateEvent(DS485_BROADCAST_ADDRESS, e->source, DSM_TRANSACTION_ID, EVENT_COMMUNICATION_LOG, EVENT_COMMUNICATION_LOG_UPSTREAM_SHORT, 10);
        DsmApi_SetEvtData16(txMsgEvt, 0, e->source);
        DsmApi_SetEvtData8(txMsgEvt, 2, e->circuit);
        DsmApi_SetEvtData8(txMsgEvt, 3, e->resend);
        DsmApi_SetEvtData8(txMsgEvt, 4, e->sensor);
        DsmApi_SetEvtData8(txMsgEvt, 5, e->key);
        DsmApi_SetEvtData8(txMsgEvt, 6, e->click);
        DsmApi_SetEvtData8(txMsgEvt, 7, e->quality);
        DsmApi_SetEvtData8(txMsgEvt, 8, e->flags);
        DsmApi_SetEvtData8(txMsgEvt, 9, (crosstalk) ? 1 : 0);
        /* send event */
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
    }
}

/*..........................................................................*/
void DsmApi_SendLongUpstreamEvent(LongUpstreamEvt const *e, bool crosstalk)
{
    DsmApi *me = &l_dsmApi;
    if (me->communicationLog.upstreamEnabled) {
        DS485MessageEvt * txMsgEvt;
        /* create event */
        txMsgEvt = DsmApi_CreateEvent(DS485_BROADCAST_ADDRESS, e->source, DSM_TRANSACTION_ID, EVENT_COMMUNICATION_LOG, EVENT_COMMUNICATION_LOG_UPSTREAM_LONG, 12);
        DsmApi_SetEvtData16(txMsgEvt, 0, e->source);
        DsmApi_SetEvtData8(txMsgEvt, 2, e->circuit);
        DsmApi_SetEvtData8(txMsgEvt, 3, e->mainqualifier);
        DsmApi_SetEvtData8(txMsgEvt, 4, e->subqualifier);
        DsmApi_SetEvtData32(txMsgEvt, 5, e->data);
        DsmApi_SetEvtData8(txMsgEvt, 9, e->quality);
        DsmApi_SetEvtData8(txMsgEvt, 10, e->flags);
        DsmApi_SetEvtData8(txMsgEvt, 11, (crosstalk) ? 1 : 0);
        /* send event */
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
    }
}

/*..........................................................................*/
void DsmApi_SendDeviceModelChangedEvent(uint16_t deviceAddress)
{
    DS485MessageEvt * txMsgEvt;
    /* create event */
    txMsgEvt = DsmApi_CreateEvent(DS485_BROADCAST_ADDRESS, deviceAddress, DSM_TRANSACTION_ID, EVENT_DEVICE_MODEL_CHANGED, 0, 2);
    DsmApi_SetEvtData16(txMsgEvt, 0, deviceAddress);
    /* send event */
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendDeviceInfoRequest(QActive * requester, uint8_t destination, uint8_t transactionId, DeviceInfoModifiers_t modifier, uint8_t localZoneId, uint16_t index)
{
}

/*..........................................................................*/
void DsmApi_SendZoneDeviceCountRequest(QActive * requester, uint8_t destination, uint8_t transactionId, DeviceInfoModifiers_t modifier, uint8_t localZoneId)
{
}

void DsmApi_SendOPCRequest(DsmApi *me, uint16_t deviceId, uint8_t modifier, uint8_t channelID, uint16_t value, uint8_t sceneID, uint8_t size)
{
    DatamodelReqDeviceIdEvt * reqEvt =
        Q_NEW(DatamodelReqDeviceIdEvt, DATAMODEL_REQ_DEVICE_ENTRY_BY_ID_SIG);
    reqEvt->basic.requester = (QActive *)me;
    reqEvt->deviceId = deviceId;
    me->deviceParameterInfo.operation = modifier;
    me->deviceParameterInfo.channel = channelID;
    me->deviceParameterInfo.value = value;
    me->deviceParameterInfo.scene = sceneID;
    me->deviceParameterInfo.is16bit = (size == 1);
    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
}

/*..........................................................................*/
void DsmApi_SendStatusEvent(uint8_t hardware, uint8_t temperature, uint8_t config, uint8_t ds485, uint8_t line)
{
    DS485MessageEvt * txMsgEvt = DsmApi_CreateEvent(DS485_BROADCAST_ADDRESS, 0, DSM_TRANSACTION_ID, EVENT_DSM_STATE_CHANGE, 0, 6);
    DsmApi_SetResData8(txMsgEvt, 0, hardware);
    DsmApi_SetResData8(txMsgEvt, 1, 0);
    DsmApi_SetResData8(txMsgEvt, 2, temperature);
    DsmApi_SetResData8(txMsgEvt, 3, config);
    DsmApi_SetResData8(txMsgEvt, 4, ds485);
    DsmApi_SetResData8(txMsgEvt, 5, line);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendDsmDoneSetDeviceEvent(uint8_t modifier, uint8_t error, uint16_t deviceId)
{
    DS485MessageEvt *txMsgEvt = DsmApi_CreateDsmDoneEvent(deviceId, modifier, error, 2);
    DsmApi_SetResData16(txMsgEvt, 0, deviceId);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendDsmDoneSetDeviceConfigEvent(uint8_t error, uint16_t deviceId, uint8_t bank, uint8_t offset)
{
    DS485MessageEvt *txMsgEvt = DsmApi_CreateDsmDoneEvent(deviceId, EVENT_DSM_DONE_SET_DEVICE_CONFIG, error, 4);
    DsmApi_SetResData16(txMsgEvt, 0, deviceId);
    DsmApi_SetResData8(txMsgEvt, 2, bank);
    DsmApi_SetResData8(txMsgEvt, 3, offset);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
void DsmApi_SendHeatingControllerStateEvent(uint16_t zoneId, uint8_t heatingState)
{
    DS485MessageEvt * txMsgEvt = DsmApi_CreateEvent(DS485_BROADCAST_ADDRESS, 0, DSM_TRANSACTION_ID, EVENT_HEATING_CONTROLLER_STATE, 0, 3);
    DsmApi_SetResData16(txMsgEvt, 0, zoneId);
    DsmApi_SetResData8(txMsgEvt, 2, heatingState);
    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
}

/*..........................................................................*/
static QState DsmApi_initial(DsmApi *me, QEvent const *e)
{
    (void)e;        /* suppress the compiler warning about unused parameter */

    /* QS dictionary entries */
    QS_OBJ_DICTIONARY(&l_dsmApi);

    QS_FUN_DICTIONARY(&DsmApi_top);
    QS_FUN_DICTIONARY(&DsmApi_idle);
    QS_FUN_DICTIONARY(&DsmApi_busy);
    QS_FUN_DICTIONARY(&DsmApi_device_info);
    QS_FUN_DICTIONARY(&DsmApi_device_state);
    QS_FUN_DICTIONARY(&DsmApi_zone_count);
    QS_FUN_DICTIONARY(&DsmApi_zone_device_count);
    QS_FUN_DICTIONARY(&DsmApi_zone_info);
    QS_FUN_DICTIONARY(&DsmApi_wait_zone_group_info);
    QS_FUN_DICTIONARY(&DsmApi_wait_scene_info);
    QS_FUN_DICTIONARY(&DsmApi_wait_dsm_info);
    QS_FUN_DICTIONARY(&DsmApi_wait_dsm_apartment_id);
    QS_FUN_DICTIONARY(&DsmApi_wait_dsm_temperature_limits);
    QS_FUN_DICTIONARY(&DsmApi_wait_flash_read);
    QS_FUN_DICTIONARY(&DsmApi_start_update);
    QS_FUN_DICTIONARY(&DsmApi_system_reset);
    QS_FUN_DICTIONARY(&DsmApi_system_reset_and_update);
    QS_FUN_DICTIONARY(&DsmApi_stopped);

    return Q_TRAN(&DsmApi_idle);
}

/*..........................................................................*/
static QState DsmApi_top(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG:
        /* subscribe events */
        QActive_subscribe((QActive *)me, DATAMODEL_ZONE_ID_SIG);
        return Q_HANDLED();
    case DATAMODEL_ZONE_ID_SIG: {
        /* save zone ID */
        DatamodelZoneIdEvt const * evt = (DatamodelZoneIdEvt const *)e;
        me->zoneIds[evt->localZoneId] = evt->zoneId;
        return Q_HANDLED();
    }
    case DS485_RX_DATA_SIG: {
        /* handle only responses */
        DS485MessageEvt const * msgEvt = (DS485MessageEvt const *)e;
        if (DS485_IsResponse(msgEvt)) {
            if (dsuid_equal(&me->pendingRequest.busAddress, &msgEvt->sourceAddr)) {
                /* message from the correct source received */
                if ((me->pendingRequest.transactionId == DsmApi_GetTransactionId(msgEvt)) &&
                    (me->pendingRequest.messageId == DsmApi_GetMessageId(msgEvt)) &&
                    (me->pendingRequest.modifierId == DsmApi_GetModifierId(msgEvt))) {
                    /* clear bus address */
                    me->pendingRequest.busAddress = DSUID_BROADCAST;
                    /* stop timer */
                    QTimeEvt_disarm(&me->pendingRequest.timer);
                    /* forward event */
                    QActive_postFIFO(me->pendingRequest.requester, e);
                }
            }
        }
        return Q_HANDLED();
    }
    case DSMAPI_REQUEST_TIMEOUT_SIG: {
        if (!dsuid_is_broadcast(&me->pendingRequest.busAddress)) {
            /* clear bus address */
            me->pendingRequest.busAddress = DSUID_BROADCAST;
            /* forward request timeout event to requester */
            QActive_postFIFO(me->pendingRequest.requester, e);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
/* For the defer/recall mechanism to work correctly Q_HANDLED() must never be called
 * in this state (idle) except for Q_ENTRY_SIG and Q_EXIT_SIG. Always use
 * Q_TRAN(&DsmApi_idle) instead.
 */
static QState DsmApi_idle(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG:
        /* recall event */
        QActive_recall((QActive *)me, &me->deferredQueue);
        return Q_HANDLED();
    case DS485_RX_LONG_DATA_SIG: {
        DS485NetlibPacketMessageEvt const * rxMsgEvt = (DS485NetlibPacketMessageEvt const *)e;
        DS485NetlibPacketMessageEvt *txMsgEvt;
        /* break if it is not a request */
        if (!DS485_IsNetlibPacketRequest(rxMsgEvt)) {
            break;
        }
        me->requestAddress = rxMsgEvt->sourceAddr;
        me->destinationAddress = rxMsgEvt->destinationAddr;
        me->transactionId = rxMsgEvt->message[DSMAPI_LONG_TRANSACTIONID_OFFSET];
        me->messageId     = rxMsgEvt->message[DSMAPI_LONG_MSGID_OFFSET];
        me->modifierId    = rxMsgEvt->message[DSMAPI_LONG_MODIFIER_OFFSET];

        switch (me->messageId) {
        case USER_PROTOBUF_MESSAGE_REQUEST: {
            uint16_t size = ((rxMsgEvt->message[DSMAPI_REQ_LEN_OFFSET_HI] << 8) & 0xff00) |
                            (rxMsgEvt->message[DSMAPI_REQ_LEN_OFFSET_LO] & 0xff);

            if (size <= VDC_USER_PROTOBUF_MESSAGE_LENGTH) {
                SelectBProtobufMessageRequestEvt *reqEvt =
                        Q_NEW(SelectBProtobufMessageRequestEvt, SELECT_B_USER_PROTOBUF_MESSAGE_REQUEST_SIG);
                reqEvt->dsuid = rxMsgEvt->destinationAddr;
                reqEvt->size = size;
                memcpy(reqEvt->request,
                       /* +1 transaction id, +1 messageId, +1 modifierId + 2 size */
                       &rxMsgEvt->message[DS485N_PACKET_DATA_OFFSET + 5],
                       reqEvt->size);

                QActive_postFIFO(AO_SelectB, (QEvent *)reqEvt);
                return Q_TRAN(&DsmApi_wait_vdc_user_protobuf_message_request);
            } else {
                txMsgEvt = DsmApi_CreateLongResponse(me, ERROR_WRONG_PARAMETER, 0);
            }
            break;
        }
        default:
            txMsgEvt = DsmApi_CreateLongResponse(me, ERROR_WRONG_MSGID, 0);
        }
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    case DS485_RX_DATA_SIG: {
        uint8_t modifier, error;
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        /* break if it is not a request */
        if (!DS485_IsRequest((DS485MessageEvt const *)e)) {
            break;
        }
        /* save request address, transaction ID, message ID and modifier ID */
        me->requestAddress = rxMsgEvt->sourceAddr;
        me->destinationAddress = rxMsgEvt->destinationAddr;
        me->transactionId = rxMsgEvt->message[DSMAPI_TRANSACTIONID_OFFSET];
        me->messageId     = rxMsgEvt->message[DSMAPI_MSGID_OFFSET];
        me->modifierId    = rxMsgEvt->message[DSMAPI_MODIFIER_OFFSET];
        /* get modifier */
        modifier = me->modifierId;
        /* set default error (wrong message ID) */
        error = ERROR_WRONG_MSGID;
        /* check if a response is expected */
        if (!DS485_IsBroadcast(rxMsgEvt)) {
            me->sendResponse = TRUE;
        } else {
            me->sendResponse = FALSE;
        }
        switch (me->messageId) {
        /* --------------------------------------------- */
        case ZONE_MODIFY: {
            uint16_t zoneId;
            uint16_t newZoneId = 0;
            /* check modifier */
            error = ERROR_WRONG_MODIFIER;
            if ((modifier == ZONE_MODIFY_ADD) || (modifier == ZONE_MODIFY_REMOVE)) {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                    /* get zone ID */
                    zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    error = ERROR_OK;
                } else {
                    error = ERROR_WRONG_SIZE;
                }
            } else if (modifier == ZONE_MODIFY_CHANGE_ID) {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 4) {
                    /* get zone IDs */
                    zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    newZoneId = DsmApi_GetReqData16(rxMsgEvt, 2);
                    error = ERROR_OK;
                } else {
                    error = ERROR_WRONG_SIZE;
                }
            }
            if (error == ERROR_OK) {
                DatamodelReqZoneModifyEvt * reqEvt =
                    Q_NEW(DatamodelReqZoneModifyEvt, DATAMODEL_REQ_ZONE_MODIFY_SIG);
                reqEvt->basic.requester = (QActive *)me;
                reqEvt->modifier = modifier;
                reqEvt->zoneId = zoneId;
                reqEvt->newZoneId = newZoneId;
                reqEvt->dsuid = rxMsgEvt->destinationAddr;
                QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                return Q_TRAN(&DsmApi_busy);
            }
            break;
        }
        /* --------------------------------------------- */
        case ZONE_DEVICE_COUNT: {
            /* check length */
            if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                /* check modifier */
                if ((modifier == ZONE_DEVICE_COUNT_ONLY_ACTIVE) ||
                    (modifier == ZONE_DEVICE_COUNT_ONLY_INACTIVE) ||
                    (modifier == ZONE_DEVICE_COUNT_ALL)) {
                    /* get zone ID */
                    uint16_t zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    DatamodelReqZoneDeviceCountEvt * reqEvt =
                        Q_NEW(DatamodelReqZoneDeviceCountEvt, DATAMODEL_REQ_ZONE_DEVICE_COUNT_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->zoneIdModifier = ZONE_INFO_BY_INDEX;
                    reqEvt->modifier = modifier;
                    reqEvt->zoneId = zoneId;
                    reqEvt->dsuid = rxMsgEvt->destinationAddr;
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_zone_device_count);
                } else {
                    error = ERROR_WRONG_MODIFIER;
                }
            } else {
                error = ERROR_WRONG_SIZE;
            }
            break;
        }
        /* --------------------------------------------- */
        case DEVICE_INFO: {
            switch (modifier) {
            case DEVICE_INFO_BY_INDEX_ONLY_ACTIVE:
                /* fall through */
            case DEVICE_INFO_BY_INDEX_ONLY_INACTIVE:
                /* fall through */
            case DEVICE_INFO_BY_INDEX: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 4) {
                    DatamodelReqDeviceInfoOfIndexEvt * reqEvt =
                        Q_NEW(DatamodelReqDeviceInfoOfIndexEvt, DATAMODEL_REQ_DEVICE_INFO_OF_INDEX_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->zoneIdModifier = ZONE_INFO_BY_INDEX;
                    reqEvt->modifier = modifier;
                    reqEvt->zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    reqEvt->deviceIndex = DsmApi_GetReqData16(rxMsgEvt, 2);
                    reqEvt->privateId = 0;
                    reqEvt->dsuid = rxMsgEvt->destinationAddr;
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_device_info);

                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DEVICE_INFO_BY_DEVICE_ID: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                    DatamodelReqDeviceIdEvt * reqEvt =
                        Q_NEW(DatamodelReqDeviceIdEvt, DATAMODEL_REQ_DEVICE_ENTRY_BY_ID_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->deviceId = DsmApi_GetReqData16( rxMsgEvt, 0 );
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_device_info);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            default: {
                error = ERROR_WRONG_MODIFIER;
            }
            }
            break;
        }
        /* --------------------------------------------- */
        case DEVICE_STATE:
            /* check length */
            if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                /* get device ID */
                uint16_t deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                /* check device ID range */
                if ((deviceId >= MIN_DEVICE) && (deviceId <= MAX_DEVICE)) {
                    DatamodelReqDeviceIdEvt * reqEvt =
                        Q_NEW(DatamodelReqDeviceIdEvt, DATAMODEL_REQ_DEVICE_STATE_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->deviceId = deviceId;
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_device_state);
                }
                /* device ID is out of range */
                error = ERROR_WRONG_PARAMETER;
            } else {
                error = ERROR_WRONG_SIZE;
            }
            break;
        /* --------------------------------------------- */
        case ZONE_COUNT: {
            DatamodelReqDsmEvt * reqEvt =
                Q_NEW(DatamodelReqDsmEvt, DATAMODEL_REQ_ZONE_COUNT_SIG);
            reqEvt->requester = (QActive *)me;
            reqEvt->dsuid = rxMsgEvt->destinationAddr;
            QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
            return Q_TRAN(&DsmApi_zone_count);
        }
        /* --------------------------------------------- */
        case ZONE_INFO: {
            switch (modifier) {
            case ZONE_INFO_BY_INDEX:
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 1) {
                    DatamodelReqZoneInfoOfIndexEvt * reqEvt =
                        Q_NEW(DatamodelReqZoneInfoOfIndexEvt, DATAMODEL_REQ_ZONE_INFO_OF_INDEX_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->zoneIndex = DsmApi_GetReqData8(rxMsgEvt, 0);
                    reqEvt->dsuid = rxMsgEvt->destinationAddr;
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_zone_info);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            case ZONE_INFO_BY_ID:
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                    DatamodelReqZoneInfoByIdEvt * reqEvt =
                        Q_NEW(DatamodelReqZoneInfoByIdEvt, DATAMODEL_REQ_ZONE_INFO_BY_ZONE_ID_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_zone_info);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            default:
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            break;
        }
        /* --------------------------------------------- */
        case ZONE_PROPERTIES: {
            switch (modifier) {
            case ZONE_PROPERTIES_SET_NAME: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == (2 + MAX_DSM_API_NAME_LEN)) {
                    DatamodelReqZoneNameEvt * reqEvt =
                        Q_NEW(DatamodelReqZoneNameEvt, DATAMODEL_REQ_ZONE_SET_NAME_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    reqEvt->dsuid = rxMsgEvt->destinationAddr;
                    DsmApi_GetReqArray8(rxMsgEvt, 2, reqEvt->name, MAX_NAME_LEN);
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case ZONE_PROPERTIES_GET_SCENE_HISTORY: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                    uint16_t zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    uint8_t localZoneId = DsmApi_GetLocalZoneId(me, zoneId);
                    if (localZoneId != 0) {
                        ZoneSceneHistoryReqEvt * reqEvt = Q_NEW(ZoneSceneHistoryReqEvt, SELECT_A_REQ_ZONE_SCENE_HISTORY_SIG);
                        reqEvt->requester = (QActive *)me;
                        reqEvt->localZoneId = localZoneId;
                        QActive_postFIFO(AO_SelectA, (QEvent *)reqEvt);
                        return Q_TRAN(&DsmApi_wait_zone_history);
                    } else {
                        error = ERROR_ZONE_NOT_FOUND;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case ZONE_PROPERTIES_GET_ROOM_STATES: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                    uint16_t zoneId;
                    uint8_t localZoneId;
                    /* get/check zone ID */
                    zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    /* find local zone ID */
                    localZoneId = DsmApi_GetLocalZoneId(me, zoneId);
                    if (localZoneId == 0) {
                        error = ERROR_ZONE_NOT_FOUND;
                    } else {
                        ZoneStatesReqEvt * reqEvt = Q_NEW(ZoneStatesReqEvt, SELECT_A_REQ_ROOM_STATES_SIG);
                        reqEvt->localZoneId = localZoneId;
                        QActive_postFIFO(AO_SelectA, (QEvent *)reqEvt);
                        return Q_TRAN(&DsmApi_wait_room_states);
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case ZONE_PROPERTIES_SET_ZONE_SENSOR: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == (3 + DSUID_LEN)) {
                    uint8_t localZoneId;
                    dsuid_t dsuid;
                    uint16_t zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    uint8_t sensorType = DsmApi_GetReqData8(rxMsgEvt, 2);
                    DsmApi_GetReqArray8(rxMsgEvt, 3, dsuid.id, DSUID_SIZE);
                    localZoneId = DsmApi_GetLocalZoneId(me, zoneId);
                    if (localZoneId == 0) {
                        error = ERROR_ZONE_NOT_FOUND;
                    } else {
                        error = Datamodel_SetSensorPushDevice(localZoneId, sensorType, &dsuid);
                        DS485MessageEvt *txMsgEvt = DsmApi_CreateResponse(me, error, 0);
                        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
                        return Q_TRAN(&DsmApi_idle);
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case ZONE_PROPERTIES_GET_ZONE_SENSOR: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    uint16_t zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    uint8_t sensorType = DsmApi_GetReqData8(rxMsgEvt, 2);
                    uint8_t localZoneId = DsmApi_GetLocalZoneId(me, zoneId);
                    if (localZoneId == 0) {
                        error = ERROR_ZONE_NOT_FOUND;
                    } else {
                        dsuid_t sensor_dsuid;
                        DS485MessageEvt *txMsgEvt = NULL;
                        if (Datamodel_GetSensorPushDevice(localZoneId, sensorType, &sensor_dsuid) == ERROR_OK) {
                            txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, DSUID_LEN);
                            DsmApi_SetResArray8(txMsgEvt, 0, sensor_dsuid.id, DSUID_LEN);
                        } else {
                            txMsgEvt = DsmApi_CreateResponse(me, ERROR_DEVICE_NOT_FOUND, 0);
                        }

                        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
                        return Q_TRAN(&DsmApi_idle);
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case ZONE_PROPERTIES_RESET_ZONE_SENSOR: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    uint16_t zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    uint8_t sensorType = DsmApi_GetReqData8(rxMsgEvt, 2);
                    uint8_t localZoneId = DsmApi_GetLocalZoneId(me, zoneId);
                    if (localZoneId == 0) {
                        error = ERROR_ZONE_NOT_FOUND;
                    } else {
                        uint8_t errorCode = Datamodel_ResetSensorPushDevice(localZoneId, sensorType);
                        DS485MessageEvt *txMsgEvt = DsmApi_CreateResponse(me, errorCode, 0);
                        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
                        return Q_TRAN(&DsmApi_idle);
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case ZONE_PROPERTIES_GET_ZONE_SENSOR_VALUE: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    uint16_t zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    uint8_t sensorType = DsmApi_GetReqData8(rxMsgEvt, 2);
                    uint8_t localZoneId = DsmApi_GetLocalZoneId(me, zoneId);
                    if (localZoneId > 0) {
                        Sensor sensor = SelectA_SensorsGetSensor(localZoneId, sensorType);
                        if (sensor.value != SENSOR_INVALID_VALUE) {
                            /* sensor value is known */
                            DS485MessageEvt *txMsgEvt;
                            /* return sensor value and age */
                            txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 6);
                            DsmApi_SetResData16(txMsgEvt, 0, sensor.value);
                            DsmApi_SetResData32(txMsgEvt, 2, SelectA_SensorsGetAge(sensor.timestamp));
                            QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
                            return Q_TRAN(&DsmApi_idle);
                        } else {
                            error = ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
                        }
                    } else {
                        error = ERROR_ZONE_NOT_FOUND;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }

            default: {
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            }
            break;
        }
        /* --------------------------------------------- */
        case DEVICE_STATISTICS_REQUEST: {
            switch (modifier) {
            case DEVICE_STATISTICS_REQUEST_GET: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    uint16_t deviceId;
                    uint8_t paramOffset;
                    /* save parameters */
                    deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    paramOffset = DsmApi_GetReqData8(rxMsgEvt, 2);
                    /* check parameters */
                    if ((deviceId >= MIN_DEVICE) && (deviceId <= MAX_DEVICE)) {
                        /* allocate get device statistics event */
                        GetDeviceStatisticsEvt * reqEvt =
                            Q_NEW(GetDeviceStatisticsEvt, SELECT_B_GET_DEVICE_STATISTICS_SIG);
                        /* set request parameters */
                        reqEvt->deviceId = deviceId;
                        reqEvt->paramOffset = paramOffset;
                        QActive_postFIFO(AO_SelectB, (QEvent *)reqEvt);
                        /* return OK */
                        error = ERROR_OK;
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            default: {
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            }
            break;
        }
        /* --------------------------------------------- */
        case DEVICE_SENSOR: {
            switch (modifier) {
            case MOD_DEV_SENSOR_GET_VALUE: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    uint16_t deviceId;
                    uint8_t sensorId;
                    /* save parameters */
                    deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    sensorId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    /* check parameters */
                    if ((deviceId >= MIN_DEVICE) && (deviceId <= MAX_DEVICE)) {
                        if (Datamodel_IsDeviceValid(deviceId)) {
                            /* allocate get device sensor event */
                            GetDeviceSensorEvt * reqEvt =
                                Q_NEW(GetDeviceSensorEvt, SELECT_B_GET_DEVICE_SENSOR_SIG);
                            /* set request parameters */
                            reqEvt->deviceId = deviceId;
                            reqEvt->sensorId = sensorId;
                            QActive_postFIFO(AO_SelectB, (QEvent *)reqEvt);
                            /* return OK */
                            error = ERROR_OK;
                        } else {
                            error = ERROR_DEVICE_NOT_FOUND;
                        }
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case MOD_DEV_SENSOR_GET_COUNT: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                    DS485MessageEvt *txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 1);
                    uint16_t deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    DsmApi_SetResData8(txMsgEvt, 0, Datamodel_GetDeviceSensorCount(deviceId));
                    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
                    return Q_TRAN(&DsmApi_idle);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case MOD_DEV_SENSOR_TABLE_ENTRY: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    DS485MessageEvt *txMsgEvt;
                    uint16_t deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    uint8_t index = DsmApi_GetReqData8(rxMsgEvt, 2), type, broadcast, pushConvert, offset = 0;
                    uint32_t pollInterval;
                    if (Datamodel_GetDeviceSensor(deviceId, index, &type, &pollInterval, &broadcast, &pushConvert)) {
                        txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 7);
                        DsmApi_SetResData8(txMsgEvt, offset++, type);
                        DsmApi_SetResData32(txMsgEvt, offset, pollInterval);
                        offset += 4;
                        DsmApi_SetResData8(txMsgEvt, offset++, broadcast);
                        DsmApi_SetResData8(txMsgEvt, offset++, pushConvert);
                    } else {
                        txMsgEvt = DsmApi_CreateResponse(me, ERROR_WRONG_PARAMETER, 0);
                    }
                    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
                    return Q_TRAN(&DsmApi_idle);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            default: {
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            }
            break;
        }
        /* --------------------------------------------- */
        case DEVICE_DSLINK_SEND: {
            switch (modifier) {
            case DEVICE_DSLINK_SEND_NOT_LAST_BYTE_NO_ANSWER:
                /* fall through */
            case DEVICE_DSLINK_SEND_LAST_BYTE_NO_ANSWER:
                /* fall through */
            case DEVICE_DSLINK_SEND_NOT_LAST_BYTE_WITH_ANSWER:
                /* fall through */
            case DEVICE_DSLINK_SEND_LAST_BYTE_WITH_ANSWER: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    uint16_t deviceId;
                    uint8_t value;
                    /* save parameters */
                    deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    value = DsmApi_GetReqData8(rxMsgEvt, 2);
                    /* check parameters */
                    if ((deviceId >= MIN_DEVICE) && (deviceId <= MAX_DEVICE)) {
                        /* allocate get device statistics event */
                        GetDevicedSLinkEvt * reqEvt =
                            Q_NEW(GetDevicedSLinkEvt, SELECT_B_GET_DEVICE_DSLINK_SIG);
                        /* set request parameters */
                        reqEvt->operation = (DSLinkOperation_t)(modifier - DEVICE_DSLINK_SEND_NOT_LAST_BYTE_NO_ANSWER);
                        reqEvt->deviceId = deviceId;
                        reqEvt->value = value;
                        QActive_postFIFO(AO_SelectB, (QEvent *)reqEvt);
                        /* return OK */
                        error = ERROR_OK;
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            default: {
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            }
            break;
        }
        /* --------------------------------------------- */
        case ZONE_GROUP_DSLINK_SEND: {
            switch (modifier) {
            case DEVICE_DSLINK_SEND_NOT_LAST_BYTE_NO_ANSWER:
                /* fall through */
            case DEVICE_DSLINK_SEND_LAST_BYTE_NO_ANSWER: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 4) {
                    uint8_t groupId, value;
                    /* get/check group and value */
                    groupId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    value = DsmApi_GetReqData8(rxMsgEvt, 3);
                    if (groupId <= MAX_GROUP) {
                        /* get zone ID */
                        uint16_t zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                        /* find local zone ID */
                        uint8_t localZoneId = DsmApi_GetLocalZoneId(me, zoneId);
                        if (zoneId == 0 || localZoneId != 0) {
                            /* allocate get device statistics event */
                            GetZoneGroupdSLinkEvt * reqEvt =
                                Q_NEW(GetZoneGroupdSLinkEvt, SELECT_B_GET_ZONE_GROUP_DSLINK_SIG);
                            /* set request parameters */
                            reqEvt->operation = (DSLinkOperation_t)(modifier - DEVICE_DSLINK_SEND_NOT_LAST_BYTE_NO_ANSWER);
                            reqEvt->localZoneId = localZoneId;
                            reqEvt->groupId = groupId;
                            reqEvt->value = value;
                            QActive_postFIFO(AO_SelectB, (QEvent *)reqEvt);
                            error = ERROR_OK;
                        } else /*(zoneId != 0 && localZoneId == 0)*/ {
                            error = ERROR_ZONE_NOT_FOUND;
                        }
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            default: {
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            }
            break;
        }
        /* --------------------------------------------- */
        case TEST_TRANSMISSION_QUALITY: {
            switch (modifier) {
            case TEST_TRANSMISSION_QUALITY_GET: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                    uint16_t deviceId;
                    /* save parameters */
                    deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    /* check parameters */
                    if ((deviceId >= MIN_DEVICE) && (deviceId <= MAX_DEVICE)) {
                        /* allocate test transmission quality event */
                        GetTransmissionQualityEvt * reqEvt =
                            Q_NEW(GetTransmissionQualityEvt, SELECT_B_TEST_TRANSMISSION_QUALITY_SIG);
                        /* set request parameters */
                        reqEvt->deviceId = deviceId;
                        QActive_postFIFO(AO_SelectB, (QEvent *)reqEvt);
                        /* return OK */
                        error = ERROR_OK;
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            default: {
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            }
            break;
        }
        /* --------------------------------------------- */
        case TEST_DEVICE_PRESENCE: {
            if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                uint16_t deviceId;
                /* save parameters */
                deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                /* check parameters */
                if ((deviceId >= MIN_DEVICE) && (deviceId <= MAX_DEVICE)) {
                    if (Datamodel_IsDeviceValid(deviceId)) {
                        /* allocate test device presence event */
                        GetDevicePresenceEvt * reqEvt =
                            Q_NEW(GetDevicePresenceEvt, SELECT_B_TEST_DEVICE_PRESENCE_SIG);
                        /* set request parameters */
                        reqEvt->deviceId = deviceId;
                        QActive_postFIFO(AO_SelectB, (QEvent *)reqEvt);
                        /* return OK */
                        error = ERROR_OK;
                    } else {
                        error = ERROR_DEVICE_NOT_FOUND;
                    }
                } else {
                    error = ERROR_WRONG_PARAMETER;
                }
            } else {
                error = ERROR_WRONG_SIZE;
            }
            break;
        }
        /* --------------------------------------------- */
        case DEVICE_CONFIG: {
            switch (modifier) {
            case DEVICE_CONFIG_GET: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 4) {
                    uint16_t deviceId;
                    uint8_t paramBank;
                    uint8_t paramOffset;
                    /* save parameters */
                    deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    paramBank = DsmApi_GetReqData8(rxMsgEvt, 2);
                    paramOffset = DsmApi_GetReqData8(rxMsgEvt, 3);
                    /* check parameters */
                    if ((deviceId >= MIN_DEVICE) && (deviceId <= MAX_DEVICE)) {
                        DatamodelReqDeviceIdEvt * reqEvt =
                            Q_NEW(DatamodelReqDeviceIdEvt, DATAMODEL_REQ_DEVICE_ENTRY_BY_ID_SIG);
                        reqEvt->basic.requester = (QActive *)me;
                        reqEvt->deviceId = DsmApi_GetReqData16( rxMsgEvt, 0 );
                        me->deviceParameterInfo.operation = DEVICE_CONFIG_GET;
                        me->deviceParameterInfo.bank = paramBank;
                        me->deviceParameterInfo.offset = paramOffset;
                        QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                        return Q_TRAN(&DsmApi_device_parameter);
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DEVICE_CONFIG_SET: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 5) {
                    uint16_t deviceId;
                    uint8_t paramBank;
                    uint8_t paramOffset;
                    uint8_t value;
                    /* save parameters */
                    deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    paramBank = DsmApi_GetReqData8(rxMsgEvt, 2);
                    paramOffset = DsmApi_GetReqData8(rxMsgEvt, 3);
                    value = DsmApi_GetReqData8(rxMsgEvt, 4);
                    /* check parameters */
                    if ((deviceId >= MIN_DEVICE) && (deviceId <= MAX_DEVICE)) {
                        DatamodelReqDeviceIdEvt * reqEvt =
                            Q_NEW(DatamodelReqDeviceIdEvt, DATAMODEL_REQ_DEVICE_ENTRY_BY_ID_SIG);
                        reqEvt->basic.requester = (QActive *)me;
                        reqEvt->deviceId = DsmApi_GetReqData16( rxMsgEvt, 0 );
                        me->deviceParameterInfo.operation = DEVICE_CONFIG_SET;
                        me->deviceParameterInfo.bank = paramBank;
                        me->deviceParameterInfo.offset = paramOffset;
                        me->deviceParameterInfo.value = value;
                        QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                        return Q_TRAN(&DsmApi_device_parameter);
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            default: {
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            }
            break;
        }
        /* --------------------------------------------- */
        case DEVICE_PROPERTIES: {
            switch (modifier) {
            case DEVICE_PROPERTIES_SET_ZONE: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 4) {
                    DatamodelReqDeviceSetZoneEvt * reqEvt =
                        Q_NEW(DatamodelReqDeviceSetZoneEvt, DATAMODEL_REQ_DEVICE_SET_ZONE_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->deviceAddress = DsmApi_GetReqData16(rxMsgEvt, 0);
                    reqEvt->zoneId = DsmApi_GetReqData16(rxMsgEvt, 2);
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DEVICE_PROPERTIES_SET_BUTTON_ACTIVE_GROUP: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    DatamodelReqDeviceSetButtonGroupEvt * reqEvt =
                        Q_NEW(DatamodelReqDeviceSetButtonGroupEvt, DATAMODEL_REQ_DEVICE_SET_BUTTON_ACTIVE_GROUP_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->deviceAddress = DsmApi_GetReqData16(rxMsgEvt, 0);
                    reqEvt->groupId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DEVICE_PROPERTIES_RESET_BUTTON_ACTIVE_GROUP: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                    DatamodelReqDeviceSetButtonGroupEvt * reqEvt =
                        Q_NEW(DatamodelReqDeviceSetButtonGroupEvt, DATAMODEL_REQ_DEVICE_SET_BUTTON_ACTIVE_GROUP_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->deviceAddress = DsmApi_GetReqData16(rxMsgEvt, 0);
                    reqEvt->groupId = DATAMODEL_GROUP_ID_NOT_APPLICABLE;
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DEVICE_PROPERTIES_SET_LOCKED_FLAG: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    DatamodelReqDeviceSetLockedEvt * reqEvt =
                        Q_NEW(DatamodelReqDeviceSetLockedEvt, DATAMODEL_REQ_DEVICE_SET_LOCKED_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->deviceAddress = DsmApi_GetReqData16(rxMsgEvt, 0);
                    reqEvt->locked = ((DsmApi_GetReqData8(rxMsgEvt, 2) == 0) ? FALSE : TRUE);
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DEVICE_PROPERTIES_SET_NAME: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == (2 + MAX_DSM_API_NAME_LEN)) {
                    DatamodelReqDeviceNameEvt * reqEvt =
                        Q_NEW(DatamodelReqDeviceNameEvt, DATAMODEL_REQ_DEVICE_SET_NAME_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->deviceAddress = DsmApi_GetReqData16(rxMsgEvt, 0);
                    DsmApi_GetReqArray8(rxMsgEvt, 2, reqEvt->name, MAX_NAME_LEN);
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DEVICE_PROPERTIES_SET_BUTTON_SET_LOCAL_PRIORITY: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    DatamodelReqDeviceSetButtonLocalPriorityEvt * reqEvt =
                        Q_NEW(DatamodelReqDeviceSetButtonLocalPriorityEvt, DATAMODEL_REQ_DEVICE_SET_SET_LOCAL_PRIO_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->deviceAddress = DsmApi_GetReqData16(rxMsgEvt, 0);
                    reqEvt->setLocalPriority = (DsmApi_GetReqData8(rxMsgEvt, 2) != 0);
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DEVICE_PROPERTIES_SET_BUTTON_SET_NO_COMING_HOME_CALL: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    DatamodelReqDeviceSetButtonNoComingHomeEvt * reqEvt =
                        Q_NEW(DatamodelReqDeviceSetButtonNoComingHomeEvt, DATAMODEL_REQ_DEVICE_SET_NO_COMING_HOME_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->deviceAddress = DsmApi_GetReqData16(rxMsgEvt, 0);
                    reqEvt->noComingHome = (DsmApi_GetReqData8(rxMsgEvt, 2) != 0);
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DEVICE_PROPERTIES_SET_BUTTON_SET_APARTMENT: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    DatamodelReqDeviceSetButtonApartmentEvt * reqEvt =
                        Q_NEW(DatamodelReqDeviceSetButtonApartmentEvt, DATAMODEL_REQ_DEVICE_SET_BUTTON_APARTMENT_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->deviceAddress = DsmApi_GetReqData16(rxMsgEvt, 0);
                    reqEvt->apartment = (DsmApi_GetReqData8(rxMsgEvt, 2) != 0);
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DEVICE_PROPERTIES_SET_BUTTON_SET_OUTPUT_CHANNEL: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    DatamodelReqDeviceSetButtonOutputChannelEvt * reqEvt =
                        Q_NEW(DatamodelReqDeviceSetButtonOutputChannelEvt, DATAMODEL_REQ_DEVICE_SET_BUTTON_OUTPUT_CHANNEL_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->deviceAddress = DsmApi_GetReqData16(rxMsgEvt, 0);
                    reqEvt->channelId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    reqEvt->sendDownstream = TRUE;
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DEVICE_PROPERTIES_SET_BUTTON_SET_GLOBAL_GROUP: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    DatamodelReqDeviceSetButtonGlobalGroupEvt * reqEvt =
                        Q_NEW(DatamodelReqDeviceSetButtonGlobalGroupEvt, DATAMODEL_REQ_DEVICE_SET_BUTTON_GLOBAL_GROUP_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->deviceAddress = DsmApi_GetReqData16(rxMsgEvt, 0);
                    reqEvt->group = DsmApi_GetReqData8(rxMsgEvt, 2);
                    reqEvt->sendDownstream = TRUE;
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            default: {
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            }
            break;
        }
        /* --------------------------------------------- */
        case DEVICE_BUTTON_INFO: {
            switch (modifier) {
            case DEVICE_BUTTON_INFO_BY_DEVICE: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                    DatamodelReqDeviceInfoEvt * reqEvt =
                        Q_NEW(DatamodelReqDeviceInfoEvt, DATAMODEL_REQ_DEVICE_ENTRY_BY_ID_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->deviceAddress = DsmApi_GetReqData16(rxMsgEvt, 0);
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_button_info);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            default: {
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            }
            break;
        }
        /* --------------------------------------------- */
        case DEVICE_BINARY_INPUT: {
            switch (modifier) {
                case DEVICE_BINARY_INPUT_GET_COUNT: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                    DS485MessageEvt *txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 1);
                    uint16_t deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    DsmApi_SetResData8(txMsgEvt, 0, Datamodel_GetDeviceBinaryInputCount(deviceId));
                    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
                    return Q_TRAN(&DsmApi_idle);
                } else {
                    error = ERROR_WRONG_SIZE;
                    break;
                }
            }
            case DEVICE_BINARY_INPUT_GET_BY_INDEX: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    DS485MessageEvt *txMsgEvt;
                    uint16_t deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    uint8_t index = DsmApi_GetReqData8(rxMsgEvt, 2), targetGroup, type, buttonId, offset = 0;
                    if (Datamodel_GetDeviceBinaryInput(deviceId, index, &targetGroup, &type, &buttonId)) {
                        txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, sizeof(BinaryInputConfig));
                        DsmApi_SetResData8(txMsgEvt, offset++, targetGroup >> 6);
                        DsmApi_SetResData8(txMsgEvt, offset++, (targetGroup & 0x3F));
                        DsmApi_SetResData8(txMsgEvt, offset++, type);
                        DsmApi_SetResData8(txMsgEvt, offset++, buttonId);
                    } else {
                        txMsgEvt = DsmApi_CreateResponse(me, ERROR_WRONG_PARAMETER, 0);
                    }
                    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
                    return Q_TRAN(&DsmApi_idle);
                } else {
                    error = ERROR_WRONG_SIZE;
                    break;
                }
                break;
            }
            default: {
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            }
            break;
        }
        /* --------------------------------------------- */
        case DEVICE_O_P_C_CONFIG: {
            switch (modifier) {
            case DEVICE_O_P_C_CONFIG_GET_CURRENT: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    uint16_t deviceId;
                    uint8_t channelID;
                    /* save parameters */
                    deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    channelID = DsmApi_GetReqData8(rxMsgEvt, 2);
                    /* check parameters */
                    if ((deviceId >= MIN_DEVICE) && (deviceId <= MAX_DEVICE)) {
                        DsmApi_SendOPCRequest(me, deviceId, modifier, channelID, 0, 0, 0);
                        return Q_TRAN(&DsmApi_device_output_channel);
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DEVICE_O_P_C_CONFIG_SET_CURRENT:
            case DEVICE_O_P_C_CONFIG_SET_CURRENT_AND_APPLY: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 6) {
                    uint16_t deviceId;
                    uint8_t channelID;
                    uint16_t value;
                    uint8_t size;
                    /* save parameters */
                    deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    channelID = DsmApi_GetReqData8(rxMsgEvt, 2);
                    size = DsmApi_GetReqData8(rxMsgEvt, 3);
                    value = DsmApi_GetReqData16(rxMsgEvt, 4);
                    /* check parameters */
                    if ((deviceId >= MIN_DEVICE) && (deviceId <= MAX_DEVICE)) {
                        DsmApi_SendOPCRequest(me, deviceId, modifier, channelID, value, 0, size);
                        return Q_TRAN(&DsmApi_device_output_channel);
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DEVICE_O_P_C_CONFIG_GET_SCENE: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 4) {
                    uint16_t deviceId;
                    uint8_t channelID;
                    uint8_t sceneID;
                    /* save parameters */
                    deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    channelID = DsmApi_GetReqData8(rxMsgEvt, 2);
                    sceneID = DsmApi_GetReqData8(rxMsgEvt, 3);
                    /* check parameters */
                    if ((deviceId >= MIN_DEVICE) && (deviceId <= MAX_DEVICE)) {
                        DsmApi_SendOPCRequest(me, deviceId, modifier, channelID, 0, sceneID, 0);
                        return Q_TRAN(&DsmApi_device_output_channel);
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DEVICE_O_P_C_CONFIG_GET_SCENE_CONFIG:
            case DEVICE_O_P_C_CONFIG_GET_DC_FLAGS: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    uint16_t deviceId;
                    uint8_t sceneID;
                    /* save parameters */
                    deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    sceneID = DsmApi_GetReqData8(rxMsgEvt, 2);
                    /* check parameters */
                    if ((deviceId >= MIN_DEVICE) && (deviceId <= MAX_DEVICE)) {
                        DsmApi_SendOPCRequest(me, deviceId, modifier, 0, 0, sceneID, 0);
                        return Q_TRAN(&DsmApi_device_output_channel);
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DEVICE_O_P_C_CONFIG_SET_SCENE: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 7) {
                    uint16_t deviceId;
                    uint8_t channelID;
                    uint8_t sceneID;
                    uint16_t value;
                    uint8_t size;
                    /* save parameters */
                    deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    channelID = DsmApi_GetReqData8(rxMsgEvt, 2);
                    size = DsmApi_GetReqData8(rxMsgEvt, 3);
                    sceneID = DsmApi_GetReqData8(rxMsgEvt, 4);
                    value = DsmApi_GetReqData16(rxMsgEvt, 5);
                    /* check parameters */
                    if ((deviceId >= MIN_DEVICE) && (deviceId <= MAX_DEVICE)) {
                        DsmApi_SendOPCRequest(me, deviceId, modifier, channelID, value, sceneID, size);
                        return Q_TRAN(&DsmApi_device_output_channel);
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DEVICE_O_P_C_CONFIG_SET_SCENE_CONFIG:
            case DEVICE_O_P_C_CONFIG_SET_DC_FLAGS: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 5) {
                    uint16_t deviceId;
                    uint8_t sceneID;
                    uint16_t value;
                    /* save parameters */
                    deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    sceneID = DsmApi_GetReqData8(rxMsgEvt, 2);
                    value = DsmApi_GetReqData16(rxMsgEvt, 3);
                    /* check parameters */
                    if ((deviceId >= MIN_DEVICE) && (deviceId <= MAX_DEVICE)) {
                        DsmApi_SendOPCRequest(me, deviceId, modifier, 0, value, sceneID, 1);
                        return Q_TRAN(&DsmApi_device_output_channel);
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            default: {
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            }
            break;
        }
        /* --------------------------------------------- */
        case DEVICE_O_P_C_TABLE: {
            switch (modifier) {
            case DEVICE_O_P_C_TABLE_GET_COUNT: {

                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                    DS485MessageEvt *txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 1);
                    uint16_t deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    DsmApi_SetResData8(txMsgEvt, 0, Datamodel_GetOpcChannelCount(deviceId));
                    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
                    return Q_TRAN(&DsmApi_idle);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DEVICE_O_P_C_TABLE_GET_BY_INDEX: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    uint16_t deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    uint8_t tableIndex = DsmApi_GetReqData8(rxMsgEvt, 2);
                    if (tableIndex < OPC_TABLE_LENGTH) {
                        uint8_t channelId = 0;
                        if (Datamodel_GetOpcChannelFromIndex(deviceId, tableIndex, &channelId)) {
                            DS485MessageEvt *txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 1);
                            DsmApi_SetResData8(txMsgEvt, 0, channelId);
                            QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
                            return Q_TRAN(&DsmApi_idle);
                        } else {
                            error = ERROR_WRONG_PARAMETER;
                        }
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            default: {
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            }
            break;
        }
        /* --------------------------------------------- */
        case CONTROLLER_HEATING: {
            switch (modifier) {
            case CONTROLLER_HEATING_SET_CONFIG: {
                if (DsmApi_GetDataLength(rxMsgEvt) == 24) {
                    HeatingConfig heating;
                    uint16_t zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    heating.controllerMode = DsmApi_GetReqData8(rxMsgEvt, 2);
                    heating.Kp = DsmApi_GetReqData16(rxMsgEvt, 3);
                    heating.Ts = DsmApi_GetReqData8(rxMsgEvt, 5);
                    heating.Ti = DsmApi_GetReqData16(rxMsgEvt, 6);
                    heating.Kd = DsmApi_GetReqData16(rxMsgEvt, 8);
                    heating.imin = DsmApi_GetReqData16(rxMsgEvt, 10);
                    heating.imax = DsmApi_GetReqData16(rxMsgEvt, 12);
                    heating.ymin = DsmApi_GetReqData8(rxMsgEvt, 14);
                    heating.ymax = DsmApi_GetReqData8(rxMsgEvt, 15);
                    heating.antiWindUp = DsmApi_GetReqData8(rxMsgEvt, 16);
                    heating.keepFloorWarm = DsmApi_GetReqData8(rxMsgEvt, 17);
                    heating.sourceZoneId = DsmApi_GetReqData16(rxMsgEvt, 18);
                    heating.offset = DsmApi_GetReqData16(rxMsgEvt, 20);
                    heating.manualValue = DsmApi_GetReqData8(rxMsgEvt, 22);
                    heating.emergencyValue = DsmApi_GetReqData8(rxMsgEvt, 23);
                    uint8_t error = Datamodel_setHeatingConfigOfvDC(&rxMsgEvt->destinationAddr, zoneId, &heating);
                    DS485MessageEvt *txMsgEvt = DsmApi_CreateResponse(me, error, 0);
                    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case CONTROLLER_HEATING_GET_CONFIG: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 2) {

                    uint16_t zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    HeatingConfig heating;
                    uint8_t error = Datamodel_getHeatingConfigOfvDC(&rxMsgEvt->destinationAddr, zoneId, &heating);

                    DS485MessageEvt * txMsgEvt;
                    txMsgEvt = DsmApi_CreateResponse(me, error, 22);
                    DsmApi_SetResData8(txMsgEvt,  0,  heating.controllerMode);
                    DsmApi_SetResData16(txMsgEvt, 1,  heating.Kp);
                    DsmApi_SetResData8(txMsgEvt,  3,  heating.Ts);
                    DsmApi_SetResData16(txMsgEvt, 4,  heating.Ti);
                    DsmApi_SetResData16(txMsgEvt, 6,  heating.Kd);
                    DsmApi_SetResData16(txMsgEvt, 8,  heating.imin);
                    DsmApi_SetResData16(txMsgEvt, 10, heating.imax);
                    DsmApi_SetResData8(txMsgEvt,  12, heating.ymin);
                    DsmApi_SetResData8(txMsgEvt,  13, heating.ymax);
                    DsmApi_SetResData8(txMsgEvt,  14, heating.antiWindUp);
                    DsmApi_SetResData8(txMsgEvt,  15, heating.keepFloorWarm);
                    DsmApi_SetResData16(txMsgEvt, 16, heating.sourceZoneId);
                    DsmApi_SetResData16(txMsgEvt, 18, heating.offset);
                    DsmApi_SetResData8(txMsgEvt,  20, heating.manualValue);
                    DsmApi_SetResData8(txMsgEvt,  21, heating.emergencyValue);
                    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
                    return Q_TRAN(&DsmApi_idle);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case CONTROLLER_HEATING_GET_INTERNALS: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                    uint16_t zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    uint8_t localZoneId = DsmApi_GetLocalZoneId(me, zoneId);
                    if (localZoneId != 0) {
                        HeatingInternals res;
                        if (!Heating_GetInternals(localZoneId, &res)) {
                            DS485MessageEvt * txMsgEvt;
                            /* return heating controller internals */
                            txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 26);
                            DsmApi_SetResData16(txMsgEvt, 0, res.t);
                            DsmApi_SetResData16(txMsgEvt, 2, res.tref);
                            DsmApi_SetResData16(txMsgEvt, 4, res.e);
                            DsmApi_SetResData16(txMsgEvt, 6, res.eprev);
                            DsmApi_SetResData32(txMsgEvt, 8, res.i);
                            DsmApi_SetResData32(txMsgEvt, 12, res.Yp);
                            DsmApi_SetResData32(txMsgEvt, 16, res.Yi);
                            DsmApi_SetResData32(txMsgEvt, 20, res.Yd);
                            DsmApi_SetResData8(txMsgEvt, 24, res.Y);
                            DsmApi_SetResData8(txMsgEvt, 25, res.antiWindUp);
                            QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
                            return Q_TRAN(&DsmApi_idle);
                        } else {
                            error = ERROR_REQUEST_CAN_NOT_BE_EXECUTED;
                        }
                    } else {
                        error = ERROR_ZONE_NOT_FOUND;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case CONTROLLER_HEATING_SET_OPERATION_MODES: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 34) {
                    uint16_t zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    uint16_t operationModes[HEATING_OPERATION_MODES_LENGTH];
                    DsmApi_GetReqArray16(rxMsgEvt, 2, operationModes, HEATING_OPERATION_MODES_LENGTH);
                    uint8_t errorState = Datamodel_setHeatingOperationModes(zoneId, operationModes, HEATING_OPERATION_MODES_LENGTH);
                    DS485MessageEvt *txMsgEvt = DsmApi_CreateResponse(me, errorState, 0);
                    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
                    return Q_TRAN(&DsmApi_idle);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case CONTROLLER_HEATING_GET_OPERATION_MODES: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                    uint16_t zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    uint16_t operationModes[HEATING_OPERATION_MODES_LENGTH];
                    uint8_t errorState = Datamodel_getHeatingOperationModesOfvDC(&rxMsgEvt->destinationAddr, zoneId,operationModes,HEATING_OPERATION_MODES_LENGTH);

                    /* return heating operation modes */
                    DS485MessageEvt * txMsgEvt;
                    txMsgEvt = DsmApi_CreateResponse(me, errorState, 32);
                    DsmApi_SetResArray16(txMsgEvt, 0, operationModes, HEATING_OPERATION_MODES_LENGTH);
                    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
                    return Q_TRAN(&DsmApi_idle);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case CONTROLLER_HEATING_SET_STATE: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    uint16_t zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    uint8_t localZoneId = DsmApi_GetLocalZoneId(me, zoneId);
                    if (localZoneId != 0) {
                        HeatingReqSetState *reqEvt = Q_NEW(HeatingReqSetState, HEATING_REQ_SET_STATE_SIG);
                        reqEvt->basic.localZoneId = localZoneId;
                        reqEvt->heatingState = DsmApi_GetReqData8(rxMsgEvt, 2);
                        if (Heating_IsHeatingStateValid(reqEvt->heatingState)) {
                            QActive_postFIFO(AO_SelectA, (QEvent *)reqEvt);
                            error = ERROR_OK;
                        } else {
                            error = ERROR_WRONG_PARAMETER;
                        }
                    } else {
                        error = ERROR_ZONE_NOT_FOUND;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case CONTROLLER_HEATING_GET_STATE: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                    uint16_t zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    uint8_t localZoneId = DsmApi_GetLocalZoneId(me, zoneId);
                    if (localZoneId != 0) {
                        uint8_t heatingState = Heating_GetHeatingState(localZoneId);
                        DS485MessageEvt * txMsgEvt;
                        /* return heating controller state */
                        txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 1);
                        DsmApi_SetResData8(txMsgEvt, 0, heatingState);
                        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
                        return Q_TRAN(&DsmApi_idle);
                    } else {
                        error = ERROR_ZONE_NOT_FOUND;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            default: {
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            }
            break;
        }

        /* --------------------------------------------- */
        case CIRCUIT_PING_DEVICES: {
            switch (modifier) {
            case CIRCUIT_PING_DEVICES_ONE_TIME_AUTOMATIC: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 4) {
                    ConfigurePingEvt * configurePingEvt = Q_NEW(ConfigurePingEvt, SELECT_B_CONFIGURE_PING_SIG);
                    configurePingEvt->modifier = CIRCUIT_PING_DEVICES_ONE_TIME_AUTOMATIC;
                    configurePingEvt->oneTimeInterval = DsmApi_GetReqData16(rxMsgEvt, 0);
                    configurePingEvt->autoInterval = DsmApi_GetReqData16(rxMsgEvt, 2);
                    QActive_postFIFO(AO_SelectB, (QEvent *)configurePingEvt);
                    error = ERROR_OK;
                } else {
                    error = ERROR_WRONG_PARAMETER;
                }
                break;
            }
            case CIRCUIT_PING_DEVICES_AUTOMATIC: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                    ConfigurePingEvt * configurePingEvt = Q_NEW(ConfigurePingEvt, SELECT_B_CONFIGURE_PING_SIG);
                    configurePingEvt->modifier = CIRCUIT_PING_DEVICES_AUTOMATIC;
                    configurePingEvt->oneTimeInterval = 0;
                    configurePingEvt->autoInterval = DsmApi_GetReqData16(rxMsgEvt, 0);
                    QActive_postFIFO(AO_SelectB, (QEvent *)configurePingEvt);
                    error = ERROR_OK;
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case CIRCUIT_PING_DEVICES_ONE_TIME_OFF: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                    ConfigurePingEvt * configurePingEvt = Q_NEW(ConfigurePingEvt, SELECT_B_CONFIGURE_PING_SIG);
                    configurePingEvt->modifier = CIRCUIT_PING_DEVICES_ONE_TIME_OFF;
                    configurePingEvt->oneTimeInterval = DsmApi_GetReqData16(rxMsgEvt, 0);
                    configurePingEvt->autoInterval = 0;
                    QActive_postFIFO(AO_SelectB, (QEvent *)configurePingEvt);
                    error = ERROR_OK;
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case CIRCUIT_PING_DEVICES_OFF: {
                static ConfigurePingEvt const configurePingEvt = { { SELECT_B_CONFIGURE_PING_SIG, 0 }, CIRCUIT_PING_DEVICES_OFF };
                QActive_postFIFO(AO_SelectB, (QEvent const *)&configurePingEvt);
                error = ERROR_OK;
                break;
            }
            default: {
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            }
            break;
        }
        /* --------------------------------------------- */
        case CIRCUIT_REMOVE_INACTIVE_DEVICES: {
            DatamodelReqBasicEvt * reqEvt =
                Q_NEW(DatamodelReqBasicEvt, DATAMODEL_REQ_REMOVE_INACTIVE_DEVICES_SIG);
            reqEvt->requester = (QActive *)me;
            QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
            return Q_TRAN(&DsmApi_busy);
        }
        /* --------------------------------------------- */
        case CIRCUIT_REMOVE_DEVICE: {
            if (modifier == CIRCUIT_REMOVE_DEVICE_BY_ID) {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                    DatamodelReqRemoveDeviceEvt * reqEvt =
                        Q_NEW(DatamodelReqRemoveDeviceEvt, DATAMODEL_REQ_REMOVE_DEVICE_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->deviceAddress = DsmApi_GetReqData16(rxMsgEvt, 0);
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
            } else if (modifier == CIRCUIT_REMOVE_DEVICE_BY_DSUID) {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == DSUID_LEN) {
                        DatamodelReqDeviceSerialEvt * reqEvt =
                            Q_NEW(DatamodelReqDeviceSerialEvt, DATAMODEL_REQ_DEVICE_ENTRY_BY_SERIAL_NUMBER_SIG);
                        reqEvt->basic.requester = (QActive *)me;
                        dsuid_t dsuid;
                        DsmApi_GetReqArray8(rxMsgEvt, 0, dsuid.id, DSUID_SIZE);
                        dsuid_to_string_upper(&dsuid, reqEvt->dsuid);
                        reqEvt->privateId = 0;
                        QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                        return Q_TRAN(&DsmApi_remove_by_serial_number);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
            } else {
                error = ERROR_WRONG_MODIFIER;
            }
            break;
        }
        /* --------------------------------------------- */
        case DEVICE_GROUP_MEMBERSHIP_MODIFY: {
            /* check length */
            if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                /* check modifier */
                if ((modifier == DEVICE_GROUP_MEMBERSHIP_MODIFY_ADD) || (modifier == DEVICE_GROUP_MEMBERSHIP_MODIFY_REMOVE)) {
                    DatamodelReqGroupMembershipModifyEvt * reqEvt =
                        Q_NEW(DatamodelReqGroupMembershipModifyEvt, DATAMODEL_REQ_GROUP_MEMBERSHIP_MODIFY_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->modifier = modifier;
                    reqEvt->deviceAddress = DsmApi_GetReqData16(rxMsgEvt, 0);
                    reqEvt->groupId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    reqEvt->sendDownstream = TRUE;
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_MODIFIER;
                }
            } else {
                error = ERROR_WRONG_SIZE;
            }
            break;
        }
        /* --------------------------------------------- */
        case ZONE_GROUP_MODIFY: {
            switch (modifier) {
            case ZONE_GROUP_MODIFY_ADD: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 4) {
                    DatamodelReqZoneGroupAddEvt * reqEvt =
                        Q_NEW(DatamodelReqZoneGroupAddEvt, DATAMODEL_REQ_ADD_GROUP_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    reqEvt->groupId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    reqEvt->stateMachineGroupId = DsmApi_GetReqData8(rxMsgEvt, 3);
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case ZONE_GROUP_MODIFY_REMOVE: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    DatamodelReqZoneGroupRemoveEvt * reqEvt =
                        Q_NEW(DatamodelReqZoneGroupRemoveEvt, DATAMODEL_REQ_REMOVE_GROUP_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    reqEvt->groupId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    reqEvt->dsuid = rxMsgEvt->destinationAddr;
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            default: {
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            }
            break;
        }
        /* --------------------------------------------- */
        case ZONE_GROUP_INFO: {
            /* check length */
            if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                /* check modifier */
                if ((modifier == ZONE_INFO_BY_INDEX) || (modifier == ZONE_INFO_BY_ID)) {
                    DatamodelReqZoneGroupInfoEvt * reqEvt =
                        Q_NEW(DatamodelReqZoneGroupInfoEvt, DATAMODEL_REQ_ZONE_GROUP_INFO_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->modifier = modifier;
                    reqEvt->zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    reqEvt->groupId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_wait_zone_group_info);
                } else {
                    error = ERROR_WRONG_MODIFIER;
                }
            } else {
                error = ERROR_WRONG_SIZE;
            }
            break;
        }
        /* --------------------------------------------- */
        case ZONE_GROUP_PROPERTIES: {
            switch (modifier) {
            case ZONE_GROUP_PROPERTIES_SET_NAME: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == (3 + MAX_DSM_API_NAME_LEN)) {
                    DatamodelReqGroupNameEvt * reqEvt =
                        Q_NEW(DatamodelReqGroupNameEvt, DATAMODEL_REQ_GROUP_SET_NAME_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    reqEvt->groupId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    DsmApi_GetReqArray8(rxMsgEvt, 3, reqEvt->name, MAX_NAME_LEN);
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case ZONE_GROUP_PROPERTIES_SET_STOP_TIMEOUT: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 4) {
                    DatamodelReqGroupStopTimeoutEvt * reqEvt =
                        Q_NEW(DatamodelReqGroupStopTimeoutEvt, DATAMODEL_REQ_GROUP_SET_STOP_TIMEOUT_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    reqEvt->groupId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    reqEvt->stopTimeout = DsmApi_GetReqData8(rxMsgEvt, 3);
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case ZONE_GROUP_PROPERTIES_SET_STATE_MACHINE: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 4) {
                    DatamodelReqGroupColorEvt * reqEvt =
                        Q_NEW(DatamodelReqGroupColorEvt, DATAMODEL_REQ_GROUP_SET_COLOR_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    reqEvt->groupId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    reqEvt->color = DsmApi_GetReqData8(rxMsgEvt, 3);
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            default: {
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            }
            break;
        }
        /* --------------------------------------------- */
        case ZONE_GROUP_SENSOR_PUSH: {
            /* check length */
            if (DsmApi_GetDataLength(rxMsgEvt) == (7 + DSUID_LEN)) {
                /* get parameters */
                uint16_t zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                uint8_t groupId = DsmApi_GetReqData8(rxMsgEvt, 2);
                // not used: DsmApi_GetReqArray8(rxMsgEvt, 3, dsuid, sizeof(dsuid));
                uint8_t sensorType = DsmApi_GetReqData8(rxMsgEvt, 20);
                uint16_t sensorValue = DsmApi_GetReqData16(rxMsgEvt, 21);
                uint8_t sensorPrecision = DsmApi_GetReqData8(rxMsgEvt, 23);
                /* find local zone ID */
                uint8_t localZoneId = DsmApi_GetLocalZoneId(me, zoneId);
                /* check if sensor push is for local zone */
                if (zoneId == 0 || localZoneId > 0) {
                    error = ERROR_OK;
                } else {
                    localZoneId = DATAMODEL_LOCAL_ZONE_ID_NOT_APPLICABLE;
                    error = ERROR_ZONE_NOT_FOUND;
                }
                /* publish sensor event */
                ZoneGroupSensorPushEvt * reqEvt = Q_NEW(ZoneGroupSensorPushEvt, SENSOR_PUSH_VALUE_SIG);
                reqEvt->zoneId = zoneId;
                reqEvt->localZoneId = localZoneId;
                reqEvt->groupId = groupId;
                reqEvt->sensorType = sensorType;
                reqEvt->sensorValue = sensorValue;
                reqEvt->sensorPrecision = sensorPrecision;
                reqEvt->isSourceExternal = TRUE;
                QF_publish((QEvent *)reqEvt);
            } else {
                error = ERROR_WRONG_SIZE;
            }
            break;
        }
        /* --------------------------------------------- */
        case ZONE_GROUP_ACTION_REQUEST: {
            switch (modifier) {
            case ZONE_GROUP_ACTION_REQUEST_ACTION_FORCE_CALL_SCENE:
                /* fall through */
            case ZONE_GROUP_ACTION_REQUEST_ACTION_CALL_SCENE: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 6) {
                    uint8_t groupId, sceneId;
                    /* get/check group and scene ID */
                    groupId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    sceneId = DsmApi_GetReqData8(rxMsgEvt, 5);
                    if ((groupId <= MAX_GROUP) && (sceneId < (2 * MAX_SCENES))) {
                        /* get zone ID */
                        uint16_t zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                        if (zoneId == 0) {
                            /* all zones */
                            SceneZoneGroupEvt * sceneZoneGroupEvt;
                            /* send broadcast call scene event to select A */
                            sceneZoneGroupEvt = Q_NEW(SceneZoneGroupEvt, CALL_SCENE_BROADCAST_SIG);
                            sceneZoneGroupEvt->localZoneId = 0;
                            sceneZoneGroupEvt->groupId = groupId;
                            sceneZoneGroupEvt->basic.sceneId = sceneId;
                            sceneZoneGroupEvt->basic.sourceZoneId = 0;
                            sceneZoneGroupEvt->basic.sourceGroupId = 0;
                            sceneZoneGroupEvt->basic.force = ((modifier == ZONE_GROUP_ACTION_REQUEST_ACTION_FORCE_CALL_SCENE) ? TRUE : FALSE);
                            sceneZoneGroupEvt->basic.outputChannel = 0;
                            QActive_postFIFO(AO_SelectA, (QEvent *)sceneZoneGroupEvt);
                            error = ERROR_OK;
                        } else {
                            /* find local zone ID */
                            uint8_t localZoneId = DsmApi_GetLocalZoneId(me, zoneId);
                            /* local zone ID found */
                            if (localZoneId != 0) {
                                SceneZoneGroupEvt * sceneZoneGroupEvt;
                                if (groupId >= MIN_GROUP) {
                                    /* send call scene zone/group event to select A */
                                    sceneZoneGroupEvt = Q_NEW(SceneZoneGroupEvt, CALL_SCENE_ZONE_GROUP_SIG);
                                } else {
                                    /* send broadcast call scene event to select A */
                                    sceneZoneGroupEvt = Q_NEW(SceneZoneGroupEvt, CALL_SCENE_BROADCAST_SIG);
                                }
                                sceneZoneGroupEvt->localZoneId = localZoneId;
                                sceneZoneGroupEvt->groupId = groupId;
                                sceneZoneGroupEvt->basic.sceneId = sceneId;
                                sceneZoneGroupEvt->basic.sourceZoneId = 0;
                                sceneZoneGroupEvt->basic.sourceGroupId = 0;
                                sceneZoneGroupEvt->basic.force = ((modifier == ZONE_GROUP_ACTION_REQUEST_ACTION_FORCE_CALL_SCENE) ? TRUE : FALSE);
                                sceneZoneGroupEvt->basic.outputChannel = 0;
                                QActive_postFIFO(AO_SelectA, (QEvent *)sceneZoneGroupEvt);
                                error = ERROR_OK;
                            } else {
                                error = ERROR_ZONE_NOT_FOUND;
                            }
                        }
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case ZONE_GROUP_ACTION_REQUEST_ACTION_SAVE_SCENE: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 6) {
                    uint16_t zoneId;
                    uint8_t groupId, sceneId, localZoneId;
                    /* get/check group and scene ID */
                    groupId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    sceneId = DsmApi_GetReqData8(rxMsgEvt, 5);
                    if ((groupId <= MAX_GROUP) && (sceneId < (2 * MAX_SCENES))) {
                        /* get/check zone ID */
                        zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                        localZoneId = 0;
                        if ((zoneId == 0) || ((localZoneId = DsmApi_GetLocalZoneId(me, zoneId)) != 0)) {
                            ActionZoneGroupSceneEvt * reqEvt = Q_NEW(ActionZoneGroupSceneEvt, SAVE_SCENE_ZONE_SIG);
                            reqEvt->localZoneId = localZoneId;
                            reqEvt->groupId = groupId;
                            reqEvt->sceneId = sceneId;
                            QActive_postFIFO(AO_SelectA, (QEvent *)reqEvt);
                            error = ERROR_OK;
                        } else {
                            error = ERROR_ZONE_NOT_FOUND;
                        }
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case ZONE_GROUP_ACTION_REQUEST_ACTION_LOCAL_STOP:
                /* fall through */
            case ZONE_GROUP_ACTION_REQUEST_ACTION_UNDO_SCENE:
                /* fall through */
            case ZONE_GROUP_ACTION_REQUEST_ACTION_BLINK:
                /* fall through */
            case ZONE_GROUP_ACTION_REQUEST_ACTION_SELECT_GROUP: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 5) {
                    uint16_t zoneId;
                    uint8_t groupId, localZoneId;
                    /* get/check group ID */
                    groupId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    if (groupId <= MAX_GROUP) {
                        /* get/check zone ID */
                        zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                        /* find local zone ID */
                        localZoneId = DsmApi_GetLocalZoneId(me, zoneId);
                        if ((zoneId == 0) || (localZoneId != 0)) {
                            QSignal signal = BLINK_ZONE_SIG;
                            switch (modifier) {
                            case ZONE_GROUP_ACTION_REQUEST_ACTION_UNDO_SCENE:
                                signal = UNDO_SCENE_ZONE_SIG;
                                break;
                            case ZONE_GROUP_ACTION_REQUEST_ACTION_SELECT_GROUP:
                                signal = SELECT_ZONE_GROUP_SIG;
                                break;
                            case ZONE_GROUP_ACTION_REQUEST_ACTION_LOCAL_STOP:
                                signal = LOCAL_STOP_SIG;
                                break;
                            }
                            ActionZoneEvt * reqEvt = Q_NEW(ActionZoneEvt, signal);
                            reqEvt->localZoneId = localZoneId;
                            reqEvt->groupId = groupId;
                            QActive_postFIFO(AO_SelectA, (QEvent *)reqEvt);
                            error = ERROR_OK;
                        } else {
                            error = ERROR_ZONE_NOT_FOUND;
                        }
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case ZONE_GROUP_ACTION_REQUEST_ACTION_PROGRAMMING_MODE_ON:
                /* fall through */
            case ZONE_GROUP_ACTION_REQUEST_ACTION_PROGRAMMING_MODE_OFF: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 5) {
                    uint16_t zoneId;
                    uint8_t groupId;
                    /* get/check group ID */
                    groupId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    if (groupId <= MAX_GROUP) {
                        /* get zone ID */
                        zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                        QSignal signal = (modifier == ZONE_GROUP_ACTION_REQUEST_ACTION_PROGRAMMING_MODE_ON) ? PROG_ON_ZONE_SIG : PROG_OFF_ZONE_SIG;
                        ProgZoneEvt * reqEvt = Q_NEW(ProgZoneEvt, signal);
                        reqEvt->zoneId = zoneId;
                        reqEvt->groupId = groupId;
                        QActive_postFIFO(AO_SelectA, (QEvent *)reqEvt);
                        error = ERROR_OK;
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case ZONE_GROUP_ACTION_REQUEST_ACTION_UNDO_SCENE_NUMBER: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 6) {
                    uint8_t groupId, sceneId;
                    /* get/check group and scene ID */
                    groupId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    sceneId = DsmApi_GetReqData8(rxMsgEvt, 5);
                    if ((groupId <= MAX_GROUP) && (sceneId < (2 * MAX_SCENES))) {
                        /* get zone ID */
                        uint16_t zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                        if (zoneId == 0) {
                            /* all zones */
                            SceneZoneGroupEvt * sceneZoneGroupEvt;
                            /* send broadcast undo scene number event to select A */
                            sceneZoneGroupEvt = Q_NEW(SceneZoneGroupEvt, UNDO_SCENE_NUMBER_BROADCAST_SIG);
                            sceneZoneGroupEvt->localZoneId = 0;
                            sceneZoneGroupEvt->groupId = groupId;
                            sceneZoneGroupEvt->basic.sceneId = sceneId;
                            sceneZoneGroupEvt->basic.sourceZoneId = 0;
                            sceneZoneGroupEvt->basic.sourceGroupId = 0;
                            sceneZoneGroupEvt->basic.force = FALSE;
                            sceneZoneGroupEvt->basic.outputChannel = 0;
                            QActive_postFIFO(AO_SelectA, (QEvent *)sceneZoneGroupEvt);
                            error = ERROR_OK;
                        } else {
                            /* find local zone ID */
                            uint8_t localZoneId = DsmApi_GetLocalZoneId(me, zoneId);
                            /* local zone ID found */
                            if (localZoneId != 0) {
                                SceneZoneGroupEvt * sceneZoneGroupEvt;
                                if (groupId >= MIN_GROUP) {
                                    /* send undo scene number zone/group event to select A */
                                    sceneZoneGroupEvt = Q_NEW(SceneZoneGroupEvt, UNDO_SCENE_NUMBER_ZONE_GROUP_SIG);
                                } else {
                                    /* send broadcast undo scene number event to select A */
                                    sceneZoneGroupEvt = Q_NEW(SceneZoneGroupEvt, UNDO_SCENE_NUMBER_BROADCAST_SIG);
                                }
                                sceneZoneGroupEvt->localZoneId = localZoneId;
                                sceneZoneGroupEvt->groupId = groupId;
                                sceneZoneGroupEvt->basic.sceneId = sceneId;
                                sceneZoneGroupEvt->basic.sourceZoneId = 0;
                                sceneZoneGroupEvt->basic.sourceGroupId = 0;
                                sceneZoneGroupEvt->basic.force = FALSE;
                                sceneZoneGroupEvt->basic.outputChannel = 0;
                                QActive_postFIFO(AO_SelectA, (QEvent *)sceneZoneGroupEvt);
                                error = ERROR_OK;
                            } else {
                                error = ERROR_ZONE_NOT_FOUND;
                            }
                        }
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case ZONE_GROUP_ACTION_REQUEST_ACTION_SET_OUTVAL: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 6) {
                    /* get/check group ID */
                    uint8_t groupId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    if (groupId <= MAX_GROUP) {
                        /* get zone ID */
                        uint16_t zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                        /* find local zone ID */
                        uint8_t localZoneId = DsmApi_GetLocalZoneId(me, zoneId);
                        if ((zoneId == 0) || (localZoneId != 0)) {
                            SetOutvalZoneGroupEvt * reqEvt = Q_NEW(SetOutvalZoneGroupEvt, SET_OUTVAL_ZONE_GROUP_SIG);
                            reqEvt->localZoneId = localZoneId;
                            reqEvt->groupId = groupId;
                            reqEvt->outval = DsmApi_GetReqData8(rxMsgEvt, 5);
                            QActive_postFIFO(AO_SelectA, (QEvent *)reqEvt);
                            error = ERROR_OK;
                        } else {
                            error = ERROR_ZONE_NOT_FOUND;
                        }
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case ZONE_GROUP_ACTION_REQUEST_ACTION_SET_SCENE_LOCAL_PRIORITY:
                /* fall through */
            case ZONE_GROUP_ACTION_REQUEST_ACTION_REGISTER_SCENE_LOCAL_PRIORITY:
                /* fall through */
            case ZONE_GROUP_ACTION_REQUEST_ACTION_CALL_SCENE_MIN: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 6) {
                    uint16_t zoneId;
                    uint8_t groupId, sceneId, localZoneId;
                    /* get/check group and scene ID */
                    groupId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    sceneId = DsmApi_GetReqData8(rxMsgEvt, 5);
                    if ((groupId <= MAX_GROUP) && (sceneId < (2 * MAX_SCENES))) {
                        /* get/check zone ID */
                        zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                        localZoneId = DsmApi_GetLocalZoneId(me, zoneId);
                        if ((localZoneId != 0) || (zoneId == 0)) {
                            QSignal signal= SET_SCENE_LOCAL_PRIO_BROADCAST_SIG;
                            switch (modifier) {
                            case ZONE_GROUP_ACTION_REQUEST_ACTION_REGISTER_SCENE_LOCAL_PRIORITY:
                                signal = REGISTER_SCENE_LOCAL_PRIO_BROADCAST_SIG;
                                break;
                            case ZONE_GROUP_ACTION_REQUEST_ACTION_CALL_SCENE_MIN:
                                signal = CALL_SCENE_MIN_BROADCAST_SIG;
                                break;
                            }
                            ActionZoneGroupSceneEvt * reqEvt = Q_NEW(ActionZoneGroupSceneEvt, signal);
                            reqEvt->localZoneId = localZoneId;
                            reqEvt->groupId = groupId;
                            reqEvt->sceneId = sceneId;
                            QActive_postFIFO(AO_SelectA, (QEvent *)reqEvt);
                            error = ERROR_OK;
                        } else {
                            error = ERROR_ZONE_NOT_FOUND;
                        }
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case ZONE_GROUP_ACTION_REQUEST_ACTION_OPC_INC:
            case ZONE_GROUP_ACTION_REQUEST_ACTION_OPC_DEC:
            case ZONE_GROUP_ACTION_REQUEST_ACTION_OPC_STOP: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 6) {
                    uint8_t groupId, channelId;
                    /* get/check group and scene ID */
                    groupId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    channelId = DsmApi_GetReqData8(rxMsgEvt, 5);
                    if (groupId <= MAX_GROUP) {
                        /* get zone ID */
                        uint16_t zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                        /* find local zone ID */
                        uint8_t localZoneId = DsmApi_GetLocalZoneId(me, zoneId);
                        if ((zoneId == 0) || (localZoneId != 0)) {
                            OPCOperation op;
                            switch (modifier) {
                            case ZONE_GROUP_ACTION_REQUEST_ACTION_OPC_INC:
                                op = OPC_INC;
                                break;
                            case ZONE_GROUP_ACTION_REQUEST_ACTION_OPC_DEC:
                                op = OPC_DEC;
                                break;
                            case ZONE_GROUP_ACTION_REQUEST_ACTION_OPC_STOP:
                                op = OPC_STOP;
                                break;
                            }
                            OPCDimZoneGroupEvt * opcDimZoneGroupEvt;
                            /* send broadcast call event to select A */
                            opcDimZoneGroupEvt = Q_NEW(OPCDimZoneGroupEvt, OPC_DIM_ZONE_GROUP_SIG);
                            opcDimZoneGroupEvt->localZoneId = localZoneId;
                            opcDimZoneGroupEvt->groupId = groupId;
                            opcDimZoneGroupEvt->channelId = channelId;
                            opcDimZoneGroupEvt->operation = op;
                            QActive_postFIFO(AO_SelectA, (QEvent *)opcDimZoneGroupEvt);
                            error = ERROR_OK;
                        } else {
                        error = ERROR_ZONE_NOT_FOUND;
                        }
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            default: {
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            }
            break;
        }
        /* --------------------------------------------- */
        case DEVICE_ACTION_REQUEST: {
            switch (modifier) {
            case ZONE_GROUP_ACTION_REQUEST_ACTION_FORCE_CALL_SCENE:
                /* fall through */
            case ZONE_GROUP_ACTION_REQUEST_ACTION_CALL_SCENE: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    uint16_t deviceId;
                    uint8_t sceneId;
                    /* get/check device and scene ID */
                    deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    sceneId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    if ((deviceId >= MIN_DEVICE) && (deviceId <= MAX_DEVICE) && (sceneId < (2 * MAX_SCENES))) {
                        SceneDeviceEvt * callSceneDeviceEvt;
                        callSceneDeviceEvt = Q_NEW(SceneDeviceEvt, CALL_SCENE_DEVICE_SIG);
                        callSceneDeviceEvt->deviceId = deviceId;
                        callSceneDeviceEvt->sceneId = sceneId;
                        callSceneDeviceEvt->force = ((modifier == ZONE_GROUP_ACTION_REQUEST_ACTION_FORCE_CALL_SCENE) ? TRUE : FALSE);
                        QActive_postFIFO(AO_SelectA, (QEvent *)callSceneDeviceEvt);
                        error = ERROR_OK;
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                }
                break;
            }
            case ZONE_GROUP_ACTION_REQUEST_ACTION_SAVE_SCENE: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    uint16_t deviceId;
                    uint8_t sceneId;
                    /* get/check device and scene ID */
                    deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    sceneId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    if ((deviceId >= MIN_DEVICE) && (deviceId <= MAX_DEVICE) && (sceneId < (2 * MAX_SCENES))) {
                        SaveSceneDeviceEvt * reqEvt = Q_NEW(SaveSceneDeviceEvt, SAVE_SCENE_DEVICE_SIG);
                        reqEvt->deviceId = deviceId;
                        reqEvt->sceneId = sceneId;
                        QActive_postFIFO(AO_SelectA, (QEvent *)reqEvt);
                        error = ERROR_OK;
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case ZONE_GROUP_ACTION_REQUEST_ACTION_UNDO_SCENE:
                /* fall through */
            case ZONE_GROUP_ACTION_REQUEST_ACTION_BLINK:
                /* fall through */
            case ZONE_GROUP_ACTION_REQUEST_ACTION_PROGRAMMING_MODE_ON:
                /* fall through */
            case ZONE_GROUP_ACTION_REQUEST_ACTION_PROGRAMMING_MODE_OFF: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                    uint16_t deviceId;
                    /* get/check device ID */
                    deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    if ((deviceId >= MIN_DEVICE) && (deviceId <= MAX_DEVICE)) {
                        QSignal signal = BLINK_DEVICE_SIG;
                        switch (modifier) {
                        case ZONE_GROUP_ACTION_REQUEST_ACTION_UNDO_SCENE:
                            signal = UNDO_SCENE_DEVICE_SIG;
                            break;
                        case ZONE_GROUP_ACTION_REQUEST_ACTION_PROGRAMMING_MODE_ON:
                            signal = PROG_ON_DEVICE_SIG;
                            break;
                        case ZONE_GROUP_ACTION_REQUEST_ACTION_PROGRAMMING_MODE_OFF:
                            signal = PROG_OFF_DEVICE_SIG;
                            break;
                        }
                        ActionDeviceEvt * reqEvt = Q_NEW(ActionDeviceEvt, signal);
                        reqEvt->deviceId = deviceId;
                        QActive_postFIFO(AO_SelectA, (QEvent *)reqEvt);
                        error = ERROR_OK;
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case ZONE_GROUP_ACTION_REQUEST_ACTION_UNDO_SCENE_NUMBER: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    uint16_t deviceId;
                    uint8_t sceneId;
                    /* get/check device and scene ID */
                    deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    sceneId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    if ((deviceId >= MIN_DEVICE) && (deviceId <= MAX_DEVICE) && (sceneId < (2 * MAX_SCENES))) {
                        SceneDeviceEvt * reqEvt = Q_NEW(SceneDeviceEvt, UNDO_SCENE_NUMBER_DEVICE_SIG);
                        reqEvt->deviceId = deviceId;
                        reqEvt->sceneId = sceneId;
                        QActive_postFIFO(AO_SelectA, (QEvent *)reqEvt);
                        error = ERROR_OK;
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case ZONE_GROUP_ACTION_REQUEST_ACTION_SET_OUTVAL: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    uint16_t deviceId;
                    /* get/check device ID */
                    deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    if ((deviceId >= MIN_DEVICE) && (deviceId <= MAX_DEVICE)) {
                        SetOutvalDeviceEvt * reqEvt = Q_NEW(SetOutvalDeviceEvt, SET_OUTVAL_DEVICE_SIG);
                        reqEvt->deviceId = deviceId;
                        reqEvt->outval = DsmApi_GetReqData8(rxMsgEvt, 2);
                        QActive_postFIFO(AO_SelectA, (QEvent *)reqEvt);
                        error = ERROR_OK;
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case ZONE_GROUP_ACTION_REQUEST_ACTION_OPC_INC:
            case ZONE_GROUP_ACTION_REQUEST_ACTION_OPC_DEC:
            case ZONE_GROUP_ACTION_REQUEST_ACTION_OPC_STOP: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 3) {
                    uint16_t deviceId;
                    uint8_t channelId;
                    /* get/check device ID */
                    deviceId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    if ((deviceId >= MIN_DEVICE) && (deviceId <= MAX_DEVICE)) {
                        /* get/check channel ID */
                        channelId = DsmApi_GetReqData8(rxMsgEvt, 2);
                        OPCOperation op;
                        switch (modifier) {
                        case ZONE_GROUP_ACTION_REQUEST_ACTION_OPC_INC:
                            op = OPC_INC;
                            break;
                        case ZONE_GROUP_ACTION_REQUEST_ACTION_OPC_DEC:
                            op = OPC_DEC;
                            break;
                        case ZONE_GROUP_ACTION_REQUEST_ACTION_OPC_STOP:
                            op = OPC_STOP;
                            break;
                        }
                        OPCDimDeviceEvt * opcDimDeviceEvt;
                        /* send broadcast call event to select A */
                        opcDimDeviceEvt = Q_NEW(OPCDimDeviceEvt, OPC_DIM_DEVICE_SIG);
                        opcDimDeviceEvt->deviceId = deviceId;
                        opcDimDeviceEvt->channelId = channelId;
                        opcDimDeviceEvt->operation = op;
                        QActive_postFIFO(AO_SelectA, (QEvent *)opcDimDeviceEvt);
                        error = ERROR_OK;
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            default: {
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            }
            break;
        }
        /* --------------------------------------------- */
        case ZONE_GROUP_SCENE_INFO: {
            /* check length */
            if (DsmApi_GetDataLength(rxMsgEvt) == 4) {
                DatamodelReqSceneInfoEvt * reqEvt =
                    Q_NEW(DatamodelReqSceneInfoEvt, DATAMODEL_REQ_SCENE_INFO_SIG);
                reqEvt->basic.requester = (QActive *)me;
                reqEvt->zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                reqEvt->groupId = DsmApi_GetReqData8(rxMsgEvt, 2);
                reqEvt->sceneIndex = DsmApi_GetReqData8(rxMsgEvt, 3);
                QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                return Q_TRAN(&DsmApi_wait_scene_info);
            } else {
                error = ERROR_WRONG_SIZE;
            }
            break;
        }
        /* --------------------------------------------- */
        case ZONE_GROUP_SCENE_PROPERTIES:
            switch (modifier) {
            case ZONE_GROUP_SCENE_PROPERTIES_SET_NAME:
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == (4 + MAX_DSM_API_NAME_LEN)) {
                    DatamodelReqSceneNameEvt * reqEvt =
                        Q_NEW(DatamodelReqSceneNameEvt, DATAMODEL_REQ_SCENE_SET_NAME_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->zoneId = DsmApi_GetReqData16(rxMsgEvt, 0);
                    reqEvt->groupId = DsmApi_GetReqData8(rxMsgEvt, 2);
                    reqEvt->sceneIndex = DsmApi_GetReqData8(rxMsgEvt, 3);
                    DsmApi_GetReqArray8(rxMsgEvt, 4, reqEvt->name, MAX_NAME_LEN);
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            default:
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            break;
        /* --------------------------------------------- */
        case DSM_CONFIG: {
            switch (modifier) {
            case DSM_CONFIG_SET_DSM_CONFIG: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == (sizeof(dsid_t)+4)) {
                    DatamodelReqSetDsmConfig * reqEvt = Q_NEW(DatamodelReqSetDsmConfig, DATAMODEL_REQ_SET_DSM_CONFIG_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    DsmApi_GetReqArray8(rxMsgEvt, 0, &reqEvt->dsmid, sizeof(dsid_t));
                    reqEvt->hwRevision = DsmApi_GetReqData32(rxMsgEvt, sizeof(dsid_t));
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DSM_CONFIG_GET_HASH: {
                DatamodelReqBasicEvt * reqEvt =
                    Q_NEW(DatamodelReqBasicEvt, DATAMODEL_REQ_GET_HASH_SIG);
                reqEvt->requester = (QActive *)me;
                QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                return Q_TRAN(&DsmApi_wait_hash);
            }
            default: {
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            }
            break;
        }
        /* --------------------------------------------- */
        case DSM_STATISTICS: {
            switch (modifier) {
            case DSM_STATISTICS_GET_POOL_MARGINS: {
                DS485MessageEvt * txMsgEvt;
                /* create response */
                txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 23);
                DsmApi_SetResData8(txMsgEvt, 0, QF_getPoolMargin(1));
                DsmApi_SetResData8(txMsgEvt, 1, QF_getPoolMargin(2));
                DsmApi_SetResData8(txMsgEvt, 2, QF_getPoolMargin(3));
                DsmApi_SetResData8(txMsgEvt, 3, QF_getQueueMargin(1));
                DsmApi_SetResData8(txMsgEvt, 4, QF_getQueueMargin(2));
                DsmApi_SetResData8(txMsgEvt, 5, QF_getQueueMargin(3));
                DsmApi_SetResData8(txMsgEvt, 6, QF_getQueueMargin(4));
                DsmApi_SetResData8(txMsgEvt, 7, QF_getQueueMargin(5));
                DsmApi_SetResData8(txMsgEvt, 8, QF_getQueueMargin(6));
                DsmApi_SetResData8(txMsgEvt, 9, 0);
                DsmApi_SetResData8(txMsgEvt, 10, 0);
                DsmApi_SetResData8(txMsgEvt, 11, 0);
                DsmApi_SetResData8(txMsgEvt, 12, QEQueue_getNFree(&l_dsmApi.deferredQueue));
                DsmApi_SetResData8(txMsgEvt, 13, DS485_GetTxMessageQueueMargin());
                DsmApi_SetResData8(txMsgEvt, 14, DS485_GetDeferredTxMessageQueueMargin());
                DsmApi_SetResData8(txMsgEvt, 15, 0);
                DsmApi_SetResData8(txMsgEvt, 16, MMI_GetDeferredQueueMargin());
                DsmApi_SetResData8(txMsgEvt, 17, SelectA_GetDeferredQueueMargin());
                DsmApi_SetResData8(txMsgEvt, 18, SelectB_GetDeferredCommandQueueMargin());
                DsmApi_SetResData8(txMsgEvt, 19, SelectB_GetDeferredDsidQueueMargin());
                DsmApi_SetResData8(txMsgEvt, 20, 0);
                DsmApi_SetResData8(txMsgEvt, 21, 0);
                DsmApi_SetResData8(txMsgEvt, 22, 0);
                /* send response */
                QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
                return Q_TRAN(&DsmApi_idle);
            }
            case DSM_STATISTICS_GET_STOP_TIMER: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 1) {
                    DS485MessageEvt * txMsgEvt;
                    uint8_t timerIndex = DsmApi_GetReqData8(rxMsgEvt, 0);
                    StopTimerEvt const * timer = SelectA_GetStopTimerByIndex(timerIndex);
                    if (timer != NULL) {
                        /* create response */
                        txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 8);
                        DsmApi_SetResData8(txMsgEvt, 0, timer->localZoneId);
                        DsmApi_SetResData8(txMsgEvt, 1, timer->groupId);
                        DsmApi_SetResData8(txMsgEvt, 2, timer->buttonId);
                        DsmApi_SetResData8(txMsgEvt, 3, Datamodel_GetStopTimeoutZoneGroup(timer->localZoneId, timer->groupId));
                        DsmApi_SetResData32(txMsgEvt, 4, timer->super.ctr);
                        /* send response */
                        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
                        return Q_TRAN(&DsmApi_idle);
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            default: {
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            }
            break;
        }
        /* --------------------------------------------- */
        case BUS_MEMBER: {
            /* modifier: BUS_MEMBER_GET_TYPE */
            DS485MessageEvt * txMsgEvt;
            /* create response */
            txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 1);
            uint8_t type = Datamodel_GetBusMemberType(&rxMsgEvt->destinationAddr);
            DsmApi_SetResData8(txMsgEvt, 0, type);
            /* send response */
            QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
            return Q_TRAN(&DsmApi_idle);
        }
        /* --------------------------------------------- */
        case DSM_DEBUG_MODE: {
            switch (modifier) {
            case DSM_DEBUG_MODE_STOP_SELECT_B: {
                static const QEvent stopEvt = { SELECT_B_STOP_SIG, 0 };
                QActive_postFIFO(AO_SelectB, &stopEvt);
                error = ERROR_OK;
                break;
            }
            case DSM_DEBUG_MODE_RESTART_REGISTRATION: {
                SelectB_sendRestartRegistrationSignal();
                error = ERROR_OK;
                break;
            }
            case DSM_DEBUG_MODE_GET_COMMUNICATION_LOG_STATE: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 0) {
                    uint8_t state = (me->communicationLog.upstreamEnabled ? DSMAPI_COMM_LOG_UPSTREAM_MASK : 0);
                    state |= (me->communicationLog.upstreamEnabled ? DSMAPI_COMM_LOG_DOWNSTREAM_MASK : 0);
                    DS485MessageEvt * txMsgEvt;
                    /* create response */
                    txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 1);
                    DsmApi_SetResData8(txMsgEvt, 0, state);
                    /* send response */
                    QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
                    return Q_TRAN(&DsmApi_idle);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DSM_DEBUG_MODE_SET_COMMUNICATION_LOG_STATE: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 1) {
                    uint8_t state = DsmApi_GetReqData8(rxMsgEvt, 0);
                    me->communicationLog.upstreamEnabled = (((state & DSMAPI_COMM_LOG_UPSTREAM_MASK) != 0) ? TRUE : FALSE);
                    me->communicationLog.downstreamEnabled = (((state & DSMAPI_COMM_LOG_DOWNSTREAM_MASK) != 0) ? TRUE : FALSE);
                    error = ERROR_OK;
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DSM_DEBUG_MODE_DISABLE_JUMBLE: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 1) {
                    uint8_t disable = DsmApi_GetReqData8(rxMsgEvt, 0);
                    DisableSelectBPartsEvt * reqEvt = Q_NEW(DisableSelectBPartsEvt, SELECT_B_DISABLE_JUMBLE_SIG);
                    reqEvt->disable = ((disable != 0) ? TRUE : FALSE);
                    QActive_postFIFO(AO_SelectB, (QEvent *)reqEvt);
                    error = ERROR_OK;
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DSM_DEBUG_MODE_INJECT_SHORT_UPSTREAM: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 9) {
                    /* allocate short upstream event */
                    ShortUpstreamEvt * e = Q_NEW(ShortUpstreamEvt, SHORT_UPSTREAM_SIG);
                    e->source = DsmApi_GetReqData16(rxMsgEvt, 0);
                    e->circuit = DsmApi_GetReqData8(rxMsgEvt, 2);
                    e->key = DsmApi_GetReqData8(rxMsgEvt, 3);
                    e->click = DsmApi_GetReqData8(rxMsgEvt, 4);
                    e->sensor = DsmApi_GetReqData8(rxMsgEvt, 5);
                    e->resend = DsmApi_GetReqData8(rxMsgEvt, 6);
                    e->quality = DsmApi_GetReqData8(rxMsgEvt, 7);
                    e->flags = DsmApi_GetReqData8(rxMsgEvt, 8);
                    /* publish short upstream event */
                    QF_publish((QEvent*)e);
                    error = ERROR_OK;
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DSM_DEBUG_MODE_INJECT_LONG_UPSTREAM: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 11) {
                    /* allocate long upstream event */
                    LongUpstreamEvt * e = Q_NEW(LongUpstreamEvt, LONG_UPSTREAM_SIG);
                    e->source = DsmApi_GetReqData16(rxMsgEvt, 0);
                    e->circuit = DsmApi_GetReqData8(rxMsgEvt, 2);
                    e->mainqualifier = DsmApi_GetReqData8(rxMsgEvt, 3);
                    e->subqualifier = DsmApi_GetReqData8(rxMsgEvt, 4);
                    e->data = DsmApi_GetReqData32(rxMsgEvt, 5);
                    e->quality = DsmApi_GetReqData8(rxMsgEvt, 9);
                    e->flags = DsmApi_GetReqData8(rxMsgEvt, 10);
                    /* publish long upstream event */
                    QF_publish((QEvent*)e);
                    error = ERROR_OK;
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            default:
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            break;
        }
        /* --------------------------------------------- */
        case DSM_INFO: {
            DatamodelReqDsmEvt * reqEvt =
                Q_NEW(DatamodelReqDsmEvt, DATAMODEL_REQ_DSM_INFO_SIG);
            reqEvt->requester = (QActive *)me;
            reqEvt->dsuid = rxMsgEvt->destinationAddr;
            QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
            return Q_TRAN(&DsmApi_wait_dsm_info);
        }
        /* --------------------------------------------- */
        case DSM_PROPERTIES: {
            switch (modifier) {
            case DSM_PROPERTIES_SET_NAME: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == MAX_DSM_API_NAME_LEN) {
                    DatamodelReqNameEvt * reqEvt =
                        Q_NEW(DatamodelReqNameEvt, DATAMODEL_REQ_DSM_SET_NAME_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->dsuid = rxMsgEvt->destinationAddr;
                    DsmApi_GetReqArray8(rxMsgEvt, 0, reqEvt->name, MAX_NAME_LEN);
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DSM_PROPERTIES_SET_APARTMENT_ID: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 1) {
                    uint8_t apartmentId = DsmApi_GetReqData8(rxMsgEvt, 0);
                    if ((apartmentId >= MIN_APARTMENT_ID) && (apartmentId <= MAX_APARTMENT_ID)){
                        DatamodelReqSetApartmentIdEvt * reqEvt =
                            Q_NEW(DatamodelReqSetApartmentIdEvt, DATAMODEL_REQ_DSM_SET_APARTMENT_ID_SIG);
                        reqEvt->basic.requester = (QActive *)me;
                        reqEvt->apartmentId = apartmentId;
                        reqEvt->dsuid = rxMsgEvt->destinationAddr;
                        QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                        return Q_TRAN(&DsmApi_busy);
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DSM_PROPERTIES_SET_TEMPERATURE_LIMITS: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                    DatamodelReqSetTemperatureLimitsEvt * reqEvt =
                        Q_NEW(DatamodelReqSetTemperatureLimitsEvt, DATAMODEL_REQ_DSM_SET_TEMPERATURE_LIMITS_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->warnTemperature = DsmApi_GetReqData8(rxMsgEvt, 0);
                    reqEvt->errorTemperature = DsmApi_GetReqData8(rxMsgEvt, 1);
                    reqEvt->dsuid = rxMsgEvt->destinationAddr;
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DSM_PROPERTIES_GET_APARTMENT_ID: {
                DatamodelReqDsmEvt * reqEvt =
                    Q_NEW(DatamodelReqDsmEvt, DATAMODEL_REQ_DSM_PROPERTIES_SIG);
                reqEvt->requester = (QActive *)me;
                reqEvt->dsuid = rxMsgEvt->destinationAddr;
                QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                return Q_TRAN(&DsmApi_wait_dsm_apartment_id);
            }
            case DSM_PROPERTIES_GET_TEMPERATURE_LIMITS: {
                DatamodelReqDsmEvt * reqEvt =
                    Q_NEW(DatamodelReqDsmEvt, DATAMODEL_REQ_DSM_PROPERTIES_SIG);
                reqEvt->requester = (QActive *)me;
                reqEvt->dsuid = rxMsgEvt->destinationAddr;
                QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                return Q_TRAN(&DsmApi_wait_dsm_temperature_limits);
            }
            case DSM_PROPERTIES_GET_CURRENT_TEMPERATURE: {
                DatamodelReqDsmEvt * reqEvt =
                    Q_NEW(DatamodelReqDsmEvt, DATAMODEL_REQ_DSM_PROPERTIES_SIG);
                reqEvt->requester = (QActive *)me;
                reqEvt->dsuid = rxMsgEvt->destinationAddr;
                QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                return Q_TRAN(&DsmApi_wait_current_temperature);
            }
            case DSM_PROPERTIES_SET_FLAGS: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 1) {
                    DatamodelReqSetDsmFlagsEvt * reqEvt =
                        Q_NEW(DatamodelReqSetDsmFlagsEvt, DATAMODEL_REQ_DSM_SET_DSM_FLAGS_SIG);
                    reqEvt->basic.requester = (QActive *)me;
                    reqEvt->flags = DsmApi_GetReqData8(rxMsgEvt, 0);
                    reqEvt->dsuid = rxMsgEvt->destinationAddr;
                    QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                    return Q_TRAN(&DsmApi_busy);
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DSM_PROPERTIES_GET_FLAGS: {
                DatamodelReqDsmEvt * reqEvt =
                    Q_NEW(DatamodelReqDsmEvt, DATAMODEL_REQ_DSM_PROPERTIES_SIG);
                reqEvt->requester = (QActive *)me;
                reqEvt->dsuid = rxMsgEvt->destinationAddr;
                QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                return Q_TRAN(&DsmApi_wait_dsm_flags);
            }
            case DSM_PROPERTIES_SET_UPSTREAM_FREQUENCY: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 1) {
                    uint8_t upstreamFrequency = DsmApi_GetReqData8(rxMsgEvt, 0);
                    if (upstreamFrequency <= UPSTREAM_FREQUENCY_MAX) {
                        DatamodelReqSetDsmUpstreamFreqEvt * reqEvt =
                            Q_NEW(DatamodelReqSetDsmUpstreamFreqEvt, DATAMODEL_REQ_DSM_SET_UPSTREAM_FREQUENCY_SIG);
                        reqEvt->basic.requester = (QActive *)me;
                        reqEvt->upstreamFrequency = upstreamFrequency;
                        reqEvt->dsuid = rxMsgEvt->destinationAddr;
                        QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                        return Q_TRAN(&DsmApi_busy);
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DSM_PROPERTIES_GET_UPSTREAM_FREQUENCY: {
                DatamodelReqDsmEvt * reqEvt =
                    Q_NEW(DatamodelReqDsmEvt, DATAMODEL_REQ_DSM_PROPERTIES_SIG);
                reqEvt->requester = (QActive *)me;
                reqEvt->dsuid = rxMsgEvt->destinationAddr;
                QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                return Q_TRAN(&DsmApi_wait_dsm_upstream_frequency);
            }
            case DSM_PROPERTIES_GET_POWER_DOWN_DURATION: {
                DS485MessageEvt * txMsgEvt;
                uint8_t state = me->longPowerDown ? 1 : 0;
                /* create response */
                txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 1);
                DsmApi_SetResData8(txMsgEvt, 0, state);
                /* send response */
                QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
                return Q_TRAN(&DsmApi_idle);
            }
            case DSM_PROPERTIES_SET_PING_INTERVALS: {
                /* check length */
                if (DsmApi_GetDataLength(rxMsgEvt) == 4) {
                    uint16_t autoPingInterval = DsmApi_GetReqData16(rxMsgEvt, 0);
                    uint16_t oneTimePingInterval = DsmApi_GetReqData16(rxMsgEvt, 2);
                    if ((autoPingInterval >= MIN_PING_INTERVAL_MS) && (oneTimePingInterval >= MIN_PING_INTERVAL_MS)) {
                        DatamodelReqSetPingIntervalsEvt * reqEvt =
                            Q_NEW(DatamodelReqSetPingIntervalsEvt, DATAMODEL_REQ_DSM_SET_PING_INTERVALS_SIG);
                        reqEvt->basic.requester = (QActive *)me;
                        reqEvt->autoPingInterval = autoPingInterval;
                        reqEvt->oneTimePingInterval = oneTimePingInterval;
                        reqEvt->dsuid = rxMsgEvt->destinationAddr;
                        QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                        return Q_TRAN(&DsmApi_busy);
                    } else {
                        error = ERROR_WRONG_PARAMETER;
                    }
                } else {
                    error = ERROR_WRONG_SIZE;
                }
                break;
            }
            case DSM_PROPERTIES_GET_PING_INTERVALS : {
                DatamodelReqDsmEvt * reqEvt =
                    Q_NEW(DatamodelReqDsmEvt, DATAMODEL_REQ_DSM_PROPERTIES_SIG);
                reqEvt->requester = (QActive *)me;
                reqEvt->dsuid = rxMsgEvt->destinationAddr;
                QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                return Q_TRAN(&DsmApi_wait_ping_intervals);
            }
            case DSM_PROPERTIES_GET_MAX_TEMPERATURE: {
                DatamodelReqDsmEvt * reqEvt =
                    Q_NEW(DatamodelReqDsmEvt, DATAMODEL_REQ_DSM_PROPERTIES_SIG);
                reqEvt->requester = (QActive *)me;
                reqEvt->dsuid = rxMsgEvt->destinationAddr;
                QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
                return Q_TRAN(&DsmApi_wait_max_temperature);
            }
            case DSM_PROPERTIES_RESET_MAX_TEMPERATURE: {
                static QEvent const resetMaxTempEvt = { DATAMODEL_RESET_MAX_TEMP_SIG, 0 };
                QActive_postFIFO(AO_Datamodel, &resetMaxTempEvt);
                error = ERROR_OK;
                break;
            }
            case DSM_PROPERTIES_GET_APARTMENT_STATE: {
                ApartmentState_t state = SelectA_GetApartmentState();
                DS485MessageEvt * txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 1);
                DsmApi_SetResData8(txMsgEvt, 0, (uint8_t)state);
                QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
                return Q_TRAN(&DsmApi_idle);
            }
            default: {
                error = ERROR_WRONG_MODIFIER;
                break;
            }
            }
            break;
        }
        /* --------------------------------------------- */
        case DSM_STATE: {
            MMIReqStateEvt * reqEvt = Q_NEW(MMIReqStateEvt, MMI_REQ_STATE_SIG);
            reqEvt->requester = (QActive *)me;
            QActive_postFIFO(AO_MMI, (QEvent *)reqEvt);
            return Q_TRAN(&DsmApi_wait_dsm_state);
        }
        /* --------------------------------------------- */
        case DSM_REBOOT: {
            if (me->sendResponse) {
                /* send the response (OK) */
                QActive_postFIFO(AO_DS485, (QEvent *)DsmApi_CreateResponse(me, ERROR_OK, 0));
            }
            return Q_TRAN(&DsmApi_system_reset);
        }
        /* --------------------------------------------- */
        case DSM_FACTORY_RESET: {
            /* check length */
            if (DsmApi_GetDataLength(rxMsgEvt) == 2) {
                uint16_t password = DsmApi_GetReqData16(rxMsgEvt, 0);
                if (password == 0xCAFE) {
                    static const QEvent reqEvt = { DATAMODEL_REQ_FACTORY_RESET_SIG, 0 };
                    QActive_postFIFO(AO_Datamodel, &reqEvt);
                    if (me->sendResponse) {
                        /* send the response (OK) */
                        QActive_postFIFO(AO_DS485, (QEvent *)DsmApi_CreateResponse(me, ERROR_OK, 0));
                    }
                    /* go to stopped */
                    return Q_TRAN(&DsmApi_stopped);
                } else {
                    error = ERROR_WRONG_PARAMETER;
                }
            } else {
                error = ERROR_WRONG_SIZE;
            }
            break;
        }
        /* --------------------------------------------- */
        }
        if (me->sendResponse) {
            QActive_postFIFO(AO_DS485, (QEvent *)DsmApi_CreateResponse(me, error, 0));
        }
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_top);
}

/*..........................................................................*/
static QState DsmApi_busy(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DS485_RX_DATA_SIG: {
        /* defer only requests */
        if (DS485_IsRequest((DS485MessageEvt const *)e)) {
            /* defer event */
            QActive_defer((QActive *)me, &me->deferredQueue, e);
            return Q_HANDLED();
        }
        break;
    }
    case DATAMODEL_RES_ERROR_SIG:
        if (me->sendResponse) {
            /* send the response (error) */
            QActive_postFIFO(AO_DS485, (QEvent *)DsmApi_CreateResponse(me, ((DatamodelErrorEvt const *)e)->error, 0));
        }
        return Q_TRAN(&DsmApi_idle);
    }
    return Q_SUPER(&DsmApi_top);
}

/*..........................................................................*/
static QState DsmApi_device_info(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_DEVICE_ENTRY_SIG: {
        uint8_t size, offset;
        dsuid_t dsuid;
        static const uint8_t dataLength = 16 +
                                          sizeof(((DatamodelResDeviceEntryEvt *)0)->device.groupMembership) +
                                          MAX_DSM_API_NAME_LEN +
                                          sizeof(dsuid);
        DatamodelResDeviceEntryEvt * const resEvt = (DatamodelResDeviceEntryEvt * const)e;
        DS485MessageEvt * txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, dataLength);
        DsmApi_SetResData16(txMsgEvt, 0, resEvt->device.deviceId);
        DsmApi_SetResData16(txMsgEvt, 2, resEvt->device.vid);
        DsmApi_SetResData16(txMsgEvt, 4, 0);
        DsmApi_SetResData16(txMsgEvt, 6, resEvt->device.fid); // used by remote programming mode (by index)
        DsmApi_SetResData16(txMsgEvt, 8, resEvt->device.ver);
        DsmApi_SetResData16(txMsgEvt,10, me->zoneIds[resEvt->device.localZoneId]);
        DsmApi_SetResData8 (txMsgEvt,12, resEvt->state);
        DsmApi_SetResData8 (txMsgEvt,13, (resEvt->device.locked) ? 1 : 0);
        DsmApi_SetResData8 (txMsgEvt,14, resEvt->device.mode); // used by remote programming mode (by index)
        DsmApi_SetResData8 (txMsgEvt,15, resEvt->device.ltmode);
        offset = 16;
        size = sizeof(resEvt->device.groupMembership);
        DsmApi_SetResArray8(txMsgEvt, offset, resEvt->device.groupMembership, size); // used by remote programming mode (by index)
        offset += size;
        DsmApi_SetResName(txMsgEvt, offset, resEvt->device.name);
        offset += MAX_DSM_API_NAME_LEN;
        dsuid_from_string(resEvt->device.dsuid, &dsuid);
        DsmApi_SetResArray8(txMsgEvt, offset, &dsuid, sizeof(dsuid));
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_button_info(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_DEVICE_ENTRY_SIG: {
        uint8_t offset;
        DatamodelResDeviceEntryEvt * const resEvt = (DatamodelResDeviceEntryEvt * const)e;
        DS485MessageEvt * txMsgEvt;
        if (Fid_HasButton(resEvt->device.fid)) {
            txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 6);
            /* separate Button ID and Group Membership */
            offset=0;
            DsmApi_SetResData8(txMsgEvt, offset, Datamodel_DeviceButtonId(&resEvt->device));
            offset += 1;
            DsmApi_SetResData8(txMsgEvt, offset, Datamodel_DeviceButtonGroup(&resEvt->device));
            offset += 1;
            DsmApi_SetResData8(txMsgEvt, offset, Datamodel_DeviceButtonActiveGroup(&resEvt->device));
            offset += 1;
            DsmApi_SetResData8(txMsgEvt, offset, resEvt->device.flags);
            offset += 1;
            DsmApi_SetResData8(txMsgEvt, offset, resEvt->device.buttonOutputChannel);
            offset += 1;
            DsmApi_SetResData8(txMsgEvt, offset, resEvt->device.buttonGlobalGroup);
        } else {
            txMsgEvt = DsmApi_CreateResponse(me, ERROR_WRONG_PARAMETER, 0);
        }
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_binary_input_count(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_DEVICE_ENTRY_SIG: {
        uint8_t index = 0;
        DatamodelResDeviceEntryEvt * const resEvt = (DatamodelResDeviceEntryEvt * const)e;
        DS485MessageEvt * txMsgEvt;
        txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 1);
        while ((index < BINARY_INPUT_TABLE_LENGTH) && (resEvt->device.binaryInputTable[index].present == 1)) {
            index++;
        }
        DsmApi_SetResData8(txMsgEvt, 0, index);
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_binary_input_table(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_DEVICE_ENTRY_SIG: {
        DatamodelResDeviceEntryEvt * const resEvt = (DatamodelResDeviceEntryEvt * const)e;
        DS485MessageEvt * txMsgEvt;
        if (resEvt->device.binaryInputTable[me->tableIndex].present == 1) {
            uint8_t offset = 0;
            txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, sizeof(BinaryInputConfig));
            DsmApi_SetResData8(txMsgEvt, offset++, resEvt->device.binaryInputTable[me->tableIndex].targetGroup >> 6);
            DsmApi_SetResData8(txMsgEvt, offset++, (resEvt->device.binaryInputTable[me->tableIndex].targetGroup & 0x3F));
            DsmApi_SetResData8(txMsgEvt, offset++, resEvt->device.binaryInputTable[me->tableIndex].type);
            DsmApi_SetResData8(txMsgEvt, offset++, resEvt->device.binaryInputTable[me->tableIndex].NUM);
        } else {
            txMsgEvt = DsmApi_CreateResponse(me, ERROR_WRONG_PARAMETER, 0);
        }
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_wait_sensor_count(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_DEVICE_ENTRY_SIG: {
        uint8_t index = 0;
        DatamodelResDeviceEntryEvt * const resEvt = (DatamodelResDeviceEntryEvt * const)e;
        DS485MessageEvt * txMsgEvt;
        txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 1);
        while ((index < SENSOR_TABLE_LENGTH) && (resEvt->device.sensorTable[index].type != 0xff)) {
            index++;
        }
        DsmApi_SetResData8(txMsgEvt, 0, index);
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_wait_sensor_table_entry(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_DEVICE_ENTRY_SIG: {
        DatamodelResDeviceEntryEvt * const resEvt = (DatamodelResDeviceEntryEvt * const)e;
        DS485MessageEvt * txMsgEvt;
        if (resEvt->device.sensorTable[me->tableIndex].type != 0xff) {
            /* return sensor entry */
            uint8_t offset = 0;
            txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 4);
            DsmApi_SetResData8(txMsgEvt, offset++, resEvt->device.sensorTable[me->tableIndex].type);
            DsmApi_SetResData32(txMsgEvt, offset, resEvt->device.sensorTable[me->tableIndex].pollInterval);
            offset = offset + 4;
            DsmApi_SetResData8(txMsgEvt, offset++, resEvt->device.sensorTable[me->tableIndex].broadcast);
            DsmApi_SetResData8(txMsgEvt, offset++, resEvt->device.sensorTable[me->tableIndex].pushConvert);
        } else {
            txMsgEvt = DsmApi_CreateResponse(me, ERROR_WRONG_PARAMETER, 0);
        }
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_device_parameter(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_DEVICE_ENTRY_SIG: {
        DatamodelResDeviceEntryEvt * const resEvt = (DatamodelResDeviceEntryEvt * const)e;

        if (me->deviceParameterInfo.operation == DEVICE_CONFIG_GET)
        {
            /* allocate get device parameter event */
            DeviceParameterEvt * reqEvt =
                Q_NEW(DeviceParameterEvt, SELECT_B_GET_DEVICE_PARAMETER_SIG);
            /* set request parameters */
            reqEvt->deviceId = resEvt->device.deviceId;
            reqEvt->deviceSerial = 0;
            reqEvt->paramOffset = me->deviceParameterInfo.offset;
            reqEvt->paramBank = me->deviceParameterInfo.bank;
            QActive_postFIFO(AO_SelectB, (QEvent *)reqEvt);
        }
        else if (me->deviceParameterInfo.operation == DEVICE_CONFIG_SET)
        {
            /* allocate set device parameter event */
            DeviceParameterEvt * reqEvt =
                Q_NEW(DeviceParameterEvt, SELECT_B_SET_DEVICE_PARAMETER_SIG);
            /* set request parameters */
            reqEvt->deviceId = resEvt->device.deviceId;
            reqEvt->deviceSerial = 0;
            reqEvt->paramOffset = me->deviceParameterInfo.offset;
            reqEvt->paramBank = me->deviceParameterInfo.bank;
            reqEvt->value = me->deviceParameterInfo.value;
            reqEvt->sendDoneModifier = EVENT_DSM_DONE_SET_DEVICE_CONFIG;
            QActive_postFIFO(AO_SelectB, (QEvent *)reqEvt);
        }
        QActive_postFIFO(AO_DS485, (QEvent *)DsmApi_CreateResponse(me, ERROR_OK, 0));
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_device_output_channel(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_DEVICE_ENTRY_SIG: {
        DatamodelResDeviceEntryEvt * const resEvt = (DatamodelResDeviceEntryEvt * const)e;

        {
            /* allocate get device parameter event */
            DeviceOutpuChannelEvt * reqEvt =
                Q_NEW(DeviceOutpuChannelEvt, SELECT_B_OUTPUT_CHANNEL_SIG);
            /* set request parameters */
            reqEvt->deviceId = resEvt->device.deviceId;
            reqEvt->deviceSerial = 0;
            reqEvt->channel = me->deviceParameterInfo.channel;
            reqEvt->scene = me->deviceParameterInfo.scene;
            reqEvt->value = me->deviceParameterInfo.value;
            reqEvt->operation = me->deviceParameterInfo.operation;
            reqEvt->is16bit = me->deviceParameterInfo.is16bit;
            QActive_postFIFO(AO_SelectB, (QEvent *)reqEvt);
        }
        QActive_postFIFO(AO_DS485, (QEvent *)DsmApi_CreateResponse(me, ERROR_OK, 0));
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_device_output_channel_wait_count(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_DEVICE_ENTRY_SIG: {
        uint8_t index = 0;
        DatamodelResDeviceEntryEvt * const resEvt = (DatamodelResDeviceEntryEvt * const)e;
        DS485MessageEvt * txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 1);
        if (FidGetSubclass(resEvt->device.fid) == FID_SUBCLASS_PRE_R200) {
            index = 1;
            while ((index < OPC_TABLE_LENGTH) && (resEvt->device.opcTable[index].channelId != 0xff)) {
                index++;
            }
        }
        DsmApi_SetResData8(txMsgEvt, 0, index);
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_device_output_channel_wait_table_entry(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_DEVICE_ENTRY_SIG: {
        DatamodelResDeviceEntryEvt * const resEvt = (DatamodelResDeviceEntryEvt * const)e;
        DS485MessageEvt * txMsgEvt;
        uint8_t channelId = resEvt->device.opcTable[me->tableIndex].channelId;
        if ((FidGetSubclass(resEvt->device.fid) == FID_SUBCLASS_PRE_R200) && ((me->tableIndex == 0) || (channelId != 0xff))) {
            /* return output channel id */
            txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 1);
            DsmApi_SetResData8(txMsgEvt, 0, channelId);
        } else {
            txMsgEvt = DsmApi_CreateResponse(me, ERROR_WRONG_PARAMETER, 0);
        }
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_device_state(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_DEVICE_STATE_SIG: {
        DatamodelResDeviceStateEvt * const resEvt = (DatamodelResDeviceStateEvt * const)e;
        DS485MessageEvt * txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 1);
        DsmApi_SetResData8(txMsgEvt, 0, (resEvt->active ? 1 : 0 ));
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}
/*..........................................................................*/
static QState DsmApi_zone_count(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_ZONE_COUNT_SIG: {
        DS485MessageEvt * txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 1);
        DsmApi_SetResData8(txMsgEvt, 0, ((DatamodelResZoneCountEvt const *)e)->zoneCount);
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_zone_device_count(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_ZONE_DEVICE_COUNT_SIG: {
        /* send the response */
        DS485MessageEvt * txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 2);
        /* device count */
        DsmApi_SetResData16(txMsgEvt, 0, ((DatamodelResZoneDeviceCountEvt const *)e)->deviceCount);
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_zone_info(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_ZONE_INFO_SIG: {
        /* send the response */
        DatamodelResZoneInfoEvt const * resZoneInfoEvt = (DatamodelResZoneInfoEvt const *)e;
        DS485MessageEvt * txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, (4 + MAX_DSM_API_NAME_LEN));
        /* zone ID */
        DsmApi_SetResData16(txMsgEvt, 0, resZoneInfoEvt->zoneId);
        DsmApi_SetResData8(txMsgEvt, 2, resZoneInfoEvt->localZoneId);
        DsmApi_SetResData8(txMsgEvt, 3, resZoneInfoEvt->groupCount);
        DsmApi_SetResName(txMsgEvt, 4, resZoneInfoEvt->name);
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_wait_scene_info(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_SCENE_INFO_SIG: {
        DatamodelResSceneInfoEvt const * resEvt = (DatamodelResSceneInfoEvt const *)e;
        DS485MessageEvt * txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, MAX_DSM_API_NAME_LEN);
        DsmApi_SetResName(txMsgEvt, 0, resEvt->name);
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}
/*..........................................................................*/
static QState DsmApi_wait_zone_group_info(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_ZONE_GROUP_INFO_SIG: {
        DatamodelResZoneGroupInfoEvt const * resEvt = (DatamodelResZoneGroupInfoEvt const *)e;
        DS485MessageEvt * txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 6 + MAX_DSM_API_NAME_LEN);
        DsmApi_SetResData8(txMsgEvt, 0, resEvt->groupId);
        DsmApi_SetResData8(txMsgEvt, 1, resEvt->stateMachineGroupId);
        DsmApi_SetResData16(txMsgEvt, 2, resEvt->countDevices);
        DsmApi_SetResName(txMsgEvt, 4, resEvt->name);
        DsmApi_SetResData8(txMsgEvt, 4 + MAX_DSM_API_NAME_LEN, resEvt->stopTimeout);
        DsmApi_SetResData8(txMsgEvt, 4 + MAX_DSM_API_NAME_LEN + 1, resEvt->remoteStopTimeout);
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_wait_zone_history(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case SELECT_A_RES_ZONE_SCENE_HISTORY_SIG: {
        ZoneSceneHistoryResEvt const * resEvt = (ZoneSceneHistoryResEvt const *)e;
        DS485MessageEvt * txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 1 + (MAX_ZONE_SCENE_HISTORY * 2));
        uint8_t index, offset;
        offset = 0;
        DsmApi_SetResData8(txMsgEvt, offset, resEvt->count);
        ++offset;
        for (index = 0; index < resEvt->count; ++index) {
            DsmApi_SetResData8(txMsgEvt, offset, resEvt->groupId[index]);
            ++offset;
            DsmApi_SetResData8(txMsgEvt, offset, resEvt->sceneId[index]);
            ++offset;
        }
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_wait_dsm_info(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_DSM_INFO_SIG: {
        DatamodelResDsmInfoEvt const * resEvt = (DatamodelResDsmInfoEvt const *)e;
        DS485MessageEvt * txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 26 + MAX_DSM_API_NAME_LEN);
        DsmApi_SetResData32(txMsgEvt,  0, resEvt->hwVersion);
        DsmApi_SetResData32(txMsgEvt,  4, resEvt->armSwVersion);
        DsmApi_SetResData32(txMsgEvt,  8, 0);
        DsmApi_SetResData16(txMsgEvt, 12, DSM_API_VERSION);
//        DsmApi_SetResArray8(txMsgEvt, 14, &resEvt->dsid, 12);
        DsmApi_SetResName(txMsgEvt, 26, resEvt->name);
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_wait_dsm_apartment_id(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_DSM_PROPERTIES_SIG: {
        DatamodelResDsmProperties const * resEvt = (DatamodelResDsmProperties const *)e;
        DS485MessageEvt * txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 1);
        DsmApi_SetResData8(txMsgEvt, 0, resEvt->apartmentId);
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_wait_dsm_temperature_limits(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_DSM_PROPERTIES_SIG: {
        DatamodelResDsmProperties const * resEvt = (DatamodelResDsmProperties const *)e;
        DS485MessageEvt * txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 2);
        DsmApi_SetResData8(txMsgEvt, 0, resEvt->warnTemperature);
        DsmApi_SetResData8(txMsgEvt, 1, resEvt->errorTemperature);
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_wait_dsm_flags(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_DSM_PROPERTIES_SIG: {
        DatamodelResDsmProperties const * resEvt = (DatamodelResDsmProperties const *)e;
        DS485MessageEvt * txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 1);
        DsmApi_SetResData8(txMsgEvt, 0, resEvt->flags);
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_wait_dsm_upstream_frequency(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_DSM_PROPERTIES_SIG: {
        DatamodelResDsmProperties const * resEvt = (DatamodelResDsmProperties const *)e;
        DS485MessageEvt * txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 1);
        DsmApi_SetResData8(txMsgEvt, 0, resEvt->upstreamFrequency);
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_wait_ping_intervals(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_DSM_PROPERTIES_SIG: {
        DatamodelResDsmProperties const * resEvt = (DatamodelResDsmProperties const *)e;
        DS485MessageEvt * txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 4);
        DsmApi_SetResData16(txMsgEvt, 0, resEvt->autoPingInterval);
        DsmApi_SetResData16(txMsgEvt, 2, resEvt->oneTimePingInterval);
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_wait_dsm_state(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case MMI_RES_STATE_SIG: {
        MMIResStateEvt const * resEvt = (MMIResStateEvt const *)e;
        DS485MessageEvt * txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 6);
        DsmApi_SetResData8(txMsgEvt, 0, resEvt->hardware);
        DsmApi_SetResData8(txMsgEvt, 1, 0);
        DsmApi_SetResData8(txMsgEvt, 2, resEvt->temperature);
        DsmApi_SetResData8(txMsgEvt, 3, resEvt->config);
        DsmApi_SetResData8(txMsgEvt, 4, resEvt->ds485);
        DsmApi_SetResData8(txMsgEvt, 5, resEvt->line);
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_wait_current_temperature(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_DSM_PROPERTIES_SIG: {
        DatamodelResDsmProperties const * resEvt = (DatamodelResDsmProperties const *)e;
        DS485MessageEvt * txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 2);
        DsmApi_SetResData16(txMsgEvt, 0, resEvt->currentTemperature);
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_wait_max_temperature(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_DSM_PROPERTIES_SIG: {
        DatamodelResDsmProperties const * resEvt = (DatamodelResDsmProperties const *)e;
        DS485MessageEvt * txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 2);
        DsmApi_SetResData16(txMsgEvt, 0, resEvt->maxTemperature);
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_wait_vdc_user_protobuf_message_request(DsmApi *me, QEvent const *e)
{

    switch (e->sig) {
    case SELECT_B_USER_PROTOBUF_MESSAGE_RESPONSE_SIG: {
        DS485NetlibPacketMessageEvt *txMsgEvt;
        SelectBProtobufMessageResponseEvt const * resEvt =
            (SelectBProtobufMessageResponseEvt const *)e;

        txMsgEvt = DsmApi_CreateLongResponse(me, ERROR_OK, resEvt->size+2);
        txMsgEvt->message[DSMAPI_LONG_RES_DATA_OFFSET]   = ((resEvt->size >> 8) & 0xff);
        txMsgEvt->message[DSMAPI_LONG_RES_DATA_OFFSET+1] = (resEvt->size & 0xff);

        // 2: offset for 2 byte length
        DsmApi_SetLongResArray8(txMsgEvt, 2, resEvt->response, resEvt->size);

        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    case SELECT_B_USER_PROTOBUF_MESSAGE_TIMEOUT_SIG: {
        QActive_postFIFO(AO_DS485, (QEvent *)DsmApi_CreateLongResponse(me, ERROR_TIMEOUT, 0));
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_wait_hash(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_GET_HASH_SIG: {
        DatamodelResHashEvt const * resEvt = (DatamodelResHashEvt const *)e;
        DS485MessageEvt * txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 12);
        DsmApi_SetResData32(txMsgEvt, 0, resEvt->hash);
        DsmApi_SetResData32(txMsgEvt, 4, resEvt->changeCount);
        DsmApi_SetResData32(txMsgEvt, 8, SelectA_GetBinaryInputEventCount());
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_wait_room_states(DsmApi *me, QEvent const *e)
{
    switch(e->sig) {
    case SELECT_A_RES_ROOM_STATES_SIG: {
        ZoneStatesResEvt const * resEvt = (ZoneStatesResEvt const *)e;
        DS485MessageEvt * txMsgEvt = DsmApi_CreateResponse(me, ERROR_OK, 10);
        DsmApi_SetResData16(txMsgEvt, 0, resEvt->roomStates);
        DsmApi_SetResData16(txMsgEvt, 2, resEvt->area1States);
        DsmApi_SetResData16(txMsgEvt, 4, resEvt->area2States);
        DsmApi_SetResData16(txMsgEvt, 6, resEvt->area3States);
        DsmApi_SetResData16(txMsgEvt, 8, resEvt->area4States);
        QActive_postFIFO(AO_DS485, (QEvent *)txMsgEvt);
        return Q_TRAN(&DsmApi_idle);
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_remove_by_serial_number(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DATAMODEL_RES_DEVICE_ENTRY_SIG: {
        /* device is known */
        DatamodelResDeviceEntryEvt const * resEvt = (DatamodelResDeviceEntryEvt const *)e;
        DatamodelReqRemoveDeviceEvt * reqEvt =
            Q_NEW(DatamodelReqRemoveDeviceEvt, DATAMODEL_REQ_REMOVE_DEVICE_SIG);
        reqEvt->basic.requester = (QActive *)me;
        reqEvt->deviceAddress = resEvt->device.deviceId;
        QActive_postFIFO(AO_Datamodel, (QEvent *)reqEvt);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&DsmApi_busy);
}

/*..........................................................................*/
static QState DsmApi_system_reset(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* wait 1 second */
        QTimeEvt_postIn(&me->timer, (QActive *)me, (1000/BSP_MS_PER_TICKS));
        return Q_HANDLED();
    }
    case DSMAPI_TIMEOUT_SIG: {
        /* publish system termination signal to reset the board */
        QF_publish(&l_sysTermEvt);
        return Q_TRAN(&DsmApi_stopped);
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState DsmApi_system_reset_and_update(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* update application after reset */
        /* wait 1 second */
        QTimeEvt_postIn(&me->timer, (QActive *)me, (1000/BSP_MS_PER_TICKS));
        return Q_HANDLED();
    }
    case DSMAPI_TIMEOUT_SIG: {
        /* publish system termination signal to reset the board */
        QF_publish(&l_sysTermEvt);
        return Q_TRAN(&DsmApi_stopped);
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState DsmApi_stopped(DsmApi *me, QEvent const *e)
{
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static uint8_t DsmApi_GetLocalZoneId(DsmApi *me, uint16_t zoneId)
{
    uint8_t localZoneId;
    for (localZoneId = 0; localZoneId <= MAX_ZONE; ++localZoneId) {
        if (me->zoneIds[localZoneId] == zoneId) {
            return localZoneId;
        }
    }
    return 0;
}

/*..........................................................................*/
static DS485MessageEvt * DsmApi_CreateResponse(DsmApi *me, uint8_t error, uint8_t dataLength)
{
    /* create message with invalid source address is OK, as dS485 will update source address before sending */
    if (dsuid_is_broadcast(&me->destinationAddress)) {
        me->destinationAddress = Datamodel_GetDsmid();
    }
    DS485MessageEvt * txMsgEvt = DS485_CreateDataMessageFromSrcDestDsuid(&me->requestAddress, &me->destinationAddress, 0, DS485_TYPE_ACK_RES,
                                                                         dataLength + DSMAPI_RESPONSE_HEADER_LEN);
    txMsgEvt->message[DSMAPI_TRANSACTIONID_OFFSET] = me->transactionId;
    txMsgEvt->message[DSMAPI_MSGID_OFFSET]         = me->messageId;
    txMsgEvt->message[DSMAPI_MODIFIER_OFFSET]      = me->modifierId;
    txMsgEvt->message[DSMAPI_ERROR_OFFSET]         = error;
    return txMsgEvt;
}

/*..........................................................................*/
static DS485NetlibPacketMessageEvt * DsmApi_CreateLongResponse(DsmApi *me, uint8_t error, uint16_t dataLength)
{
    if (dsuid_is_broadcast(&me->destinationAddress)) {
        me->destinationAddress = Datamodel_GetDsmid();
    }
    /* create message with invalid source address is OK, as dS485 will update source address before sending */
    DS485NetlibPacketMessageEvt * txMsgEvt = DS485_CreateLongDataMessage(&me->requestAddress, &me->destinationAddress, 0, DS485_TYPE_ACK_RES,
                                                                         dataLength + DSMAPI_LONG_RESPONSE_HEADER_LEN);
    txMsgEvt->message[DSMAPI_LONG_TRANSACTIONID_OFFSET] = me->transactionId;
    txMsgEvt->message[DSMAPI_LONG_MSGID_OFFSET]         = me->messageId;
    txMsgEvt->message[DSMAPI_LONG_MODIFIER_OFFSET]      = me->modifierId;
    txMsgEvt->message[DSMAPI_LONG_ERROR_OFFSET]         = error;
    return txMsgEvt;
}
/*..........................................................................*/
static DS485MessageEvt * DsmApi_CreateRequestFromDsid(dsuid_t const *destination, uint16_t source, uint8_t transactionId, uint8_t msgid, uint8_t modifier, uint8_t dataLength)
{
    /* create message with invalid source address is OK, as dS485 will update source address before sending */
    DS485MessageEvt * txMsgEvt = DS485_CreateDataMessageFromId(destination, source, 0, DS485_TYPE_REQ, dataLength + DSMAPI_REQUEST_HEADER_LEN);
    txMsgEvt->message[DSMAPI_TRANSACTIONID_OFFSET] = transactionId;
    txMsgEvt->message[DSMAPI_MSGID_OFFSET]         = msgid;
    txMsgEvt->message[DSMAPI_MODIFIER_OFFSET]      = modifier;
    return txMsgEvt;
}

/*..........................................................................*/
static DS485MessageEvt * DsmApi_CreateRequestFromVdcDsuid(dsuid_t const *destination, dsuid_t const *vdcDsuid, uint8_t transactionId, uint8_t msgid, uint8_t modifier, uint8_t dataLength)
{
    DS485MessageEvt * txMsgEvt = DS485_CreateDataMessageFromSrcDestDsuid(destination, vdcDsuid, 0, DS485_TYPE_REQ, dataLength + DSMAPI_REQUEST_HEADER_LEN);
    txMsgEvt->message[DSMAPI_TRANSACTIONID_OFFSET] = transactionId;
    txMsgEvt->message[DSMAPI_MSGID_OFFSET]         = msgid;
    txMsgEvt->message[DSMAPI_MODIFIER_OFFSET]      = modifier;
    return txMsgEvt;
}

/*..........................................................................*/
static DS485MessageEvt * DsmApi_CreateEvent(uint8_t destination, uint16_t source, uint8_t transactionId, uint8_t msgid, uint8_t modifier, uint8_t dataLength)
{
    /* create message with invalid source address is OK, as dS485 will update source address before sending */
    DS485MessageEvt * txMsgEvt = DS485_CreateDataMessageFromId(&l_broadcastDsuid, source, 0, DS485_TYPE_EVENT, dataLength + DSMAPI_EVENT_HEADER_LEN);
    txMsgEvt->message[DSMAPI_TRANSACTIONID_OFFSET] = transactionId;
    txMsgEvt->message[DSMAPI_MSGID_OFFSET]         = msgid;
    txMsgEvt->message[DSMAPI_MODIFIER_OFFSET]      = modifier;
    txMsgEvt->message[DSMAPI_ERROR_OFFSET]         = 0;
    return txMsgEvt;
}

/*..........................................................................*/
static DS485MessageEvt * DsmApi_CreateDsmDoneEvent(uint16_t source, uint8_t modifier, uint8_t error, uint8_t dataLength)
{
    /* create message with invalid source address is OK, as dS485 will update source address before sending */
    DS485MessageEvt * txMsgEvt = DS485_CreateDataMessageFromId(&l_broadcastDsuid, source, 0, DS485_TYPE_EVENT, dataLength + DSMAPI_EVENT_HEADER_LEN);
    txMsgEvt->message[DSMAPI_TRANSACTIONID_OFFSET] = DSM_TRANSACTION_ID;
    txMsgEvt->message[DSMAPI_MSGID_OFFSET]         = EVENT_DSM_DONE;
    txMsgEvt->message[DSMAPI_MODIFIER_OFFSET]      = modifier;
    txMsgEvt->message[DSMAPI_ERROR_OFFSET]         = error;
    return txMsgEvt;
}

